/* CouchDB init & cleanup functions
 *
 * CouchDB C API (libcouchdb) - Copyright (c) 2006 Dirk Schalge
 *
 * Thanks to Damien Katz and all other involved developers for this great
 * database!
 *
 * See doc/LICENSE for licensing details.
 */

#include "libcouchdb.h"
#include "couch.h"
#include <curl/curl.h>
#include <libxml/parser.h>

Couch *couch_init ( char *server, char *db )
{
	Couch *new;

	if ( !server )
	{
		errno = EINVAL;
		return NULL;
	}

	new = calloc ( 1, sizeof(Couch) );
	if ( !new )
		return NULL;

	new->server = strdup ( server );
	if ( !new->server )
	{
		couch_cleanup ( &new );
		return NULL;
	}

	if ( db )
	{
		new->db = strdup ( db );
		if ( !new->db )
		{
			couch_cleanup ( &new );
			return NULL;
		}
	}

	new->result = calloc ( 1, sizeof(CouchResult) );
	if ( !new->result )
	{
		couch_cleanup ( &new );
		return NULL;
	}

	new->curl = curl_easy_init ( );
	if ( !new->curl )
	{
		couch_cleanup ( &new );
		return NULL;
	}

	return new;
}

void couch_cleanup ( Couch **c )
{
	if ( !c )
	{
		errno = EINVAL;
		return;
	}

	if ( !(*c) )
	{
		/* uninitialized couch */
		errno = 0;
		return;
	}

	if ( (*c)->server )
		free ( (*c)->server );

	if ( (*c)->db )
		free ( (*c)->db );

	if ( (*c)->result )
		couch_result_cleanup ( &(*c)->result );

	if ( (*c)->xml )
		free ( (*c)->xml );

	/* cleanup libcurl & libxml2 stuff */
	curl_easy_cleanup ( (*c)->curl );
	curl_global_cleanup ( );
	xmlCleanupParser ( );

	free ( (*c) );
	(*c) = NULL;
}

int couch_db_use ( Couch *c, char *db )
{
	if ( !c || !db )
	{
		errno = EINVAL;
		return -1;
	}

	c->db = strdup ( db );
	if ( !c->db )
		return -1;

	return 1;
}

void couch_result_clear ( Couch *c )
{
	if ( !c )
	{
		errno = EINVAL;
		return;
	}

	/* cleanup result */
	c->result->code = 0;
	c->result->type = 0;

	if ( c->result->error )
		free ( c->result->error );
	if ( c->result->docid )
		free ( c->result->docid );
	if ( c->result->revid )
		free ( c->result->revid );

	c->result->error = NULL;
	c->result->docid = NULL;
	c->result->revid = NULL;

	/* cleanup XML result */
	if ( c->xml )
		free ( c->xml );

	c->xml = NULL;
}

void couch_result_cleanup ( CouchResult **r )
{
	if ( !r )
	{
		errno = EINVAL;
		return;
	}

	if ( !(*r) )
	{
		/* uninitialized couch-result */
		errno = 0;
		return;
	}

	if ( (*r)->error )
		free ( (*r)->error );

	if ( (*r)->docid )
		free ( (*r)->docid );

	if ( (*r)->revid )
		free ( (*r)->revid );

	free ( (*r) );
	(*r) = NULL;
}
