/* CouchDB document-handling functions
 *
 * CouchDB C API (libcouchdb) - Copyright (c) 2006 Dirk Schalge
 *
 * Thanks to Damien Katz and all other involved developers for this great
 * database!
 *
 * See doc/LICENSE for licensing details.
 */

#include "libcouchdb.h"
#include "couch.h"
#include "xml.h"
#include "curl.h"

static int couch_doc_cleanup_func ( void *data, void *data_user );
static int couch_doc_list_cleanup_func ( void *data, void *data_user );
static void couch_doc_field_cleanup ( CouchDocumentField *f );
static int couch_doc_field_cleanup_func ( void *data, void *data_user );
static void couch_doc_field_content_cleanup ( CouchDocumentFieldContent *c );
static CouchDocumentField *couch_doc_field_find ( List *l, char *name );
static int couch_doc_field_find_func ( void *data, void *data_user );


CouchDocument *couch_doc_create ( char *name )
{
	CouchDocument *new;

	new = calloc ( 1, sizeof(CouchDocument) );
	if ( !new )
		return NULL;

	if ( !list_init(&new->content) )
	{
		free ( new );
		return NULL;
	}

	if ( name )
	{
		new->docid = strdup ( name );
		if ( !new->docid )
		{
			couch_doc_cleanup ( &new );
			return NULL;
		}
	}

	return new;
}

CouchDocument *couch_doc_get ( Couch *c, char *docid )
{
	CouchDocument *d = NULL;

	if ( !c || !docid )
	{
		errno = EINVAL;
		return NULL;
	}

	/* send GET to server */
	couch_curl_doc_get ( c, docid );

	d = couch_xml_parse_doc_get ( c->xml );

	if ( !d )
	{
		/* parse error */
		couch_xml_parse_response ( c );
		return NULL;
	}

	return d;
}

List *couch_doc_get_all ( Couch *c )
{
	List *l = NULL;

	if ( !c )
	{
		errno = EINVAL;
		return NULL;
	}

	/* send GET to server */
	couch_curl_doc_get_all ( c );

	l = couch_xml_parse_doc_get_all ( c->xml );

	if ( l )
	{
		c->result->type = 1;
		return l;
	}
	else
	{
		/* parse error */
		couch_xml_parse_response ( c );
		return NULL;
	}
}

void couch_doc_set ( CouchDocument *d, char *name, ... )
{
	CouchDocumentField *field = NULL;
	va_list args;
	char *a = NULL;

	if ( !d || !d->content || !name )
	{
		errno = EINVAL;
		return;
	}

	/* does this field already exist? */
	field = couch_doc_field_find ( d->content, name );

	if ( !field )
	{
		/* field not found, create an empty new field */
		field = couch_doc_field_create ( d, name );
	}
	else
	{
		/* field found. clear the current content. */
		list_foreach ( field->values, couch_doc_field_cleanup_func, NULL );
		if ( !list_cleanup(&field->values) )
		{
			couch_doc_field_cleanup ( field );
			return;
		}

		if ( !list_init(&field->values) )
		{
			couch_doc_field_cleanup ( field );
			return;
		}
	}

	/* parse the values for the field */
	va_start ( args, name );
	while ( (a = va_arg(args, char*)) != NULL )
	{
		couch_doc_field_content_create ( field, 0, a );
	}
	va_end ( args );
}

int couch_doc_save ( Couch *c, CouchDocument *d )
{
	CouchResult *r = NULL;
	char *xml = NULL;
	int i = 0;

	if ( !c || !d )
	{
		errno = EINVAL;
		return -1;
	}

	/* gimme da xml for da fucking document */
	xml = couch_xml_doc_save ( d );

	/* send da fucking document to da fucking serva */
	couch_curl_doc_post ( c, xml );

	/* parse xml */
	i = couch_xml_parse_response ( c );

	free ( xml );

	if ( i < 1 )
		return i;

	/* update document & revision id if necessary */
	r = c->result;

	if ( r->type == 1 && !d->docid )
		if ( !(d->docid = strdup(r->docid)) )
			return -1;

	if ( d->revid )
		free ( d->revid );

	if ( !(d->revid = strdup(r->revid)) )
		return -1;

	/* done! :-) */
	return 1;
}

int couch_doc_remove ( Couch *c, char *docid )
{
	if ( !c || !docid )
	{
		errno = EINVAL;
		return -1;
	}

	/* send DELETE to server */
	couch_curl_doc_delete ( c, docid );

	/* parse xml */
	return couch_xml_parse_response ( c );
}

void couch_doc_cleanup ( CouchDocument **d )
{
	if ( !d )
	{
		errno = EINVAL;
		return;
	}

	if ( !(*d) )
	{
		/* uninitialized couch document */
		errno = 0;
		return;
	}

	if ( (*d)->docid )
		free ( (*d)->docid );

	if ( (*d)->revid )
		free ( (*d)->revid );

	if ( (*d)->content )
	{
		list_foreach ( (*d)->content, couch_doc_cleanup_func, NULL );
		list_cleanup ( &(*d)->content );
	}

	free ( (*d) );
	(*d) = NULL;
}

void couch_doc_list_cleanup ( List **l )
{
	if ( !l )
	{
		errno = EINVAL;
		return;
	}

	if ( !(*l) )
	{
		errno = 0;
		return;
	}

	/* cleanup list content */
	list_foreach ( (*l), couch_doc_list_cleanup_func, NULL );

	/* cleanup the list itself */
	list_cleanup ( l );

	(*l) = NULL;
}


/*	##########################
 *	##### END PUBLIC API #####
 *	##########################
 */

static int couch_doc_cleanup_func ( void *data, void *data_user )
{
	if ( !data || data_user )
		return -1;

	couch_doc_field_cleanup ( (CouchDocumentField *)data );

	return 1;
}

static int couch_doc_list_cleanup_func ( void *data, void *data_user )
{
	CouchDocument *d = NULL;

	if ( !data || data_user )
		return -1;

	d = (CouchDocument *)data;

	couch_doc_cleanup ( &d );

	return 1;
}

CouchDocumentField *couch_doc_field_create ( CouchDocument *d, char *name )
{
	CouchDocumentField *field = NULL;
	
	field = calloc ( 1, sizeof(CouchDocumentField) );
	if ( !field )
		return NULL;

	field->name = strdup ( name );
	if ( !field->name )
	{
		couch_doc_field_cleanup ( field );
		return NULL;
	}

	if ( !list_init(&field->values) )
	{
		couch_doc_field_cleanup ( field );
		return NULL;
	}

	/* add the field to the list! :-) */
	if ( !list_append(d->content, field) )
	{
		couch_doc_field_cleanup ( field );
		return NULL;
	}

	return field;
}

static void couch_doc_field_cleanup ( CouchDocumentField *f )
{
	if ( !f )
		return;

	if ( f->name )
		free ( f->name );

	if ( f->values )
	{
		list_foreach ( f->values, couch_doc_field_cleanup_func, NULL );
		list_cleanup ( &f->values );
	}

	free ( f );
}

static int couch_doc_field_cleanup_func ( void *data, void *data_user )
{
	if ( !data || data_user )
		return -1;

	couch_doc_field_content_cleanup ( (CouchDocumentFieldContent *)data );

	return 1;
}

CouchDocumentFieldContent *couch_doc_field_content_create ( CouchDocumentField *f, int type, char *value )
{
		CouchDocumentFieldContent *content = NULL;

		if ( !f || type < 0 || type > 1 || !value )
			return NULL;

		content = calloc ( 1, sizeof(CouchDocumentFieldContent) );
		if ( !content )
			return NULL;

		content->type = type;
		content->value = strdup ( value );
		if ( !content->value )
		{
			couch_doc_field_content_cleanup ( content );
			return NULL;
		}

		if ( !list_append(f->values, content) )
		{
			couch_doc_field_content_cleanup ( content );
			return NULL;
		}

		return content;
}

static void couch_doc_field_content_cleanup ( CouchDocumentFieldContent *c )
{
	if ( !c )
		return;

	if ( c->value )
		free ( c->value );

	free ( c );
}



static CouchDocumentField *couch_doc_field_find ( List *l, char *name )
{
	return (CouchDocumentField *)list_search ( l, couch_doc_field_find_func, name );
}

static int couch_doc_field_find_func ( void *data, void *data_user )
{
	CouchDocumentField *f = NULL;

	if ( !data || !data_user )
		return -1;

	f = (CouchDocumentField *)data;

	if ( !f->name )
		return -1;

	if ( !strcasecmp ( f->name, (char *)data_user ) )
		return 1;
	else
		return 0;
}
