<?php
/*
	(c) 2006 Jan Lehnardt <jan@php.net>
	
	All rights reserved.
	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:
	* Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.
	* Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.
	* Neither the name of Couch PHP Library nor the names of its contributors may 
	  be used to endorse or promote products derived from this software 
	  without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

// dev help
// error_reporting(E_ALL);
// ini_set('display_errors', 'On');


//include PEAR HTTP_Requerst
require_once "HTTP_Request/Request.php";

/**
 * CouchDb PHP Library
 *
 * This library allows you to talk to a CouchDb server from within PHP. It
 * saves you writing access functions to CouchDb's REST API. The primary
 * goal of this library is ease of use. To achieve this, trade-offs
 * had to be taken. They include speed optimiziation and cleaner code. A minor
 * but also important goal was to allow this library to be used from PHP 
 * versions 4 to 6. Libraries with less code can be written for PHP 5 and 6
 * and they are planned already, but this is what is here now and it can be
 * used by everyone. 
 *
 * A note on object oriented programming (OOP): This library makes use of 
 * PHP's OOP syntax but it does not neccessarily adhere to more advanced
 * OOP design principles. The purpose of the two clases here is
 * to encapsulate data and namespace functions. You can easily build a pure
 * OOP Library on top of this library or from scratch.
 * 
 * The CouchDb PHP Library makes use of the PEAR package HTTP_Request which
 * in turn uses the PEAR packge itself as well as Net_Socket and Net_URL.
 *
 * To learn more about CouchDb, visit http://couchdb.com/
 *
 * Usage Example:
 * <pre> 
 * &lt;?php
 * include &apos;CouchDb/Couch.php&apos;;
 * 
 * $couch = new Couch(&apos;localhost&apos;, &apos;test&apos;); 
 * if(!$couch-&gt;running()) {
 * 	die(&quot;The Couch is not running\n&quot;);
 * }
 * 
 * $doc = new CouchDocument;
 * $doc-&gt;set(&apos;Name&apos;, &apos;Micha&apos;);
 * $doc-&gt;set('Age', 24);
 * $doc-&gt;set('Cool', 'true');
 * $docid = $couch-&gt;save($doc);
 * echo &quot;Successfully inserted $docid\n&quot;;
 * ?&gt;
 * </pre>
 * 
 * Enjoy!
 * 
 * @author Jan Lehnardt <jan@php.net>
 * @version 0.6.0
 * @copyright Jan Lehnardt, 2006
 * @package Couch
 **/
class Couch
{
	/**
	 * The host where CouchDb runs on
	 *
	 * @var string
	 **/
	var $host = '';
	
	/**
	 * The IP port that CouchDb listenes on
	 *
	 * @var integer
	 **/
	var $port = 0;
	
	var $request_options = array();
	
	var $_res = array();
	var $_curDoc = false;
	var $_curRow = 0;
	var $_curCol = '';
	var $_curTag = '';
	var $_tagCount = 0;
	var $_lastCount = 0;
	var $documents = array();
	
	/**
	 * Prefix for the fields of the document that holds the computed tables
	 *
	 * @var string
	 **/
	var $table_prefix = '$table_';
	
	/**
	 * Document Id of the document that holds the computed tables
	 *
	 * @var string
	 **/
	var $table_doc_id = '$tables';
	
	var $php4 = false;
	var $silent = false;
	
	var $tables = false;
	
	/**
	 * Constructor
	 *
	 * @param string the host where CouchDb runs on
	 * @param string database name that is used with this instance of Couch
	 * @param string username (unused in 0.6.0)
	 * @param string password (unused in 0.6.0)
	 * @param integer the port where CouchDb listens on
	 * @param array additional request options. See http://pear.php.net/package/HTTP_Request/docs/latest/apidoc/HTTP_Request/HTTP_Request.html#methodHTTP_Request
	 * @return void
	 **/
	function Couch($host = 'localhost', $database = '', $user = '', $pass = '', $port = 8888, $request_options = array())
	{
		$this->host = $host;
		$this->port = $port;
		$this->user = $user;
		$this->pass = $pass;
		$this->_request_options = $request_options;
		$this->db = $database;
		$this->url_set();
		if(version_compare($thisVersion = phpversion(), $php5Version = '5.0.0', $lowerThan = 'lt')) {
			$this->php4 = true;
		}
	}

	function url_set()
	{
		$this->url = 'http://'.$this->host.':'.$this->port.'/'.$this->db;
	}

	function db_set($_db)
	{
		$this->db = $_db;
		$this->url_set();
	}

	/**
	 * Determines if CouchDb is running
	 *
	 * @return bool
	 **/
	function running()
	{
		$socket = @fsockopen($this->host, $this->port, $errno = -1, $errstr = '', $timeout = 10);
		if(!$socket) {
			$this->throw_error("Cannot connect to CouchDb on {$this->host}:{$this->port}", E_USER_NOTICE);
			return false;
		}
		return true;
	}
	
	/**
	 * Creates the CouchDb database
	 *
	 * @return bool
	 **/
	function db_create()
	{
		if(!$this->_sendRequest('', 'PUT')) {
			$this->throw_error("Cannot create database \"{$this->db}\" to CouchDb on {$this->host}:{$this->port}", E_USER_NOTICE);
			return false;
		}

		return true;
	}
	
	/**
	 * Deletes the CouchDB Database
	 *
	 * @return void
	 **/
	function db_delete()
	{
		if(!$this->_sendRequest('', 'DELETE')) {
			$this->throw_error("Cannot delete database \"{$this->db}\" to CouchDb on {$this->host}:{$this->port}", E_USER_NOTICE);
			return false;
		}

		return true;
	}
	
	/**
	 * Determines if the CouchDb database exists
	 *
	 * @return void
	 **/
	function db_exists()
	{
		$this->_sendRequest('', 'GET');
		// echo "Debug: ".$this->lastStatusCode."\n";
		return ($this->lastStatusCode == '200');
	}
	
	/**
	 * Adds a single or multiple computed tables to the CouchDb database
	 *
	 * @param mixed table name or table structure array
	 * @param array table structure (if $name is a string)
	 * @return void
	 **/
	function table_add($name, $structure = false)
	{
		if(!$this->exists($this->table_doc_id)) {
			$tableDoc = new CouchDocument;
			$tableDoc->set('id', $this->table_doc_id);
			$this->save($tableDoc, $this->table_doc_id);
		}
		
		$this->table_update($name, $structure);
	}
	
	/**
	 * Determines if a table already exists in the $tables CouchDocument
	 *
	 * @return void
	 * @param mixed table name or table structure array
	 * @param array table structure (if $name is a string)
	 **/
	function table_exists($name)
	{
		return (bool)$this->get($this->table_doc_id);
	}
	
	/**
	 * Updates a single or multiple table definitions
	 *
	 * @return void
	 **/
	function table_update($name, $structure = false)
	{
		$tableDoc = $this->get($this->table_doc_id);

		if(is_array($name)) {
			foreach($name AS $tableName => $tableStructure) {
				$tableDoc->set($this->table_prefix.$tableName, $tableStructure);
			}
		} else {
			$tableDoc->set($this->table_prefix.$name, $structure);
		}		

		$this->save($tableDoc, $this->table_doc_id, $tableDoc->get('revision'), $escape = false);
	}
	
	/**
	 * Deletes a table form CouchDb's $tables CouchDocument
	 *
	 * @return void
	 * @param string tablename
	 **/
	function table_delete($name)
	{
		$tableDoc = $this->get($this->table_doc_id);
		$tableDoc->set($this->table_prefix.$name, NULL);
		$this->save($tableDoc);
	}
	
	/**
	 * Determines if a CouchDocument with a given ID exists
	 *
	 * @return bool
	 * @param id string docuemnt id
	 **/
	function exists($id)
	{
		return (bool)$this->get($id);
	}
	
	/**
	 * Retrieves a single CouchDocument from the CouchDb database
	 *
	 * @return mixed CouchDocument or false on error
	 * @param string CouchDocument id
	 * @param array additional table request options. Not yet in use.
	 **/
	function get($id, $options = array())
	{
		if(empty($id)) {
			$this->throw_error("Couch::get(): Empty id given.", E_USER_NOTICE);
			return false;
		}
		$optionsUrl = '';
		foreach($options as $option => $value) {
			$optionsUrl .= $option.'='.$value.'&';
		}
		
		$url = "/$id?$optionsUrl";
		$resp = $this->_sendRequest($url, 'GET');
		if(!$resp) {
			return false;
		}
		$this->_parseResult($resp);
	
		return $this->_curDoc;
	}
	
	/**
	 * Retrieves all documents from a table
	 *
	 * @return array of CouchDocuments or false on error
	 * @param string tablename
	 * @param array additional table request options, see http://www.couchdbwiki.com/index.php?title=HTTP_Computed_Tables_API#Querying_Options (Example: array('key' => 'foo', 'reverse' => 'true'))
	 **/
	function get_all($table, $options = array())
	{
		if(empty($table)) {
			$this->throw_error("Couch::get_all(): Empty table given.", E_USER_NOTICE);
			return false;
		}
		
		$optionsUrl = '?';
		foreach($options as $option => $value) {
			$optionsUrl .= $option.'='.$value.'&';
		}
		
		$url = "/$this->table_doc_id:$table$optionsUrl";
		
		$resp = $this->_sendRequest($url, 'GET');
		if(!$resp) {
			return false;
		}
		$this->_parseResult($resp);
		return $this->documents;
	}
	
	function get_all_docs($db)
	{
		$options = $this->request_options;
		$options['method'] = 'GET';
		$options['user'] = $this->user;
		$options['pass'] = $this->pass;

		$req = new HTTP_Request("http://{$this->host}:{$this->port}/$db\$all_docs", $options);
		$resp = $req->sendRequest();
		if(!$resp) {
			return false;
		}
		$this->_parseResult($req->getResponseBody());
		return $this->documents;
	}
	
	function get_all_dbs()
	{
		$options = $this->request_options;
		$options['method'] = 'GET';
		$options['user'] = $this->user;
		$options['pass'] = $this->pass;

		$req = new HTTP_Request("http://{$this->host}:{$this->port}/\$all_dbs", $options);
		$resp = $req->sendRequest();
		if(!$resp) {
			return false;
		}
		
		$matches = array();
		preg_match_all('/<db>([^<]+)<\/db>/', $req->getResponseBody(), $matches);
		return $matches[1];
	}
	
	function update($id, $values)
	{
		if(empty($id)) {
			$this->throw_error("Couch::update(): Empty id given.", E_USER_NOTICE);
			return false;
		}
		// echo 'in';
		$doc = $this->get($id);
		// var_dump($doc);
		foreach($values AS $field => $value) {
			$doc->set($field, $value);
		}
		// var_dump($doc);
		$this->save($doc, $doc->get('id'), $doc->get('revision'));
		// echo 'out';
	}
	
	/**
	 * Saves a CouchDocument
	 *
	 * @return mixed saved CouchDocument id or false on error
	 * @param CouchDocument
	 * @param string CouchDocument id
	 * @param string CouchDocument revision
	 * @param bool xml escape data
	 **/
	function save($doc, $id = '', $rev = '', $escape = true)
	{
		if(!is_object($doc)) {
			$this->throw_error("Couch::save(): Invalid document given.", E_USER_NOTICE);
			return false;
		}
		
		$idxml = $revxml = '';
		if((strlen($id) > 0) || strlen(($id = $doc->get('id')) > 0)) {
			$idxml = ' id="'.$id.'"';
		}

		if((strlen($rev) > 0) || strlen(($rev = $doc->get('revision')) > 0)) {
			$revxml = ' rev="'.$rev.'"';
		}

		$xml = <<<EOX
<?xml version="1.0" ?>
<doc$idxml$revxml>
EOX;
		$fields = get_object_vars($doc);
		foreach($fields AS $field => $value) {
			if(($field == 'id') || ($field == 'revision')) {continue;}
			$xml .= "<field name=\"$field\">";

			if(is_array($value)) {
				foreach($value AS $listValue) {
					$listValue = $this->_xmlEscape($listValue, $escape);
					$xml .= "<text>$listValue</text>";
				}
			} else {
				$value = $this->_xmlEscape($value, $escape);
				$xml .= "<text>$value</text>";
			}
			$xml .= '</field>';
			
		}

		$xml .= '</doc>';
		if(!$this->_sendRequest('', 'POST', $xml)) {
			return false;
		}
		
		return $this->lastDocId;
	}

	/**
	 * Deletes a CouchDocument
	 *
	 * @return bool success
	 * @param string CouchDocument id
	 **/
	function delete($id)
	{
		if(empty($id)) {
			return false;
		}
		
		$url = '/'.$id;
		if(!$this->_sendRequest($url, 'DELETE')) {
			$this->throw_error("Cannot delete document $id from CouchDb on {$this->host}:{$this->port}", E_USER_NOTICE);
			return false;
		}
		
		return true;
	}
	
	
	function _sendRequest($url, $method, $post_data = false)
	{
		if(empty($method)) {
			return false;
		}
		
		$options = $this->request_options;
		$options['method'] = $method;
		$options['user'] = $this->user;
		$options['pass'] = $this->pass;
		
		$req = new HTTP_Request($this->url.$url, $options);
		if($post_data) {
			$req->addRawPostData($post_data);
			
		}
		$resp = $req->sendRequest();
		if(PEAR::isError($resp)) {
			return false;
		}
		
		$this->lastStatusCode = $req->getResponseCode();
		
		if(!in_array($this->lastStatusCode, range(200, 299))) {
			return false;
		}
		
		$this->lastRevision = $req->getResponseHeader('x-couch-revid');
		$this->lastDocId = $req->getResponseHeader('x-couch-docid');
		return $req->getResponseBody();
	}
	
	/**
	 * Returns the lastest error message
	 *
	 * @return string error message
	 **/	
	function get_latest_error()
	{
		$status = $this->lastStatusCode;
		
		if($status == 200) {
			return 'No Error (Code 200).';
		}
		
		switch($status) {
			case 409:
				return "Resource Exists (Code $status).";
			break;
			
			default:
				return "Unknown Error. Sorry about that! (Code $status).";
			break;
		}
		
		return false;
	}

	
	function throw_error($message, $type)
	{
		if(!$this->silent) {
			trigger_error($message, $type);
		}
	}
	
	function _parseResult($xml)
	{
		$parser = xml_parser_create();
		if($this->php4) {
			xml_set_element_handler($parser, array(&$this, '_parserStart'), array(&$this, '_parserEnd'));
			xml_set_character_data_handler($parser, array(&$this, '_parserCdata'));
		} else {
			xml_set_element_handler($parser, array($this, '_parserStart'), array($this, '_parserEnd'));
			xml_set_character_data_handler($parser, array($this, '_parserCdata'));	
		}
		xml_parser_set_option($parser, XML_OPTION_CASE_FOLDING, false);
		$this->_curRow = 0;
		$this->_curTag = '';
		$this->_curCol = '';
		$this->documents = array();
		xml_parse($parser, $xml);
	}
	
	function _parserStart($parser, $tag, $attributes)
	{
		static $elements = array('table', 'column', 'text', 'row', 'field', 'doc');
		if(!in_array($tag, $elements)) {
			return;
		}

		$this->_curTag = $tag;
		$this->_tagCount++;
		if(('row' == $tag) || ('doc' == $tag)) {
			$attributeName = 'id';
			if('row' == $tag) {
				$attributeName = 'docid';
			}
			$this->_curRow = $attributes[$attributeName];
			if($this->php4) {
				$this->_curDoc =& new CouchDocument;
				$this->documents[$this->_curRow] =& $this->_curDoc;
			} else {
				//php5 code
				$this->_curDoc = new CouchDocument;
				$this->documents[$this->_curRow] = $this->_curDoc;
			}
			$this->_curDoc->set('id', $this->_curRow);
			$this->_curDoc->set('revision', $attributes['rev']);
		}

		if($tag == 'table') {
			$this->totalRows = $attributes['total_rows'];
		}
		
		if($tag == 'field') {
			$this->_curCol = $attributes['name'];
		}
		
		if($tag == 'column') {
			$this->_curCol = $attributes['name'];
		}
	}
	
	function _parserEnd($parser, $tag)
	{
		//do nothing

	}
	
	function _parserCdata($parser, $data)
	{
		if(strlen($this->_curTag) == 0) {
			return;
		}

		if(strlen(trim($data)) == 0) {
			return;
		}
		/*
		if we have multiple cdata elements inside a tag, concatenate them into one field
		if we have multiple fields with the same name, array them
		*/
		$append = false;
		if($this->_lastCount == $this->_tagCount) {
			$append = true;
		}
		
		// if(isset($this->_res[$this->_curRow][$this->_curCol]) && !is_array($this->_res[$this->_curRow][$this->_curCol]) && !$append) {
		// 	$old = $this->_res[$this->_curRow][$this->_curCol];
		// 	$this->_res[$this->_curRow][$this->_curCol] = array();
		// 	$this->_res[$this->_curRow][$this->_curCol][] = $old;
		// 	$this->_res[$this->_curRow][$this->_curCol][] = $data;
		// }
		
		if($this->_curDoc->_isset($this->_curCol) && 
			!is_array($this->_curDoc->get($this->_curCol)) && 
			!$append) {
			$old = $this->_curDoc->_get($this->_curCol);
			$this->_curDoc->set($this->_curCol, array());
			$this->_curDoc->append($this->_curCol, $old);
			$this->_curDoc->append($this->_curCol, $data);
		}
		
		// if(is_array($this->_res[$this->_curRow][$this->_curCol])) {
		if(is_array($this->_curDoc->get($this->_curCol))) {
			if($append) {
				// $this->_res[$this->_curRow][$this->_curCol][count($this->_res[$this->_curRow][$this->_curCol])-1] .= $data;
				$this->_curDoc->append($this->_curCol, $data, $appendToPrevious = true);
				
			} else {
				$this->_curDoc->append($this->_curCol, $data);
			}
		} else {
			if($append) {
				// $this->_res[$this->_curRow][$this->_curCol] .= $data;
				$this->_curDoc->append($this->_curCol, $data);
			} else {
				// $this->_res[$this->_curRow][$this->_curCol] = $data;
				$this->_curDoc->set($this->_curCol, $data);
			}
		}
		
		$this->_lastCount = $this->_tagCount;
	}
	
	function _xmlEscape($xml, $escape = true)
	{
		if(!$escape) {
			return $xml;
		}
		
		$from[0] = '<';
		$to[0] = '&lt;';

		$from[1] = '>';
		$to[1] = '&gt;';

		$from[2] = '"';
		$to[2] = '&quot;';

		$from[3] = "'";
		$to[3] = '&apos;';
		
		$from[4] = '&';
		$to[4] = '&amp;';
		
		return str_replace($from, $to, $xml);
	}
	function _xmlUnEscape($xml)
	{
		$from[0] = '<';
		$to[0] = '&lt;';

		$from[1] = '>';
		$to[1] = '&gt;';

		$from[2] = '"';
		$to[2] = '&quot;';

		$from[3] = "'";
		$to[3] = '&apos;';
		
		$from[4] = '&';
		$to[4] = '&amp;';
		
		return str_replace($to, $from, $xml);
	}
	
	/**
	 * Set the CouchDb Library to display error messages. Only useful and needed during development. Don't use in production!
	 *
	 * @return void
	 * @see be_verbose(), squeal()
	 **/
	function shout()
	{
		$this->be_verbose();
	}

	/**
	 * Set the CouchDb Library to display error messages. Only useful and needed during development. Don't use in production!
	 *
	 * @return void
	 * @see be_verbose(), shout()
	 **/
	function squeal()
	{
		$this->be_verbose();
	}

	/**
	 * Set the CouchDb Library to display error messages. Only useful and needed during development. Don't use in production!
	 *
	 * @return void
	 * @see shout(), squeal()
	 **/
	function be_verbose()
	{
		$this->silent = false;
	}

	/**
	 * Set the CouchDb Library to be quiet about errors. Use in production.
	 *
	 * @return void
	 * @see be_quiet(), shut_up()
	 **/
	function shush()
	{
		$this->beQuiet();
	}

	/**
	 * Set the CouchDb Library to be quiet about errors. Use in production.
	 *
	 * @return void
	 * @see be_quiet(), shush()
	 **/
	function shut_up()
	{
		$this->be_quiet();
	}
	
	/**
	 * Set the CouchDb Library to be quiet about errors. Use in production.
	 *
	 * @return void
	 * @see shush(), shut_up()
	 **/
	function be_quiet()
	{
		$this->silent = true;
	}
}

/**
 * Encapsulates a the data in a CouchDb document
 *
 * @package default
 * @author Jan Lehnardt <jan@php.net>
 **/
class CouchDocument
{
	var $id;
	var $revision;
	
	function set($name, $value)
	{
		@$this->$name = $value;
	}
	
	function _isset($name)
	{
		return @isset($this->$name);
	}
	
	function get($name)
	{
		return @$this->$name;
	}
	
	function append($name, $value, $appendToPrevious = false)
	{
		if($appendToPrevious) {
			$this->$name[$count($this->$name) - 1] = $val;
		} else {
			if(is_array($this->$name)) {
				@$this->{$name}[] = $value;
			} else {
				@$this->$name .= $value;
			}
		}
	}
}