/*
Fabric formula engine server. For use with Erlang and CouchDb
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

// This file is the C port driver for Erlang. It provides a low overhead
// means of calling into C code, however unlike the Fabric engine, coding
// errors in this module can crash the entire Erlang server.

#include "erl_driver.h"
#include "unicode/ucol.h"
#include "unicode/ucasemap.h"
#ifndef WIN32
#include <strings.h> // for memcpy
#endif

#define CMD_COLLATE 0
#define CMD_COLLATE_INSENSITIVE 1


typedef struct {
    ErlDrvPort port;	
	UCollator* collCaseInsensitive;
	UCollator* collCaseSensitive;
} couch_drv_data;

static void free_drv_data(couch_drv_data* pData)
{
	if (pData->collCaseSensitive) {
		ucol_close(pData->collCaseSensitive);
	}
	if (pData->collCaseInsensitive) {
		ucol_close(pData->collCaseInsensitive);
	}
	driver_free((char*)pData);	
}

static ErlDrvData couch_drv_start(ErlDrvPort port, char *buff)
{
	UErrorCode status = U_ZERO_ERROR;
    couch_drv_data* pData = (couch_drv_data*)driver_alloc(sizeof(couch_drv_data));
	if (pData == NULL) {
		return ERL_DRV_ERROR_GENERAL;
	}

    pData->port = port;
	pData->collCaseSensitive = NULL;
	pData->collCaseInsensitive = NULL;

    return (ErlDrvData)pData;
}

static void couch_drv_stop(ErlDrvData handle)
{
    driver_free((char*)handle);
}

static int return_control_result(void* pLocalResult, int localLen, char **ppRetBuf, int returnLen)
{
	if (*ppRetBuf == NULL || localLen > returnLen) {
		*ppRetBuf = (char*)driver_alloc_binary(localLen);
		if(*ppRetBuf == NULL) {
			return -1;
		}
	}
	memcpy(*ppRetBuf, pLocalResult, localLen);
	return localLen;
}

// this populates the pItor with the data at *ppBuf and increments
// the values at ppBuf and pBufLen to "consume" the buffer.
static int get_string(const char ** ppBuf, int* pBufLen, UCharIterator* pItor)
{
	int strLen;

	if (*pBufLen < sizeof(int)) {
		return -1;
	}
	memcpy(&strLen, *ppBuf, sizeof(int));
	*ppBuf += sizeof(int);
	*pBufLen -= sizeof(int);
	
	if (*pBufLen < strLen) {
		return -1;
	}

	uiter_setUTF8(pItor, *ppBuf, strLen);

	*ppBuf += strLen;
	*pBufLen -= strLen;

	return 0;
}


static int couch_drv_control(ErlDrvData drv_data, unsigned int command, const char *pBuf, 
		   int bufLen, char **rbuf, int rlen)
{
    couch_drv_data* pData = (couch_drv_data*)drv_data;

    switch(command) {
    case CMD_COLLATE: //case collate
    case CMD_COLLATE_INSENSITIVE: //case collate, case insensitive
		{		
		UErrorCode status = U_ZERO_ERROR;
		int collResult;
		char response;
		UCharIterator iterA;
		UCharIterator iterB;

		if (get_string(&pBuf, &bufLen, &iterA) == -1) {
			return -1;
		}
		if (get_string(&pBuf, &bufLen, &iterB) == -1) {
			return -1;
		}

		if (command ==  CMD_COLLATE) {
			if (pData->collCaseSensitive == NULL) {
				pData->collCaseSensitive = ucol_open("", &status);
				if (U_FAILURE(status)) {
					return -1;
				}
			}
			collResult = ucol_strcollIter(pData->collCaseSensitive, &iterA, &iterB, &status);
		}
		else {
			if (pData->collCaseSensitive == NULL) {
				pData->collCaseInsensitive = ucol_open("", &status);
				ucol_setAttribute(pData->collCaseInsensitive, UCOL_STRENGTH, UCOL_PRIMARY, &status);				
				if (U_FAILURE(status)) {
					return -1;
				}
			}
			collResult = ucol_strcollIter(pData->collCaseInsensitive, &iterA, &iterB, &status);
		}
		if (collResult < 0) {
			response = 0; //lt
		}
		else if (collResult > 0) {
			response = 1; //gt
		}
		else {
			response = 2; //eq
		}
		return return_control_result(&response, sizeof(response), rbuf, rlen);
		}

	default:
		return -1;
	}
}

ErlDrvEntry couch_driver_entry = {
    NULL,                       /* F_PTR init, N/A */
    couch_drv_start,          /* L_PTR start, called when port is opened */
    couch_drv_stop,           /* F_PTR stop, called when port is closed */
    NULL,         /* F_PTR output, called when erlang has sent */
    NULL,                       /* F_PTR ready_input, called when input descriptor ready */
    NULL,                       /* F_PTR ready_output, called when output descriptor ready */
    "couch_erl_driver",              /* char *driver_name, the argument to open_port */
    NULL,                       /* F_PTR finish, called when unloaded */
	NULL,						/* Not used */
    couch_drv_control,         /* F_PTR control, port_command callback */
    NULL,                       /* F_PTR timeout, reserved */
    NULL                        /* F_PTR outputv, reserved */
};

DRIVER_INIT(couch_erl_driver) /* must match name in driver_entry */
{
    return &couch_driver_entry;
}


