%% CouchDb
%% Copyright (C) 2006  Damien Katz
%% 
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%% 
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%% 
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-module(couch_fabric).
-behaviour(gen_server).

-include("couch_db.hrl").

-export([start_link/1, test/0]).

-export([init/1, terminate/2, handle_call/3, handle_cast/2, handle_info/2,code_change/3,stop/0]).
-export([check/2, check_all/1, table_compute/2, table_columns/1]).
-export([start_tables/1, tables_compute_docs/2, stop_tables/1]).

-define(ERR_HANDLE, {Port, {exit_status, Status}} -> {stop, {unknown_error, Status}, {unknown_error, Status}, Port}).

start_link(FabricExec) ->
	gen_server:start_link({local, couch_fabric}, couch_fabric, FabricExec, []).
	
stop() ->
	exit(whereis(couch_fabric), close).
	
start_tables(Formulas) ->
	{ok, Cmd} = gen_server:call(couch_fabric, get_exe_cmd),
	
	Port = open_port({spawn, Cmd}, [{packet, 4}, exit_status, hide]),
	
	% indicate table computation start
	send_dword(Port, 2),
	
	% send the number of formulas
	send_dword(Port, length(Formulas)), 
	
	% send the formulas
	lists:foreach(fun(Formula) ->
			send_string(Port, Formula)
		end,
		Formulas),
			
	% get the results
	lists:foreach(fun(_Formula) ->
			case get_dword(Port) of
			0 ->
				ok;
			1 ->
				throw (get_error(Port))
			end
		end,
		Formulas),
		
	{ok, Port}.
	
tables_compute_docs(TablesPid, Docs) ->		
	% send the documents
	lists:foreach(fun(Doc) ->
			% indicate table computation
			send_dword(TablesPid, 3),
			
			send_doc(TablesPid, Doc)
		end,
		Docs),
	
	% receive the results
	Results =
	lists:map(fun(_Doc) ->
			% get the number of table formulas	
			FormulaCount = get_dword(TablesPid),
	
			get_table_results(TablesPid, FormulaCount, [])
		end,
		Docs),
	{ok, Results}.
	
% loop for each formula
get_table_results(_TablesPid, 0, Results) ->
	lists:reverse(Results);
get_table_results(TablesPid, N, Results) ->
	% did the formula match?
	Result = 
	case get_dword(TablesPid) of
	0 -> % no match
		no_match;
	1 -> % did match
		{ok, get_values(TablesPid)}
	end,
	get_table_results(TablesPid, N -1, [Result | Results]).
	
	
	
stop_tables(TablesPid) ->
	port_close(TablesPid).

check(Formula, Type) ->
	{ok, [Result]} = check_all([{Formula,Type}]),
	Result.
	
check_all(Formulas) ->
	gen_server:call(couch_fabric, {check, Formulas}).

timeout() ->
	% in Msecs. maybe need to make this a configurable value
	30000.
	
	
table_columns(Formulas) ->
	gen_server:call(couch_fabric, {table_columns, Formulas}).

init(FabricExec) ->
	Cmd =
	case erlang:system_info(os_type) of
	{unix, _} ->
	    lists:flatten(io_lib:format("~s launched_from_erlang", [FabricExec]));
	{win32, _} ->
	    lists:flatten(io_lib:format("cmd /c\"~s\" launched_from_erlang", [FabricExec]))
    end,
	Port = open_port({spawn, Cmd}, [{packet, 4}, exit_status, hide]),
	{ok, {Cmd, Port}}.

terminate(_Reason, _Server) ->
	ok.
	
	
handle_call(get_exe_cmd, _From, {Cmd, Port}) ->
	{reply, {ok, Cmd}, {Cmd, Port}};
handle_call({check, Formulas}, _From, {Cmd, Port}) ->	
	% send the formulas
	lists:foreach(fun({Formula, Type}) ->
			send_dword(Port, 1), % indicate compilation
			case Type of
			simple ->
				send_dword(Port, 0);
			table ->
				send_dword(Port, 1)
			end,
			send_string(Port, Formula)
		end,
		Formulas),
	
	Results = lists:map(fun(_Formula) ->
			case get_dword(Port) of
			0 ->
				ok;
			1 ->
				get_error(Port)
			end
		end,
		Formulas),
	{reply, {ok, Results}, {Cmd, Port}};
handle_call({table_columns, Formulas}, _From, {Exe, Port}) ->
	% indicate table columns
	send_dword(Port, 4),
	% send the number of formulas
	send_dword(Port, length(Formulas)), 		
	% send the formulas
	Results = lists:map(fun(Formula) ->
			send_string(Port, Formula),
			case get_dword(Port) of
			0 -> % compile error
				get_error(Port);
			1 ->
				Names = get_list(Port),
				{ok, [list_to_atom(Name) || Name <- Names]}
			end
		end,
		Formulas),
	{reply, {ok, Results}, {Exe, Port}}.
	
handle_cast(_Whatever, {Exe, Port}) ->
	{noreply, {Exe, Port}}.
	
handle_info({Port, {exit_status, Status}}, {Exe, Port}) ->
	{stop, {fabric_server_exited, Status}, {Exe, Port}};
handle_info(_Whatever, {Exe, Port}) ->
	{noreply, {Exe, Port}}.
	
code_change(_OldVsn, State, _Extra) ->
	{ok, State}.

send_dword(Port, Num) when Num =< 16#FFFFFFFF, Num >= 0 ->
	port_command(Port, <<Num:32/native >>).

get_dword(Port) ->
	receive
	{Port, {data, Bytes}} ->
		<<Result:32/native>> = list_to_binary(Bytes),
		Result;
	?ERR_HANDLE
	end.
	
get_float(Port) ->
	receive
	{Port, {data, Bytes}} ->
		<<Result:64/float>> = list_to_binary(Bytes),
		RoundInt = round(Result),
		case float(RoundInt) == Result of
			true -> RoundInt;
			false -> Result
		end,
		Result;
	?ERR_HANDLE
	end.
	
get_error(Port) ->
	ErrorAtom = get_atom(Port),
	ErrorString = get_string(Port),
	{ErrorAtom, ErrorString}.
	

send_float(Port, Num) ->
	port_command(Port, <<Num:64/float>>).

send_string(Port, []) ->
	% can't send null, send a single zeroed byte
	port_command(Port, [0]);
send_string(Port, String) ->
	port_command(Port, String).

get_string(Port) ->
	receive
	{Port, {data, [0]}} ->
		% we can't send or receive null strings, so to represent a
		% null string we send a zeroed 1 byte string and then
		% convert to a null string
		[];
	{Port, {data, String}} ->
		String;
	?ERR_HANDLE
	end.

get_atom(Port) ->
	list_to_atom(get_string(Port)).
	
send_doc(Port, Doc) ->
	Fields = [
		{"$docid", [Doc#doc.uuid]},
		{"$revisions", [couch_util:revid_to_list(Rev) || Rev <- Doc#doc.revisions]}
		| dict:to_list(Doc#doc.summary_fields)],
	send_fields(Port, Fields).
	
	
	
send_fields(Port, Fields) ->
	% send the number of fields
	send_dword(Port, length(Fields)),
	% send the fields
	lists:foreach(fun({FieldName, FieldValue}) ->
			send_string(Port, FieldName),
			send_list(Port, FieldValue)
		end,
		Fields).
	
get_values(Port) ->
	get_values2(Port, get_dword(Port), []).
	
get_values2(_Port, 0, Acc) ->
	lists:reverse(Acc);
get_values2(Port, Count, Acc) ->
	Value = get_list(Port),
	get_values2(Port, Count-1, [Value | Acc]).
	

send_list(Port, List) ->
	send_dword(Port, length(List)),
	send_list2(Port, List).

send_list2(_Port, []) ->
	ok;
send_list2(Port, [Element | Rest]) when is_list(Element) ->
	send_dword(Port, 0), % indicate string
	send_string(Port, Element),
	send_list2(Port, Rest);
send_list2(Port, [Element | Rest]) when is_number(Element) ->
	send_dword(Port, 1), % indicate float
	send_float(Port, Element),
	send_list2(Port, Rest).
	
get_list(Port) ->
	get_list2(Port, get_dword(Port), []).

get_list2(_Port, 0, Acc) ->
	lists:reverse(Acc);
get_list2(Port, Count, Acc) ->
	Element =
	case get_dword(Port) of
	0 ->
		get_string(Port);
	1 ->
		get_float(Port)
	end,
	get_list2(Port, Count - 1, [Element | Acc]).
	
test() ->
	couch_server:start(),
	ok = check("damien", simple),
	io:format("Formula compiled~n"),
	{ok, Columns} = table_columns([" select Foo == \"bar\"; COLUMN Food := Foo + Bar2", " select Foo != \"bar\";"]),
	io:format("Table Columns: ~p~n", [Columns]),
	Doc = couch_doc:new([{"Foo", ["bar"]}, {"Bar2", ["bar2"]}, {"Num", [1.0]}, {"gak", ["glopf"]}]),
	Doc2 = couch_doc:new([{"Foo", ["bar"]}, {"Bar2", ["bbbbbbbabs"]}, {"Num", [2.0]}, {"gak", ["glopf"]}]),
	{ok, [Result, ResultB]} = table_compute([Doc, Doc2], [" select Foo == \"bar\"; COLUMN Food := Num + 2", " select Foo != \"bar\";"]),
	 io:format("~p~n", [Result]),
	  io:format("ResultB: ~p~n", [ResultB]),
	{ok, [Result2]} = table_compute([Doc], [" select Foo != \"bar\";"]),
	io:format("~p~n", [Result2]),
	{ok, [Result3]} = table_compute([Doc], [" select Num == 1"]),
	io:format("~p~n", [Result3]),
	ok.
	
table_compute(Docs, FormulaStrings) ->
	{ok, TableFormulas} = start_tables(FormulaStrings),
	Result = tables_compute_docs(TableFormulas, Docs),
	stop_tables(TableFormulas),
	Result.
	