%% CouchDb
%% Copyright (C) 2006  Damien Katz
%% 
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%% 
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%% 
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-module(couch_server_sup).
-behaviour(supervisor).

-define(DEFAULT_INI, "couch.ini").

-export([start_link/1,stop/0]).

%% supervisor callbacks
-export([init/1]).

start_link(IniFilename) ->
	case whereis(couch_server_sup) of
	undefined ->
		start_server(IniFilename);
	_Else ->
		{error, already_started}
	end.
	
start_server(IniFilename) ->
	{ok, IniBin} =
	case IniFilename of
	"" ->
		% no ini file specified, check the command line args
		case init:get_argument(couch) of
		{ok, Args} ->
			case lists:keysearch("ini", 1, Args) of
			{value, {"ini", Filename}} ->
				file:read_file(Filename);
			_Else ->
				get_default_ini()
			end;
		_Else ->
			get_default_ini()
		end;
	IniFilename ->
		% if IniFilename is specified, it is used
		file:read_file(IniFilename)
	end,
	{ok, Ini} = couch_util:parse_ini(binary_to_list(IniBin)),
	ConsoleStartupMsg = list_to_atom(couch_util:ini_get_value(Ini, "Couch", "ConsoleStartupMsg", "CouchDb server starting")),
	ConsoleStartupCompleteMsg = list_to_atom(couch_util:ini_get_value(Ini, "Couch", "ConsoleStartupCompleteMsg", "CouchDb server started")),
	LogLevel = list_to_atom(couch_util:ini_get_value(Ini, "Couch", "LogLevel", "error")),
	DbRootDir = couch_util:abs_pathname(couch_util:ini_get_value(Ini, "Couch", "DbRootDir", ".")),
	HttpConfigFile = couch_util:abs_pathname(couch_util:ini_get_value(Ini, "Couch", "HttpConfigFile", "couch_httpd.conf")),	
	FabricServer = couch_util:abs_pathname(couch_util:ini_get_value(Ini, "Couch", "FabricServer", "FabricServer")),
	LogFile = couch_util:abs_pathname(couch_util:ini_get_value(Ini, "Couch", "LogFile", "couchdb.log")),
	UtilDriverDir = couch_util:abs_pathname(couch_util:ini_get_value(Ini, "Couch", "UtilDriverDir", ".")),
	
	io:format("couch ~s (LogLevel=~s)~n", [couch_server:get_version(), LogLevel]),
	
	io:format("~s~n", [ConsoleStartupMsg]),
	
	% start the couch logger
	ok = couch_log:start(LogFile, LogLevel),
	
	couch_log:debug("Startup Info:~n\tDbRootDir=~p~n\tHttpConfigFile=~p~n\tFabricServer=~p~n\tLogFile=~p~n\tUtilDriverDir=~p",
				[DbRootDir, HttpConfigFile, FabricServer, LogFile, UtilDriverDir]),

    StartResult = supervisor:start_link({local, couch_server_sup}, couch_server_sup, {DbRootDir, FabricServer, HttpConfigFile, UtilDriverDir}),
    
    case StartResult of
	{ok,_} ->
		% only output when startup was successful
		io:format("~s~n", [ConsoleStartupCompleteMsg]);
	_ -> ok
	end,
	StartResult.
    
stop() ->
	catch exit(whereis(couch_server_sup), normal),
	couch_log:stop().

get_default_ini() ->
	case file:read_file(?DEFAULT_INI) of
	{ok, Ini} ->
		{ok, Ini};
	{error, enoent} ->
		{ok, <<>>} % default ini missing, just return a blank
	end.

init({DbServerRoot, FabricServer, HttpConfigFile, UtilDriverDir}) ->	
    {ok, {{one_for_one, 10, 3600}, 
		[{couch_server,
			{couch_server, sup_start_link, [DbServerRoot]},
			permanent,
			brutal_kill,
			worker,
			[couch_server]},
		{couch_util,
			{couch_util, start_link, [UtilDriverDir]},
			permanent,
			brutal_kill,
			worker,
			[couch_util]},
		{couch_fabric,
			{couch_fabric, start_link, [FabricServer]},
			permanent,
			brutal_kill,
			worker,
			[couch_fabric]},
		{httpd,
			{httpd, start_link, [HttpConfigFile]},
			permanent,
			1000,
			supervisor,
			[httpd]}
		]}}.