%% CouchDb
%% Copyright (C) 2006  Damien Katz
%% 
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%% 
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%% 
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-module(couch_stream).
-behaviour(gen_server).

-export([test/1]).
-export([open/1, open/2, read/3, read_term/2, write/2, write_term/2, get_state/1]).
-export([init/1, terminate/2, handle_call/3]).
-export([handle_cast/2,code_change/3,handle_info/2]).

-include("couch_db.hrl").

-define(FILE_POINTER_BYTES, 8).
-define(FILE_POINTER_BITS, 8*(?FILE_POINTER_BYTES)).

-define(STREAM_OFFSET_BYTES, 4).
-define(STREAM_OFFSET_BITS, 8*(?STREAM_OFFSET_BYTES)).


-record(write_stream,
	{fd = 0,
	current_pos = 0,
	bytes_remaining = 0,
	alloc_sizer_func = fun(NeededSize) -> NeededSize end % default is to return Needed size unchanged
	}).

-record(stream,
	{
	pid,
	fd
	}).


%%% Interface functions %%%

open(Fd) ->
	open(nil, Fd).

open(nil, Fd) ->
	open({0,0}, Fd);
open(State, Fd) ->
	{ok, Pid} = gen_server:start_link(couch_stream, {State, Fd}, []),
	{ok, #stream{pid = Pid, fd = Fd}}.

close(#stream{pid = Pid, fd = _Fd}) ->
	gen_server:call(Pid, close).

get_state(#stream{pid = Pid, fd = _Fd}) ->
	gen_server:call(Pid, get_state).

read(#stream{pid = _Pid, fd = Fd}, Sp, Num) ->
	read(Fd, Sp, Num);	
read(Fd, Sp, Num) ->
	{ok, RevBinList, Sp2} = read_data(Fd, Sp, Num, []),
	Bin = list_to_binary(lists:reverse(RevBinList)),
	{ok, Bin, Sp2}.

read_term(#stream{pid = _Pid, fd = Fd}, Sp) ->
	read_term(Fd, Sp);
read_term(Fd, Sp) ->
	{ok, <<TermLen:(?STREAM_OFFSET_BITS)>>, Sp2}
		= read(Fd, Sp, ?STREAM_OFFSET_BYTES),
	{ok, Bin, _Sp3} = read(Fd, Sp2, TermLen),
	{ok, binary_to_term(Bin)}.	

write_term(Stream, Term) ->
	Bin = term_to_binary(Term),
	Size = size(Bin),
	Bin2 = <<Size:(?STREAM_OFFSET_BITS), Bin/binary>>,
	write(Stream, Bin2).
	
write(#stream{pid = Pid}, Bin) ->
	gen_server:call(Pid, {write, Bin}).	
	

init({{Pos, BytesRemaining}, Fd}) ->
	{ok, #write_stream
		{fd = Fd,
		current_pos = Pos,
		bytes_remaining = BytesRemaining
		}}.
	
terminate(_Reason, _Stream) ->
	ok.
	
handle_call(get_state, _From, Stream) ->
	#write_stream{current_pos=Pos, bytes_remaining = BytesRemaining} = Stream,
	{reply, {Pos, BytesRemaining}, Stream};
handle_call({write, Bin}, _From, Stream) ->
	AllocSizer = Stream#write_stream.alloc_sizer_func,
	% ensure init is called first so we can get a pointer to the begining of the binary
	{ok, Sp, Stream2} = ensure_init(Stream, AllocSizer(size(Bin))),
	{ok, Stream3} = write_data(Stream2, Bin),
	{reply, {ok, Sp}, Stream3};
handle_call(close, _From, Stream) ->
	#write_stream{current_pos=Pos, bytes_remaining = BytesRemaining} = Stream,
	{stop, normal, {ok, {Pos, BytesRemaining}}, Stream}.
	
handle_cast(_Msg, State) ->
	{noreply,State}.

code_change(_OldVsn, State, _Extra) ->
	{ok, State}.
	
handle_info(_Info, State) ->
	{noreply, State}.

%%% Internal function %%%

read_data(_Fd, Sp, 0, RevBinList) ->
	{ok, RevBinList, Sp};
read_data(Fd, {Pos, 0}, Num, RevBinList) ->
	{ok, <<NextPos:(?FILE_POINTER_BITS), NextOffset:(?STREAM_OFFSET_BITS)>>}
		= couch_file:pread(Fd, Pos, ?FILE_POINTER_BYTES + ?STREAM_OFFSET_BYTES),
	Sp = {NextPos, NextOffset},
	read_data(Fd, Sp, Num, RevBinList);
read_data(Fd, {Pos, Offset}, Num, RevBinList) ->
	ReadAmount = lists:min([Num, Offset]),
	{ok, Bin} = couch_file:pread(Fd, Pos, ReadAmount),
	Sp = {Pos + ReadAmount, Offset - ReadAmount},
	read_data(Fd, Sp, Num - ReadAmount, [Bin | RevBinList]).

write_data(Stream, <<>>) ->
	{ok, Stream};
write_data(#write_stream{fd=Fd, current_pos=Pos, bytes_remaining=0, alloc_sizer_func=AllocSizer} = Stream, Bin) ->
	% no space in the current segment, must alloc a new segment
	NewSize = AllocSizer(size(Bin)),
	{ok, NewPos} = couch_file:expand(Fd, NewSize + ?FILE_POINTER_BYTES + ?STREAM_OFFSET_BYTES),
	ok = couch_file:pwrite(Fd, Pos, <<NewPos:(?FILE_POINTER_BITS), NewSize:(?STREAM_OFFSET_BITS)>>),
	Stream2 = Stream#write_stream{
		current_pos=NewPos,
		bytes_remaining=NewSize},
	write_data(Stream2, Bin);
write_data(#write_stream{fd=Fd, current_pos=Pos, bytes_remaining=BytesRemaining} = Stream, Bin) ->
	BytesToWrite = lists:min([size(Bin), BytesRemaining]),
	{WriteBin, Rest} = split_binary(Bin, BytesToWrite),
	ok = couch_file:pwrite(Fd, Pos, WriteBin),
	Stream2 = Stream#write_stream{
		bytes_remaining=BytesRemaining - BytesToWrite,
		current_pos=Pos + BytesToWrite
		},
	write_data(Stream2, Rest).
	
ensure_init(#write_stream{current_pos=0, bytes_remaining=0}=Stream, InitialSize) ->
	% null stream initialize
	{ok, NewPos} = couch_file:expand(Stream#write_stream.fd, InitialSize + ?FILE_POINTER_BYTES + ?STREAM_OFFSET_BYTES),
	{ok, {NewPos, InitialSize}, Stream#write_stream{current_pos = NewPos, bytes_remaining = InitialSize}};
ensure_init(#write_stream{current_pos=Pos, bytes_remaining=0, alloc_sizer_func=AllocSizer, fd=Fd}=Stream, InitialSize) ->
	% no space in the current segment, must alloc a new segment
	NewSize = AllocSizer(InitialSize),
	{ok, NewPos} = couch_file:expand(Fd, InitialSize + ?FILE_POINTER_BYTES + ?STREAM_OFFSET_BYTES),
	ok = couch_file:pwrite(Fd, Pos, <<NewPos:(?FILE_POINTER_BITS), NewSize:(?STREAM_OFFSET_BITS)>>),
	{ok, {NewPos, NewSize}, Stream#write_stream{current_pos = NewPos, bytes_remaining = NewSize}};
ensure_init(#write_stream{current_pos=Pos, bytes_remaining=NextOffset}=Stream, _InitialSize) ->
	{ok, {Pos, NextOffset}, Stream}.
	


%%% Tests %%%

	
test(Term) ->
	{ok, Fd} = couch_file:open("foo", [write]),
	{ok, Stream} = open({0,0}, Fd),
	{ok, Pos} = write_term(Stream, Term),
	{ok, Pos2} = write_term(Stream, {Term, Term}),
	close(Stream),
	couch_file:close(Fd),
	{ok, Fd2} = couch_file:open("foo", [read, write]),
	{ok, Stream2} = open({0,0}, Fd2),
	{ok, Term1} = read_term(Fd2, Pos),
	io:format("Term1: ~w ~n",[Term1]),
	{ok, Term2} = read_term(Fd2, Pos2),
	io:format("Term2: ~w ~n",[Term2]),
	{ok, PointerList} = deep_write_test(Stream2, Term, 1000, []),
	deep_read_test(Fd2, PointerList),
	close(Stream2),
	couch_file:close(Fd2).
	
deep_read_test(_Fd, []) ->
	ok;
deep_read_test(Fd, [Pointer | RestPointerList]) ->
	{ok, _Term} = read_term(Fd, Pointer),
	deep_read_test(Fd, RestPointerList).
	
deep_write_test(_Stream, _Term, 0, PointerList) ->
	{ok, PointerList};	
deep_write_test(Stream, Term, N, PointerList) ->
	WriteList = lists:duplicate(random:uniform(N), Term),
	{ok, Pointer} = write_term(Stream, WriteList),
	deep_write_test(Stream, Term, N-1, [Pointer | PointerList]).