%% CouchDb
%% Copyright (C) 2006  Damien Katz
%% 
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%% 
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%% 
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-module(couch_table_group).
-behaviour(gen_server).

-export([start_manager/4,stop/1,
		open/4,open/5,fold/5,fold/6,
		update_group/3,update_table_proc/5,free_groups/2,
		update_table_proc/7,get_group_async/3,
		less_fun/2]).
-export([init/1,terminate/2,handle_call/3,handle_cast/2,handle_info/2,code_change/3]).

-include("couch_db.hrl").

% arbitrarily chosen amount of memory to use before flushing to disk
-define(FLUSH_MAX_MEM, 1000000).

-record(table_group,
	{db,
	named_queries,
	tables,
	current_seq,
	uuid_btree,
	update_notif_fun,
	fabric_tables = nil
	}).
	
-record(table,
	{id,
	name,
	btree,
	columns,
	query_string
	}).
	
-record(mgr,
	{db,
	fd,
	update_state_by_uuid = dict:new(),
	uuids_by_pid = dict:new(),
	get_group_info_fun,
	update_group_info_fun,
	cached_groups=dict:new()
	}).
	

start_manager(Db, Fd, GetGroupInfoFun, UpdateGroupInfoFun) ->
	gen_server:start_link(couch_table_group, {Db, Fd, GetGroupInfoFun, UpdateGroupInfoFun}, []).

stop(MgrPid) ->
	gen_server:cast(MgrPid, stop).
	
init({Db, Fd, GetGroupInfoFun, UpdateGroupInfoFun}) ->
	process_flag(trap_exit, true),
	{ok, #mgr{db=Db, fd=Fd, get_group_info_fun=GetGroupInfoFun, update_group_info_fun=UpdateGroupInfoFun}}.

terminate(_Reason, _Mgr) ->
	ok.

get_group_async(MgrPid, GroupId, OrigFrom) ->
	gen_server:cast(MgrPid, {get_group_async, GroupId, OrigFrom}).


update_group(MgrPid, GroupId, UpdateNotifFun) ->
	gen_server:cast(MgrPid, {update_group, GroupId, UpdateNotifFun}).

% stops any processing of the table and frees and cached values
free_groups(MgrPid, GroupIds) ->
	gen_server:call(MgrPid, {free_groups, GroupIds}).


% called from the update process
handle_call({group_cache_update, GroupId, Group}, {FromPid, _FromRef}, Mgr) ->
	Group2 = Group#table_group{fabric_tables=nil},
	% the process may have been killed by the free groups call
	% so check to make sure its alive.
	case is_process_alive(FromPid) of
	true ->
		#mgr{cached_groups=CachedGroups} = Mgr,
		CachedGroups2 = dict:store(GroupId, Group2, CachedGroups),
		{reply, ok, Mgr#mgr{cached_groups=CachedGroups2}};
	false ->
		{reply, ok, Mgr}
	end;	
handle_call({free_groups, GroupIds}, _From, Mgr) ->
	Mgr2 =
	lists:foldl(fun(GroupId, MgrAcc) ->
			#mgr{cached_groups=CachedGroups, update_state_by_uuid=ProcsDict} = MgrAcc,
			CachedGroups2 = dict:erase(GroupId, CachedGroups),
			case dict:find(GroupId, ProcsDict) of
			{ok, {processing_request, _NotifyFuns, Pid}} ->
				exit(Pid, freed);
			{ok, {processing_and_pending_request, _NotifyFuns, _PendingUpdateNotifFuns, Pid}} ->
				exit(Pid, freed);
			_Else ->
				ok
			end,
			MgrAcc#mgr{cached_groups=CachedGroups2}
		end,
		Mgr, GroupIds),
	{reply, ok, Mgr2}.
	
handle_cast({get_group_async, GroupId, OrigFrom}, Mgr) ->
	#mgr{
		db=Db,
		fd=Fd,
		get_group_info_fun=GetGroupInfoFun,
		cached_groups=CachedGroups
		} = Mgr,
	case dict:find(GroupId, CachedGroups) of
	{ok, CachedGroup} ->
		gen_server:reply(OrigFrom, {ok, CachedGroup}),
		{noreply, Mgr};
	error ->
		{Mgr2, Reply} = 
			case GetGroupInfoFun(GroupId) of
			{ok, {NamedQueries, GroupState}} ->
				{ok, Group} = open(Db, Fd, NamedQueries, GroupState),
				NewMgr = Mgr#mgr{cached_groups=dict:store(GroupId, Group, CachedGroups)},
				{NewMgr, {ok, Group}};
			Else ->
				{Mgr, Else}
			end,
		gen_server:reply(OrigFrom, Reply),
		{noreply, Mgr2}
	end;
handle_cast({update_group, GroupId, UpdateNotifFun}, Mgr) ->
	#mgr{
		update_state_by_uuid=ProcsDict,
		uuids_by_pid=GroupIdDict,
		db=Db,
		fd=Fd,
		get_group_info_fun=GetGroupInfoFun,
		update_group_info_fun=UpdateGroupInfoFun,
		cached_groups=CachedGroups
		} = Mgr,
	case dict:find(GroupId, ProcsDict) of
	{ok, {processing_request, NotifyFuns, Pid}} ->
		ProcsDict2 = dict:store(
			GroupId,
			{processing_and_pending_request, NotifyFuns, [UpdateNotifFun], Pid},
			ProcsDict
			),
		{noreply, Mgr#mgr{update_state_by_uuid=ProcsDict2}};
	{ok, {processing_and_pending_request, NotifyFuns, PendingUpdateNotifFuns, Pid}} ->
		ProcsDict2 = dict:store(
			GroupId,
			{processing_and_pending_request, NotifyFuns, [UpdateNotifFun | PendingUpdateNotifFuns], Pid},
			ProcsDict
			),
		{noreply, Mgr#mgr{update_state_by_uuid=ProcsDict2}};
	error ->
		case dict:find(GroupId, CachedGroups) of
		{ok, Group} ->
			Pid = spawn_link(couch_table_group, update_table_proc, [self(), Group, GroupId, UpdateGroupInfoFun, [UpdateNotifFun]]);
		error ->
			Pid = spawn_link(couch_table_group, update_table_proc, [self(), Db, Fd, GroupId, GetGroupInfoFun, UpdateGroupInfoFun, [UpdateNotifFun]])
		end,
		ProcsDict2 = dict:store(GroupId, {processing_request, [UpdateNotifFun], Pid}, ProcsDict),
		GroupIdDict2 = dict:store(Pid, GroupId, GroupIdDict),
		{noreply, Mgr#mgr{update_state_by_uuid=ProcsDict2, uuids_by_pid=GroupIdDict2}}
	end;
handle_cast(stop, Mgr) ->
	{stop, normal, Mgr}. % causes terminate to be called
	
handle_info({'EXIT', FromPid, Reason}, Mgr) ->
	#mgr{
		update_state_by_uuid=ProcsDict,
		uuids_by_pid=GroupIdDict,
		db=Db,
		fd=Fd,
		get_group_info_fun=GetGroupInfoFun,
		update_group_info_fun=UpdateGroupInfoFun,
		cached_groups=CachedGroups
		} = Mgr,
	case dict:find(FromPid, GroupIdDict) of
	{ok, GroupId} ->
		case dict:find(GroupId, ProcsDict) of
		{ok, {processing_request, NotifyFuns, _Pid}} ->
			GroupIdDict2 = dict:erase(FromPid, GroupIdDict),
			ProcsDict2 = dict:erase(GroupId, ProcsDict);
		{ok, {processing_and_pending_request, NotifyFuns, NextNotifyFuns, _Pid}} ->
			case dict:find(GroupId, CachedGroups) of
			{ok, Group} ->
				Pid = spawn_link(couch_table_group, update_table_proc, [self(), Group, GroupId, UpdateGroupInfoFun, NextNotifyFuns]);
			error ->
				Pid = spawn_link(couch_table_group, update_table_proc, [self(), Db, Fd, GroupId, GetGroupInfoFun, UpdateGroupInfoFun, NextNotifyFuns])
			end,
			GroupIdDict2 = dict:store(Pid, GroupId, dict:erase(FromPid, GroupIdDict)),
			ProcsDict2 = dict:store(GroupId, {processing_request, NextNotifyFuns, Pid}, ProcsDict)
		end,
		case Reason of
		normal ->
			ok;
		_Else ->
			% process returned abnormally, notify any waiting listeners
			[catch NotifyFun(Reason) || NotifyFun <- NotifyFuns]
		end,
		Mgr2 = Mgr#mgr{update_state_by_uuid=ProcsDict2, uuids_by_pid=GroupIdDict2},
		{noreply, Mgr2};
	error ->
		% a linked process must have died, we propagate the error
		exit(Reason)
	end.
	
code_change(_OldVsn, State, _Extra) ->
	{ok, State}.
	
	
open(Db, Fd, NamedQueries, GroupState) ->
	open(Db, Fd, NamedQueries, GroupState, fun(_Table) -> ok end).
	
open(Db, Fd, NamedQueries, GroupState, UpdateNotifFun) ->
	{TableBtreeStates, CurrentSeq, GroupIdBtreeState} = 
		case GroupState of
		nil -> % new table group, init GroupState to nils
			{[ nil || _Query <- NamedQueries], 0, nil};
		GroupState ->
			GroupState
		end,
	{Tables, _Count} =
		lists:mapfoldl(fun({{Name, QueryString}, BtreeState}, Count) ->
				{ok, Btree} = couch_btree:open(BtreeState, Fd, fun less_fun/2),
				{#table{name=Name, id=Count, btree=Btree, query_string=QueryString}, Count+1}
			end,
		0, lists:zip(NamedQueries, TableBtreeStates)),
	{ok, GroupIdBtree} = couch_btree:open(GroupIdBtreeState, Fd),
	TableGroup = #table_group{db=Db,
		tables=Tables,
		current_seq=CurrentSeq,
		uuid_btree=GroupIdBtree,
		update_notif_fun=UpdateNotifFun,
		named_queries=NamedQueries},
	{ok, TableGroup}.
	
	
get_info(#table_group{named_queries=NamedQueries, tables=Tables, current_seq=CurrentSeq, uuid_btree=GroupIdBtree} = _TableGroup) ->
	TableBtreeStates = [couch_btree:get_state(Table#table.btree) || Table <- Tables],
	{NamedQueries, {TableBtreeStates, CurrentSeq, couch_btree:get_state(GroupIdBtree)}}.

fold(TableGroup, TableName, Dir, Fun, Acc) ->
	fold_int(TableGroup#table_group.tables, Dir, TableName, Fun, Acc).

fold_int([], _TableName, _Dir, _Fun, _Acc) ->
	{not_found, missing_named_table};
fold_int([Table | _RestTables], TableName, Dir, Fun, Acc) when Table#table.name == TableName ->
	TotalRowCount = couch_btree:row_count(Table#table.btree),
	{ok, [{ok, ColumnNames}]} = couch_fabric:table_columns([Table#table.query_string]),
	WrapperFun = fun({{KeyColumn, DocId}, [Rev|Columns]}, Offset, WrapperAcc) ->
		case ColumnNames of
		[] -> 
			Fun({DocId,Rev}, KeyColumn, [], Offset, TotalRowCount, WrapperAcc);
		_ ->
			Fun({DocId,Rev}, KeyColumn,  lists:zip(ColumnNames, [KeyColumn | Columns]), Offset, TotalRowCount, WrapperAcc)
		end
	end,
	{ok, AccResult} = couch_btree:fold(Table#table.btree, Dir, WrapperFun, Acc),
	{ok, TotalRowCount, AccResult};
fold_int([_Table | RestTables], TableName, Dir, Fun, Acc) ->
	fold_int(RestTables, TableName, Dir, Fun, Acc).
	
	
fold(TableGroup, TableName, StartKey, Dir, Fun, Acc) ->
	fold_int(TableGroup#table_group.tables, TableName, StartKey, Dir, Fun, Acc).

fold_int([], _TableName, _StartKey, _Dir, _Fun, _Acc) ->
	{not_found, missing_named_table};
fold_int([Table | _RestTables], TableName, StartKey, Dir, Fun, Acc) when Table#table.name == TableName ->
	TotalRowCount = couch_btree:row_count(Table#table.btree),
	{ok, [{ok, ColumnNames}]} = couch_fabric:table_columns([Table#table.query_string]),
	WrapperFun = fun({{KeyColumn, DocId}, [Rev|Columns]}, Offset, WrapperAcc) ->
		case ColumnNames of
		[] -> 
			Fun({DocId, Rev}, KeyColumn, [], Offset, TotalRowCount, WrapperAcc);
		_ ->
			Fun({DocId, Rev}, KeyColumn, lists:zip(ColumnNames, [KeyColumn | Columns]), Offset, TotalRowCount, WrapperAcc)
		end
	end,
	{ok, AccResult} = couch_btree:fold(Table#table.btree, StartKey, Dir, WrapperFun, Acc),
	{ok, TotalRowCount, AccResult};
fold_int([_Table | RestTables], TableName, StartKey, Dir, Fun, Acc) ->
	fold_int(RestTables, TableName, StartKey, Dir, Fun, Acc).
	



cmp_element(A, B) when is_list(A) and is_list(B) ->
	couch_util:collate(A, B, [nocase]);
cmp_element(A, B) when A == B ->
	0;
cmp_element(A, B) when A < B ->
	-1;
cmp_element(A, B) when A > B ->
	1;
cmp_element(_A, _B) ->
	0.
	
cmp_list([], []) ->
	0;
cmp_list([A|RestA], [B|RestB]) ->
	case cmp_element(A, B) of
	0 ->	cmp_list(RestA, RestB);
	Else ->	Else
	end;
cmp_list(_A, []) ->
	1;
cmp_list([], _B) ->
	-1.
	
% a case insenstive comparison function for keys in a table
% Note! Because of case insensitivity, this will not produce 
% a stable sort! However, there is a way to get a stable sort
% and the desired insensitivity sematics by using the right
% collations at the right time. TODO
less_fun({A, DocIdA}, {B, DocIdB}) when is_list(A) and is_list(B) ->
	case cmp_list(A, B) of
	0 ->
		DocIdA < DocIdB;
	-1 -> 
		true;
	1 ->
		false
	end;
less_fun({A, DocIdA}, {B, DocIdB}) ->
	{A, DocIdA} < {B, DocIdB}.
	

notify(MgrPid, UpdateStatus, TableGroup, GroupId, UpdateGroupInfoFun, StatusNotifyFuns) ->
	GroupInfo = get_info(TableGroup),
	ok = gen_server:call(MgrPid, {group_cache_update, GroupId, TableGroup}),
	ok = UpdateGroupInfoFun(GroupId, UpdateStatus, GroupInfo),
	
	StatusNotifyFuns2 = lists:foldl(fun(NotifyFun, AccFuns) ->
			case (catch NotifyFun(UpdateStatus)) of
			NewNotifyFun when is_function(NewNotifyFun) ->
				[NewNotifyFun | AccFuns];
			_Else ->
				AccFuns
			end
		end,
		[], StatusNotifyFuns),
	fun(UpdateStatus2, TableGroup2) ->
		notify(MgrPid, UpdateStatus2, TableGroup2, GroupId, UpdateGroupInfoFun, StatusNotifyFuns2)
	end.

update_table_proc(MgrPid, #table_group{} = TableGroup, GroupId, UpdateGroupInfoFun, StatusNotifyFuns) ->
	NotifyFun =
		fun(UpdateStatus, TableGroup2) ->
			notify(MgrPid, UpdateStatus, TableGroup2, GroupId, UpdateGroupInfoFun, StatusNotifyFuns)
		end,
	update_int(TableGroup, NotifyFun).
	
update_table_proc(MgrPid, Db, Fd, GroupId, GetGroupInfoFun, UpdateGroupInfoFun, StatusNotifyFuns) ->
	case GetGroupInfoFun(GroupId) of
	{ok, {NamedQueries, GroupState}} ->
		{ok, TableGroup} = open(Db, Fd, NamedQueries, GroupState),
		update_table_proc(MgrPid, TableGroup, GroupId, UpdateGroupInfoFun, StatusNotifyFuns);
	Error ->
		exit(Error)
	end.
	
update_int(TableGroup, NotifyFun) ->
	#table_group{
		db=Db,
		tables=Tables,
		current_seq=CurrentSeq
		} = TableGroup,		
	TableEmptyKVs = [{Table, []} || Table <- Tables],
	
	% compute on all docs modified since we last computed.		
	{ok, {UncomputedDocs, TableGroup2, TableKVsToAdd, GroupIdTableIdKeys, NewSeq, NotifyFun2}}
		= couch_db:enum_docs_since(
			Db,
			CurrentSeq,
			fun(DocInfo, Acc) -> process_doc(Db, DocInfo, Acc) end,
			{[], TableGroup, TableEmptyKVs, [], CurrentSeq, NotifyFun}
			),
			
	
	{TableGroup3, Results} = table_compute(TableGroup2, UncomputedDocs),
	{TableKVsToAdd2, GroupIdTableIdKeys2} = table_insert_query_results(UncomputedDocs, Results, TableKVsToAdd, GroupIdTableIdKeys),
	
	{ok, TableGroup4} = write_changes(TableGroup3, TableKVsToAdd2, GroupIdTableIdKeys2, NewSeq),
	NotifyFun2(complete, TableGroup4),
	{ok, TableGroup4}.
	
		
process_doc(Db, DocInfo, {Docs, TableGroup, TableKVs, GroupIdTableIdKeys, _LastSeq, NotifyFun}) ->
	% This fun computes once for each document			
	#doc_info{uuid=GroupId,
		update_seq=Seq,
		conflict_revs=ConflictRevs} = DocInfo,
	{Docs2, GroupIdTableIdKeys2} =
		case couch_db:open_doc(Db, DocInfo) of
		{ok, Doc} ->
			ConflictRevTextList = [couch_util:revid_to_list(Rev) || Rev <- ConflictRevs],
			Doc2 =
			case ConflictRevTextList of
				[] -> Doc; % no conflicts found
				_  -> couch_doc:set_field(Doc, "$conflicts", ConflictRevTextList)
			end,
			{[Doc2 | Docs], GroupIdTableIdKeys};
		{not_found, deleted} ->
			{Docs, [{GroupId, []} | GroupIdTableIdKeys]}
		end,
	case process_info(self(), memory) of
	{memory, Mem} when Mem > ?FLUSH_MAX_MEM ->
		{TableGroup1, Results} = table_compute(TableGroup, Docs2),	
		{TableKVs3, GroupIdTableIdKeys3} = table_insert_query_results(Docs2, Results, TableKVs, GroupIdTableIdKeys2),
		{ok, TableGroup2} = write_changes(TableGroup1, TableKVs3, GroupIdTableIdKeys3, Seq),
		garbage_collect(),
		NotifyFun2 = NotifyFun(partial, TableGroup2),
		TableEmptyKeyValues = [{Table, []} || Table <- TableGroup2#table_group.tables],
		{ok, {[], TableGroup2, TableEmptyKeyValues, [], Seq, NotifyFun2}};
	_Else ->
		{ok, {Docs2, TableGroup, TableKVs, GroupIdTableIdKeys2, Seq, NotifyFun}}
	end.

table_insert_query_results([], [], TableKVs, GroupIdTableIdKeysAcc) ->
	{TableKVs, GroupIdTableIdKeysAcc};
table_insert_query_results([Doc|RestDocs], [QueryResults | RestResults], TableKVs, GroupIdTableIdKeysAcc) ->
	{NewTableKVs, NewTableIdKeys} = table_insert_doc_query_results(Doc, QueryResults, TableKVs, [], []),
	NewGroupIdTableIdKeys = [{Doc#doc.uuid, NewTableIdKeys} | GroupIdTableIdKeysAcc],
	table_insert_query_results(RestDocs, RestResults, NewTableKVs, NewGroupIdTableIdKeys).
	
	
table_insert_doc_query_results(_Doc, [], [], TableKVsAcc, TableIdKeysAcc) ->
	{lists:reverse(TableKVsAcc), lists:reverse(TableIdKeysAcc)};
table_insert_doc_query_results(#doc{uuid=DocId, revisions=[Rev|_]}=Doc, [{ok, Columns}|RestResults], [{Table, KVs}|RestTableKVs], TableKVsAcc, TableIdKeysAcc) ->
	case Columns of
	[] ->
		NewKVs = [{{[], DocId}, [Rev]}];
	[KeyColumn | ValueColumns] ->
		NewKVs = [{{KeyColumn, DocId}, [Rev|ValueColumns]}]
	end,
	NewTableIdKeys = [{Table#table.id, Key} || {{Key, _GroupId}, _Value} <- NewKVs],
	NewTableKVsAcc = [{Table, NewKVs ++ KVs} | TableKVsAcc],
	NewTableIdKeysAcc = NewTableIdKeys ++ TableIdKeysAcc,	
	table_insert_doc_query_results(Doc, RestResults, RestTableKVs, NewTableKVsAcc, NewTableIdKeysAcc);
table_insert_doc_query_results(Doc, [no_match|RestResults], [{Table, KVs}|RestTableKVs], TableKVsAcc, TableIdKeysAcc) ->
	NewTableKVsAcc = [{Table, KVs} | TableKVsAcc],
	table_insert_doc_query_results(Doc, RestResults, RestTableKVs, NewTableKVsAcc, TableIdKeysAcc).

table_compute(TableGroup, []) ->
	{TableGroup, []};
table_compute(#table_group{fabric_tables=FabricTables}=TableGroup, Docs) ->
	Tables =
	case FabricTables of
	nil ->
		{ok, Tables0} = couch_fabric:start_tables(queries(TableGroup)),
		Tables0;
	_ ->
		FabricTables
	end,
	{ok, Results} = couch_fabric:tables_compute_docs(Tables, Docs),
	{TableGroup#table_group{fabric_tables=Tables}, Results}.
	
queries(TableGroup) ->
	[Table#table.query_string || Table <- TableGroup#table_group.tables].
	

dict_find(Key, DefaultValue, Dict) ->
	case dict:find(Key, Dict) of
	{ok, Value} ->
		Value;
	error ->
		DefaultValue
	end.
		
write_changes(TableGroup, TableKeyValuesToAdd, GroupIdTableIdKeys, NewSeq) ->
	#table_group{uuid_btree=GroupIdBtree} = TableGroup,
			
	AddGroupIdTableIdKeys = [{GroupId, TableIdKeys} || {GroupId, TableIdKeys} <- GroupIdTableIdKeys, TableIdKeys /= []],
	RemoveGroupIds = [GroupId || {GroupId, TableIdKeys} <- GroupIdTableIdKeys, TableIdKeys == []],
	LookupGroupIds = [GroupId || {GroupId, _TableIdKeys} <- GroupIdTableIdKeys],
	
	{ok, LookupResults, GroupIdBtree2}
		= couch_btree:query_modify(GroupIdBtree, LookupGroupIds, AddGroupIdTableIdKeys, RemoveGroupIds),
	KeysToRemoveByTable = lists:foldl(
		fun(LookupResult, KeysToRemoveByTableAcc) ->
			case LookupResult of
			{ok, {GroupId, TableIdKeys}} ->
				lists:foldl(
					fun({TableId, Key}, KeysToRemoveByTableAcc2) ->
						dict:append(TableId, {Key, GroupId}, KeysToRemoveByTableAcc2)
					end,
					KeysToRemoveByTableAcc, TableIdKeys);
			{not_found, _} ->
				KeysToRemoveByTableAcc
			end
		end,
		dict:new(), LookupResults),
		
	Tables2 = [
		begin
			KeysToRemove = dict_find(Table#table.id, [], KeysToRemoveByTable),
			{ok, TableBtree2} = couch_btree:add_remove(Table#table.btree, AddKeyValues, KeysToRemove),
			Table#table{btree = TableBtree2}
		end
	||
		{Table, AddKeyValues} <- TableKeyValuesToAdd
	],
	
	TableGroup2 = TableGroup#table_group{tables=Tables2, current_seq=NewSeq, uuid_btree=GroupIdBtree2},
	
	{ok, TableGroup2}.
	
