%% CouchDb
%% Copyright (C) 2006  Damien Katz
%% 
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%% 
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%% 
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-module(couch_util).
-behaviour(gen_server).

-export([start_link/0,start_link/1, new_uuid/0, rand32/0, implode/2,parse_ini/1,ini_get_value/4,collate/3]).
-export([abs_pathname/1,abs_pathname/2,trim/1, test/0]).
-export([revid_to_list/1, list_to_revid/1]).

-export([init/1, terminate/2, handle_call/3]).
-export([handle_cast/2,code_change/3,handle_info/2]).

	
start_link() ->	
	start_link(".").
		
start_link(LibDir) ->	
	case erl_ddll:load_driver(LibDir, "couch_erl_driver") of
	ok -> ok;
	{error, already_loaded} -> ok;
	{error, ErrorDesc} -> exit({error, ErrorDesc})
	end,
	gen_server:start_link({local, couch_util}, couch_util, [], []).


new_uuid() ->
	gen_server:call(couch_util, new_uuid).

rand32() ->
	gen_server:call(couch_util, rand32).

abs_pathname(Filename) ->
	{ok, Cwd} = file:get_cwd(),
	abs_pathname(Filename, Cwd).
	
abs_pathname(Filename, Dir) ->
	Name = filename:absname(Filename, Dir ++ "/"),
	RevComponents = lists:reverse(filename:split(Name)),
	OutFilename = filename:join(pathname_work(RevComponents, [])),
	case string:right(Filename,1) of
	"/" ->
		OutFilename ++ "/";
	"\\" ->
		OutFilename ++ "/";
	_Else->
		OutFilename
	end.

% Is a character whitespace?
is_whitespace($\s)-> true;
is_whitespace($\t)-> true;
is_whitespace($\n)-> true;
is_whitespace($\r)-> true;
is_whitespace(_Else) -> false.

	
% removes leading and trailing whitespace from a string
trim(String) ->
	String2 = lists:dropwhile(fun is_whitespace/1, String),
	lists:reverse(lists:dropwhile(fun is_whitespace/1, lists:reverse(String2))).
	

pathname_work([], Acc) ->
	Acc;
pathname_work(["..", _PrevDir], _Acc) ->
	% last two elements, this means backup over the drive/root spec
	% therefore an error
	throw({error, invalid_dir});
pathname_work(["..", _PrevDir | Rest], Acc) ->
	pathname_work(Rest, Acc);
pathname_work([Dir | Rest], Acc) ->
	pathname_work(Rest, [Dir | Acc]).
	
	
implode(List, Sep) ->
	implode(List, Sep, []).

implode([], _Sep, Acc) ->
	lists:flatten(lists:reverse(Acc));
implode([H], Sep, Acc) ->
	implode([], Sep, [H|Acc]);
implode([H|T], Sep, Acc) ->
	implode(T, Sep, [Sep,H|Acc]).


% This is a simple ini parser. it just converts the string
% contents of a file like this:
%
%; comments are ignored
%;commentedoutvariable=foo
%this is line that gets ignored because it has no equals sign
%[this line gets ignored because it starts with a bracket but doesn't end with one
%bloodtype=Ragu
%[Some Section]
%timeout=30
%Default=zuh ; another comment (leading space or tab before a semi is necessary to be a comment if not at beginning of line)
%[Another Section]
%key with spaces=a value with stuff; and no comment
%oops="it doesn't qet quoted strings with semis quite right ; it thinks it's part comment"
%
%And converts it into this:
%[{{"","bloodtype","Ragu"},
% {{"thingy section","timeout"},"30"},
% {{"thingy section","Default"}, "zuh"},
% {{"another thingy section", "key with spaces"}, "a value with stuff; and no comment"},
% {{"another thingy section", "oops"}, "\"it doesn't qet quoted strings with semis quite right"}]
%

parse_ini(FileContents) ->
	{ok, Lines} = regexp:split(FileContents, "\r\n|\n|\r|\032"),
	{_, ParsedIniValues} =
	lists:foldl(fun(Line, {AccSectionName, AccValues}) ->
			case string:strip(Line) of
			"[" ++ Rest ->
				case regexp:split(Rest, "\\]") of
				{ok, [NewSectionName, ""]} ->
					{NewSectionName, AccValues};
				_Else -> % end bracket not at end, ignore this line
					{AccSectionName, AccValues}
				end;
			";" ++ _Comment ->
				{AccSectionName, AccValues};
			Line2 ->
				case regexp:split(Line2, "=") of
				{ok, [_SingleElement]} -> % no "=" found, ignore this line
					{AccSectionName, AccValues};
				{ok, [""|_LineValues]} -> % line begins with "=", ignore
					{AccSectionName, AccValues};					
				{ok, [ValueName|LineValues]} -> % yeehaw, got a line!
					RemainingLine = implode(LineValues, "="),
					{ok, [LineValue | _Rest]} = regexp:split(RemainingLine, " ;|\t;"), % removes comments
					{AccSectionName, [{{AccSectionName, ValueName}, LineValue} | AccValues]}
				end
			end					
		end, {"", []}, Lines),
	{ok, lists:reverse(ParsedIniValues)}.
	
ini_get_value(ParsedIniValues, Section, Key, DefaultValue) ->
	case lists:keysearch({Section, Key}, 1, ParsedIniValues) of
	{value, {_SectionKey, Value}} ->
		Value;
	_Else ->
		DefaultValue
	end.
	
init([]) ->
	{A,B,C} = erlang:now(),
	random:seed(A,B,C),
	{ok, dummy_server}.

terminate(_Reason, _Server) ->
	ok.
	
handle_call(new_uuid, _From, Server) ->
	{reply, new_uuid_int(), Server};
handle_call(rand32, _From, Server) ->
	{reply, rand32_int(), Server}.
	
handle_cast(_Msg, State) ->
	{noreply,State}.

code_change(_OldVsn, State, _Extra) ->
	{ok, State}.
	
handle_info(_Info, State) ->
	{noreply, State}.
	

new_uuid_int() ->
	% eventually make this a C callout for a real guid (collisions are far less likely
	% when using a proper generation function). For now we just fake it.
	Num1 = rand32_int(),
	Num2 = rand32_int(),
	Num3 = rand32_int(),
	Num4 = rand32_int(),
	lists:flatten(io_lib:format("~8.16.0B~8.16.0B~8.16.0B~8.16.0B", [Num1, Num2, Num3, Num4])).
	


rand32_int() ->
	random:uniform(4294967296) - 1.
	
drv_port() ->
	case get(couch_drv_port) of
	undefined ->
		Port = open_port({spawn, "couch_erl_driver"}, []),
		put(couch_drv_port, Port),
		Port;
	Port ->
	 Port
	end.
		
	
collate(A, B, Options) when is_list(A), is_list(B) ->
	Operation = 
	case lists:member(nocase, Options) of
		true -> 1;
		false -> 0
	end,
	Port = drv_port(),
	LenA = length(A),
	LenB = length(B),
	Bin = list_to_binary([<<LenA:32/native>>, A, <<LenB:32/native>>, B]),
	case erlang:port_control(Port, Operation, Bin) of
		[0] -> -1;
		[1] -> 1;
		[2] -> 0
	end.


revid_to_list(RevId) ->
	io_lib:format("~.16B", [RevId]).

list_to_revid(RevisionStr) ->
	case io_lib:fread("~16u", RevisionStr) of 
	{ok, [Int], []} ->
		Int;
	_Else ->
		% parse error
		throw({error_response, {bad_revision_id, RevisionStr}})
	end.


	
test() ->
	start_link("debug"),
	collate("a","b",[]).