%% CouchDb
%% Copyright (C) 2006  Damien Katz
%% 
%% This program is free software; you can redistribute it and/or
%% modify it under the terms of the GNU General Public License
%% as published by the Free Software Foundation; either version 2
%% of the License, or (at your option) any later version.
%% 
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%% 
%% You should have received a copy of the GNU General Public License
%% along with this program; if not, write to the Free Software
%% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

-module(mod_couch).

-include("couch_db.hrl").

-export([do/1, load/2, test/0, element_to_utf8/2, parse_xml/1]).
-export([doc_to_xml/1,doc_to_xml/2,doc_to_xmerl/2,xml_to_doc/1,parse_response/1]).

-include_lib("inets/src/httpd.hrl").
-include_lib("xmerl/include/xmerl.hrl").

-record(uri_parts,
	{db = "",
	doc = "",
	resource = "",
	querystr = ""}).

-record(doc_query_args,
	{full=false,
	options = [],
	revs = [],
	all_revs = false
	}).
	

-define(DOC_NAMESPACE, "http://couchdb.org/doc").
-define(TABLE_NAMESPACE, "http://couchdb.org/table").
	

%% do
	
do(ModData) ->
	#mod{request_uri=Uri,request_line=Request, parsed_header=Header,entity_body=Body} = ModData,
	case Uri of
	"/$utils/" ++ RestURI ->
		DocumentRoot = httpd_util:lookup(ModData#mod.config_db, document_root, ""),
		RealName = DocumentRoot ++ "/" ++ RestURI,
		{Path, AfterPath} = httpd_util:split_path(RealName),
		{proceed, [{real_name, {Path, AfterPath}} | ModData#mod.data]};
	_ ->
		couch_log:info("HTTP Request: ~s~nHeader:~p", [Request, Header]),
		couch_log:debug("Body:~P", [Body, 100]),
		Parts = parse_uri(Uri),
		{ok, ResponseCode} =
			case Parts of
			#uri_parts{db="$all_dbs"} ->
				return_all_dbs(ModData, Parts);
			#uri_parts{db=""} ->
				% empty request
				show_couch_welcome(ModData);	
			_Else ->
				case (catch handle_request(ModData, Parts)) of
				{ok, Response} ->
					{ok, Response};
				{error_response, Error} ->
					return_error(ModData, Error);
				Error ->
					return_error(ModData, Error)
				end
			end,
		couch_log:info("HTTP Response Code:~p~n", [ResponseCode]),
		{proceed, [{response, {already_sent, ResponseCode, 0}} | ModData#mod.data]}
	end.
	
	
parse_uri(RequestUri) ->
	% seperate out the path and query portions and
	% strip out leading slash and question mark.
	case httpd_util:split_path(RequestUri) of
	{[$/|UriPath], [$?|QueryStr]}	-> ok;
	{[$/|UriPath], QueryStr}		-> ok % there is no query
	end,
	
	case UriPath of
	"" ->
		#uri_parts{querystr=QueryStr};	
	UriPath ->
		case lists:last(UriPath) of
		$/ -> % ends with a slash means database only
			DbId = lists:sublist(UriPath, length(UriPath) - 1), % strip off trailing slash
			#uri_parts{db=DbId,querystr=QueryStr};
		_Else ->
			% lets try to parse out the UriPath
			{ok, SplitUrl} = regexp:split(UriPath, "/"),
			case SplitUrl of
			[DbId] -> % single element list.
				#uri_parts{db=DbId,querystr=QueryStr};
			_Else2 ->
				% multi element list. treat last element as the docid
				[DocIdDocResource | RevDbPath] = lists:reverse(SplitUrl),
				DbId = couch_util:implode(lists:reverse(RevDbPath), "/"),
				% Everything before first ":" is the docId,
				% and everything after is the resource id.
				{ok, [DocId | DocResourceList]} = regexp:split(DocIdDocResource, ":"), 
				ResourceId = couch_util:implode(DocResourceList, ":"),
				#uri_parts{db=DbId, doc=DocId, resource=ResourceId, querystr=QueryStr}
			end
		end
	end.

resp_header(ModData) ->
	resp_header(ModData, "text/html; charset=utf-8").
	
resp_xml_header(ModData) ->
	resp_header(ModData, "application/xml; charset=utf-8").

resp_header(#mod{http_version=Version}, ContentType) ->
	[{"cache-control", "no-cache"},
	{"pragma", "no-cache"},
	{"expires", httpd_util:rfc1123_date()},
	{"damien", "awesome"},
	{"content-type", ContentType},
	case Version == "HTTP/1.1" of
	true ->
		{"transfer-encoding", "chunked"};
	false ->
		{"connection", "close"}
	end].
	

send_header(ModData, RespCode, Headers) ->
	 couch_log:debug("HTTP response headers (code ~w):~p", [RespCode, Headers]),
	httpd_response:send_header(ModData, RespCode, Headers).
	
send_chunk(ModData, Data) ->
	httpd_response:send_chunk(ModData, Data, false).

send_final_chunk(ModData) ->
	httpd_response:send_final_chunk(ModData, false).
	
doc_to_xml(Doc) ->
	doc_to_xml(Doc, true).

doc_to_xml(Doc, DoFullDoc) ->
	lists:flatten(xmerl:export_simple([doc_to_xmerl(Doc, DoFullDoc)], xmerl_xml, [])).
	
revisions_to_xmerl(RevisionsList) ->
	Revs = [{rev,[{rev, couch_util:revid_to_list(Revision)}], []} || Revision <- RevisionsList],
	[{revisions, 
		[],
		Revs}].

doc_to_xmerl(Doc, DoFullDoc) ->
	FieldNodes = couch_doc:map(Doc, fun(Name, Value) -> 
			{field, [{name, Name}], list_to_xmerl(Value)}
		end),
	Attribs = [{id, Doc#doc.uuid}] ++ 
				case Doc#doc.revisions of
				[] -> [];
				_ -> [{rev, couch_util:revid_to_list(lists:nth(1, Doc#doc.revisions))}]
				end ++
				case Doc#doc.deleted of
				true -> [{deleted, "true"}];
				_ -> []
				end ++
					[{xmlns, ?DOC_NAMESPACE}],
	case DoFullDoc of
	true ->
		% put the revisions into a meta node. In the future other meta data
		% will go here.
		Revisions = revisions_to_xmerl(Doc#doc.revisions),
		MetaNode = {meta, [], Revisions},
		{doc, Attribs, [MetaNode | FieldNodes]};
	false ->
		{doc, Attribs, FieldNodes}
	end.



%converts a field/column list element to XML
list_to_xmerl(List) ->
	[element_to_xmerl(E) || E <- List].
	
element_to_xmerl(Element) when is_list(Element) ->
	{text, [Element]};
element_to_xmerl(Element) when is_float(Element) ->
	[Text] = io_lib:format("~.20f", [Element]),
	% Text is a now a number that looks something like "1.21000000000"
	% remove those trailing zeros
	StrippedRevText = lists:dropwhile(fun(C) -> C == $0 orelse C == $. end, lists:reverse(Text)),
	{number, [lists:reverse(StrippedRevText)]};
element_to_xmerl(Element) when is_integer(Element) ->
	{number, io_lib:format("~B", [Element])}.
	

show_couch_welcome(ModData) ->
	send_header(ModData, 200, resp_header(ModData)),
	send_chunk(ModData, "<body><h1>CouchDb</h1><br>"),
	send_chunk(ModData, "<h2>Welcome!</h2>"),
	send_chunk(ModData, "CouchDb-" ++ couch_server:get_version()),
	send_chunk(ModData, "</body>"),
	send_final_chunk(ModData),
	{ok, 200}.
	
handle_request(#mod{method="POST",entity_body=RawBody}=ModData, #uri_parts{db="$fabric_check", doc=""}) ->	
	case couch_fabric:check(RawBody, table) of
	ok ->
		{ok, [{ok, ColumnNames}]} = couch_fabric:table_columns([RawBody]),
		return_success(ModData, 200, [{columns, [], [{column, [{id, Name}], []} || Name <- ColumnNames]}]);
	Error ->
		return_error(ModData, Error)
	end;
handle_request(#mod{method="POST"}=ModData, #uri_parts{db=DbName, doc="$missing_revs"}=UriParts) ->
	handle_missing_revs_request(ModData, UriParts,  open_db(DbName));
handle_request(#mod{method="POST"}=ModData, #uri_parts{db=DbName, doc=""}=UriParts) ->
	% POST with no doc specified, therefore document addition
	handle_db_request(ModData#mod{method="PUT"}, UriParts,  open_db(DbName));
handle_request(#mod{method="PUT"}=ModData, #uri_parts{db=DbName, doc=""}=_UriParts) ->
	% put with no doc specified, therefore database creation
	case couch_server:create(DbName, []) of
	{ok, _Db} ->
		return_success(ModData, 201, []);
	{error, database_already_exists} ->
		Msg = io_lib:format("Database ~p already exists.", [DbName]),
		throw({error_response, {database_already_exists, Msg}});
	Error ->		
		Msg = io_lib:format("Error creating database ~p: ~p", [DbName, Error]),
		throw({error_response, {unknown_error, Msg}})
	end;
handle_request(#mod{method="DELETE"}=ModData, #uri_parts{db=DbName, doc=""}=_UriParts) ->
	% delete with no doc specified, therefore database delete
	case couch_server:delete(DbName) of
	ok ->
		return_success(ModData, 202, []);
	Error ->
		throw({error_response, Error})
	end;
handle_request(ModData, #uri_parts{db=DbName}=UriParts) ->
	handle_db_request(ModData, UriParts,  open_db(DbName)).
	
% returns db, otherwise throws exception. Note, no {ok,_}.
open_db(DbName) ->
	case couch_server:open(DbName) of
	{ok, Db} ->
		Db;
	Error ->
		throw({error_response, Error})
	end.
	
	
element_to_utf8(#xmlElement{attributes=Attributes, content=Bodies}=Entity, GlobalState) ->
	% convert attributes and text back to utf8
	Attributes2 = [Attr#xmlAttribute{value=xmerl_ucs:to_utf8(Str)}
		|| #xmlAttribute{value=Str}=Attr <- Attributes],
	Bodies2 = (catch lists:map(
		fun(#xmlText{value=Value}=TextEntity) ->
			TextEntity#xmlText{value=xmerl_ucs:to_utf8(Value)};
		(OtherEntity) ->
			OtherEntity
		end,
		Bodies)),
	{Entity#xmlElement{attributes=Attributes2, content=Bodies2}, GlobalState};
element_to_utf8(Entity, GlobalState) ->
	{Entity, GlobalState}.

parse_xml(XmlBody) ->
	case (catch xmerl_scan:string(XmlBody, [{quiet, true}, {hook_fun, fun element_to_utf8/2, []}])) of
	{'EXIT', {fatal, XmlError}} ->
		throw({error_response, {xml_error, XmlError}});
	{'EXIT', XmlError} ->
		throw({error_response, {xml_error, XmlError}});
	{XmerlDoc, _RemainingBytes} ->
		{ok, XmerlDoc}
	end.

parse_missing_rev_req([], DocIdRevsAcc) ->
	DocIdRevsAcc;
parse_missing_rev_req([#xmlElement{name=doc,content=Content}=Element], DocIdRevsAcc) ->
	% extract out the doc id and revisions
	[{get_req_attrib(id, Element), get_revisions(Content, [])} | DocIdRevsAcc];
parse_missing_rev_req([_ | Rest], DocIdRevsAcc) ->
	parse_missing_rev_req(Rest, DocIdRevsAcc).
	

handle_missing_revs_request(#mod{entity_body=RawXml}=ModData, _UriParts, Db) ->
	{ok, XmerlDoc} = parse_xml(RawXml),
	#xmlElement{name=missing_revs,content=Content} = XmerlDoc,
	DocIdRevsList = parse_missing_rev_req(Content, []),
	{ok, Results} = couch_db:get_missing_revs(Db, DocIdRevsList),
	MissingRevsXerml = [{doc, [{id, DocId}], revisions_to_xmerl(MissingRevs)} || {DocId, MissingRevs} <- Results],
	ResultXmerl = [{missing_revs, [], MissingRevsXerml}],
	return_body(ModData, 200, ResultXmerl).
	

handle_db_request(#mod{method="DELETE"}=ModData, #uri_parts{doc=DocId,querystr=QueryStr}, Db) ->
	% handle doc deletion
		
	#doc_query_args{revs=Revs} = doc_parse_query(QueryStr),
	
	Revs2 = case Revs of [] -> all; _ -> Revs end,
	
	case couch_db:delete_doc(Db, DocId, Revs2) of
	{ok, Results} ->
		
		XmerlDocResults = 
		lists:map(fun(Result) ->
				case Result of
				{ok, RevId} ->
					{updateinfo, [{success,"true"}, {rev, couch_util:revid_to_list(RevId)}], []};
				{conflict, RevId} ->
					{updateinfo, [{success,"false"}, {rev, couch_util:revid_to_list(RevId)}], []}
				end
			end, Results),
		return_success(ModData, 202, XmerlDocResults);
	Error ->
		throw({error_response, {delete_error, Error}})
	end;
handle_db_request(#mod{method="PUT",entity_body=RawBody}=ModData, #uri_parts{doc=DocId,querystr=QueryStr}, Db) ->
	% handle doc creation
	% just assume the content
	% is application/xml utf-8
	
	#doc_query_args{options=SaveOptions} = doc_parse_query(QueryStr),
	{ok, Xmerl} = parse_xml(RawBody),
	case Xmerl of 
	#xmlElement{name=docs, content=SubXmerlDocs} ->	
		% convert all the doc elements to native doc
		Docs =
		lists:reverse(
			lists:foldl(fun(SubXmerl, DocAcc) ->
					case SubXmerl of
					#xmlElement{name=doc} ->
						{Doc, _} = xmerl_to_doc(SubXmerl),
						[Doc#doc{uuid=DocId} | DocAcc];
					_Else ->
						DocAcc
					end
				end, [], SubXmerlDocs)),
				
		% save them
		{ok, Results} = couch_db:save_doc_revs(Db, Docs, SaveOptions),
		
		XmerlDocResults = 
		lists:map(fun(Result) ->
				case Result of
				{ok, RevId} ->
					{updateinfo, [{success,"true"}, {rev, couch_util:revid_to_list(RevId)}], []}
				end
			end, Results),
		Xmerl,
				
		XmerlInfo = [{update_results, [], XmerlDocResults}],
		return_body(ModData, 201, XmerlInfo, []);
		
	_Else ->
		{Doc, PrevRevId} = xmerl_to_doc(Xmerl),
		
		%if the docid is specified in the URL, use that (we override the XML documents ID).
		Doc2 =
		case DocId of
		[] ->
			Doc;
		_ ->
			Doc#doc{uuid = DocId}
		end,
		Doc3 =
		case PrevRevId of
		nil ->
			Doc2;
		PrevRevId ->
			% there is a "based on" revisionid
			% load the document from the database and see if its the same
			% revision. If so then snag the revision history and use it.
			case couch_db:open_doc(Db, Doc2#doc.uuid, [allow_stub]) of
			{ok, DiskDoc} ->
				case PrevRevId == lists:nth(1, DiskDoc#doc.revisions) of
				true ->
					Doc2#doc{revisions = DiskDoc#doc.revisions};
				false ->
					% oops, the revision on disk isn't what this doc is based
					% upon, a conflict!
					throw({error_response, {update_error, conflict}})
				end;
			{not_found, missing} ->
				Doc2;
			OldOpenError ->
				throw({error_response, {update_error, OldOpenError}})
			end
		end,
		case couch_db:save_doc(Db, Doc3, SaveOptions) of
		{ok, NewRevId} ->
			XmerlInfo = [{updateinfo, [{docid, Doc#doc.uuid}, {rev, couch_util:revid_to_list(NewRevId)}], []}],
			return_success(ModData, 201, XmerlInfo, [{"x-couch-docid", Doc#doc.uuid}, {"x-couch-rev", couch_util:revid_to_list(NewRevId)}]);
		Error ->
			throw({error_response, {update_error, Error}})
		end
	end;
handle_db_request(#mod{method="GET"}=ModData,
		#uri_parts{doc=DocId, resource=Resource}=UriParts, Db) ->
	% handle doc retrieval
	case {DocId, Resource} of
	{"", ""} ->
		return_database(ModData, UriParts, Db);
	{"$all_docs", ""} ->
		return_all_docs(ModData, UriParts, Db);
	{"$all_docs_by_update_seq", ""} ->
		return_all_docs_by_seq(ModData, UriParts, Db);
	{_DocId, ""} ->
		return_doc(ModData, UriParts, Db);
	{_DocId, _TableId} ->
		return_table(ModData, UriParts, Db)
	end.	
		

doc_parse_query(QueryStr) ->
	QueryList = httpd:parse_query(QueryStr),
	lists:foldl(fun({Key,Value}, Args) ->
			case {Key, Value} of
			{"full", "true"} ->
				Args#doc_query_args{full=true};
			{"latest", "true"} ->
				Options = [latest | Args#doc_query_args.options],
				Args#doc_query_args{options=Options};
			{"rev", RevIdStr} ->
				Revs = [couch_util:list_to_revid(RevIdStr) | Args#doc_query_args.revs],
				Args#doc_query_args{revs=Revs};
			{"all_revs", "true"} ->
				Args#doc_query_args{revs=true};
			{"new_edits", "true"} ->
				Options = [new_edits | Args#doc_query_args.options],
				Args#doc_query_args{options=Options};
				
			_Else -> % unknown key value pair, ignore.
				Args
			end
		end,
		#doc_query_args{}, QueryList).
	
xml_to_doc(RawXml) ->
	{ok, XmerlDoc} = parse_xml(RawXml),
	xmerl_to_doc(XmerlDoc).
	
xmerl_to_doc(#xmlElement{name=doc,content=DocContent}=DocElement) ->
	Deleted =
	case get_attrib(deleted, DocElement, false) of
		"true" -> true;	
		false -> false		
	end,
	Doc =
	case get_attrib(id, DocElement, nil) of
	nil ->
		#doc{deleted=Deleted}; % gets a new UUID by default
	Uuid ->
		#doc{deleted=Deleted, uuid=Uuid}
	end,
	PrevRevId =
	case get_attrib(rev, DocElement, nil) of
	nil ->
		nil;
	RevStr ->
		couch_util:list_to_revid(RevStr)
	end,	
	Doc2 =
	lists:foldl(
		fun(#xmlElement{name=meta,content=Content}, DocAcc) ->
			get_meta_data(Content, DocAcc);
		(#xmlElement{name=field,content=Content}=Element, DocAcc) ->
			FieldName = get_req_attrib(name, Element),
			Value = get_field_values(Content, []),
			couch_doc:set_field(DocAcc, FieldName, Value);
		(_Element, DocAcc) ->
			% let unknown elements pass. assume they are from the future!
			DocAcc
		end,
		Doc, DocContent),
	{Doc2, PrevRevId};
xmerl_to_doc(DocElement) ->
	% not a document, throw exception!
	throw({error_response, {invalid_element, {doc, DocElement}}}).
			

get_req_attrib(Atom, Element) ->
	case get_attrib(Atom, Element, nil) of
	nil ->
		throw({error_response, {missing_req_attribute, {Atom, Element}}});
	Value ->
		Value
	end.

get_attrib(Atom, #xmlElement{attributes=Attribs}, Default) ->
	Result =
	[AttribValue || #xmlAttribute{name=AttribAtom, value=AttribValue} <- Attribs, AttribAtom == Atom],
	case Result of
	[Value|_] ->
		Value;
	[] ->
		Default
	end.

get_meta_data([], Doc) ->
	Doc;
get_meta_data([#xmlElement{name=revisions,content=Content}|Rest], Doc) ->
	Doc2 = Doc#doc{revisions=get_revisions(Content, [])},
	get_meta_data(Rest, Doc2);
get_meta_data([_UnknownElement|Rest], Doc) ->	
	get_meta_data(Rest, Doc).
	

get_revisions([], Acc) ->
	lists:reverse(Acc);
get_revisions([#xmlElement{name=revisions,content=Content}|_], Acc) ->
	get_revisions(Content, Acc);
get_revisions([#xmlElement{name=rev}=Element|Rest], Acc) ->
	get_revisions(Rest, [couch_util:list_to_revid(get_req_attrib(rev,Element)) | Acc]);
get_revisions([_Element|Rest], Acc) ->
	% unknown element, ignore
	get_revisions(Rest, Acc).	

parse_float(Str, Element) ->
	case (catch list_to_integer(Str)) of
	Int when is_integer(Int) ->
		Int;
	_Else ->
		case (catch list_to_float(Str)) of
		Float when is_float(Float) ->
			Float;
		_Else ->
			throw({bad_number, {Str, Element}})
		end
	end.

get_field_values([], Acc) ->
	lists:reverse(Acc);
get_field_values([#xmlElement{name=text,content=Content} | Rest], Acc) ->
	Value = lists:flatten( [Value || #xmlText{value=Value} <- Content]),
	get_field_values(Rest, [Value|Acc]);
get_field_values([#xmlElement{name=number,content=Content}=Element | Rest], Acc) ->
	Value = lists:flatten( [Value || #xmlText{value=Value} <- Content]),
	get_field_values(Rest, [parse_float(couch_util:trim(Value), Element)| Acc]);
get_field_values([_| Rest], Acc) ->
	% unknown element, ignore
	get_field_values(Rest, Acc).
	
return_database(ModData, #uri_parts{db=DbName}, Db) ->
	{ok, InfoList} = couch_db:get_info(Db),
	Xmerl = [{database, [{id, DbName} | InfoList], []}],
	Xml = lists:flatten(xmerl:export_simple(Xmerl, xmerl_xml, [])),
	ok = send_header(ModData, 200, resp_xml_header(ModData)),
	ok = send_chunk(ModData, Xml),
	ok = send_final_chunk(ModData),
	{ok, 200}.
	
return_doc(ModData, #uri_parts{doc=DocId,querystr=QueryStr}, Db) ->	
	#doc_query_args{full=Full, options=OpenOptions, revs=Revs} = doc_parse_query(QueryStr),
	case Revs of
	[] ->
		case couch_db:open_doc(Db, DocId, OpenOptions) of
		{ok, Doc} ->
			return_body(ModData, 200, [doc_to_xmerl(Doc, Full)]);
		Error ->
			throw({error_response, Error})
		end;
	Revs ->
		{ok, Results} = couch_db:open_doc_revs(Db, DocId, Revs, OpenOptions),
		ok = send_header(ModData, 200, resp_xml_header(ModData)),
		ok = send_chunk(ModData, "<?xml version=\"1.0\"?><doc_revs id=\"" ++ DocId ++ "\">"),
		lists:foreach(fun(Result) ->
				case Result of
				{ok, Doc} ->
					XmerlDoc=doc_to_xmerl(Doc, Full),
					Xml = lists:flatten(xmerl:export_simple([XmerlDoc], xmerl_xml, [#xmlAttribute{name=prolog,value=""}])),
					ok = send_chunk(ModData, Xml);
				{{not_found, missing}, RevId} ->
					Xmerl = [{doc, [{id, DocId}, {rev, couch_util:revid_to_list(RevId)}, {missing, "true"}], []}],
					Xml = lists:flatten(xmerl:export_simple(Xmerl, xmerl_xml, [#xmlAttribute{name=prolog,value=""}])),
					ok = send_chunk(ModData, Xml)
				end
			end, Results),
		
		ok = send_chunk(ModData, "</doc_revs>"),
		ok = send_final_chunk(ModData),
		{ok, 200}
	end.

return_body(ModData, Code, XmerlBody) ->	
	return_body(ModData, Code, XmerlBody, []).
	
return_body(ModData, Code, XmerlBody, AdditionalHeaders) ->
	Xml = lists:flatten(xmerl:export_simple(XmerlBody, xmerl_xml, [])),
	ok = send_header(ModData, Code, resp_xml_header(ModData) ++ AdditionalHeaders),
	ok = send_chunk(ModData, Xml),
	ok = send_final_chunk(ModData),
	{ok, Code}.

return_success(ModData, Code, Xmerl) ->	
	return_success(ModData, Code, Xmerl, []).
	
return_success(ModData, Code, Xmerl, AdditionalHeaders) ->
	return_body(ModData, Code, [{success, [], Xmerl}], AdditionalHeaders).


-record(query_args,
	{start_keys = [],
	end_keys = [],
	count = 10000000000,	% a huge huge default number. Picked so we don't have
							% to do different logic for when there is no count limit
	update = true,
	direction = fwd,
	start_docid = nil,
	skip = 0
	}).
	
table_parse_query(QueryStr) ->
	QueryList = httpd:parse_query(QueryStr),
	QueryArgs =
	lists:foldl(fun({Key,Value}, Args) ->
			case {Key, Value} of
			{"key", Value} ->
				Keys = [Value|Args#query_args.start_keys],
				Args#query_args{start_keys=Keys,end_keys=Keys};
			{"startdoc", DocId} -> % old name of startkeys_docid, leaving to prevent breakage.
				Args#query_args{start_docid=DocId};
			{"startkeys_docid", DocId} ->
				Args#query_args{start_docid=DocId};
			{"startkey", Value} ->
				StartKeys = [Value|Args#query_args.start_keys],
				Args#query_args{start_keys=StartKeys};		
			{"endkey", Value} ->
				EndKeys = [Value|Args#query_args.end_keys],
				Args#query_args{end_keys=EndKeys};
			{"count", Value} ->
				case (catch list_to_integer(Value)) of
				Count when is_integer(Count) ->
					Args#query_args{count=Count};
				_Error ->
					Msg = io_lib:format("Bad URL query value, number expected: count=~s", [Value]),
					throw({error_response, {query_parse_error, Msg}})
				end;
			{"update", "false"} ->
				Args#query_args{update=false};
			{"reverse", "true"} ->
				Args#query_args{direction=rev};
			{"skip", Value} ->
				case (catch list_to_integer(Value)) of
				Count when is_integer(Count) ->
					Args#query_args{skip=Count};
				_Error ->
					Msg = io_lib:format("Bad URL query value, number expected: skip=~s", [Value]),
					throw({error_response, {query_parse_error, Msg}})
				end;
			_Else -> % unknown key value pair, ignore.
				Args
			end
		end,
		#query_args{}, QueryList),
	StartKeys = lists:reverse(QueryArgs#query_args.start_keys),
	EndKeys = lists:reverse(QueryArgs#query_args.end_keys),
	QueryArgs#query_args{start_keys=StartKeys,end_keys=EndKeys}.
	
		
xml_escape(Text) ->
	xmerl:export_simple([#xmlText{value=Text}], xmerl_xml, [#xmlAttribute{name=prolog,value=""}]).
	
	
	
row_to_xmerl({Uuid, Rev}, _Key, Columns) ->
	XmerlColumns =
	lists:map(fun({Name, Value}) ->
			{column, [{name, Name}], list_to_xmerl(Value)}
		end,
		Columns),
	{row, [{docid, Uuid}, {rev,couch_util:revid_to_list(Rev)}], XmerlColumns}.
	
	
make_table_fold_fun(ModData, #uri_parts{doc=DocId, resource=TableId}, QueryArgs) ->
	#query_args{
		end_keys=EndKeys,
		direction=Dir} = QueryArgs,
	case Dir of
	fwd ->
		% bin values sort higher. By appending to the list of keys,
		% it ensures that the stop key, even if shorter than the table key
		% will sort higher. This gives us the desired EndKey semantics.
		% If no end key was provided, then the stop key will sort higher than
		% all *valid* keys. Also use a bin as the default UUID, same reason.
		StopTableKey = {EndKeys ++ [<<>>], <<>>};
	rev ->
		StopTableKey = {EndKeys, nil}
	end,
	
	fun({Uuid, Rev}, Keys, Columns, Offset, TotalTableCount, {AccCount, AccSkip, HeaderSent}) ->
		PassedEnd =
		case Dir of
		fwd ->
			couch_table_group:less_fun(StopTableKey, {Keys, Uuid});
		rev ->
			couch_table_group:less_fun({Keys, Uuid}, StopTableKey)
		end,
		case PassedEnd of
		true ->
			% The stop key has been passed, stop looping.
			{stop, {AccCount, AccSkip, HeaderSent}};
		false ->
			
			case AccCount > 0 of
			true ->
				case AccSkip > 0 of
				true ->
					{ok, {AccCount, AccSkip - 1, HeaderSent}};
				false ->
					case HeaderSent of
					header_sent -> ok;
					_Else ->
						% We do this the first time through, NOT before folding,
						% because we want to make sure the table exists efficiently.
						ok = send_header(ModData, 200, resp_xml_header(ModData)),
						XmlBegin = lists:flatten(
							io_lib:format("<?xml version=\"1.0\"?><table id=\"~s:~s\" total_rows=\"~w\" result_start=\"~w\" xmlns=\"" ++ ?TABLE_NAMESPACE ++ "\">",
								[xml_escape(DocId), xml_escape(TableId), TotalTableCount, Offset + 1])),
						ok = send_chunk(ModData, XmlBegin)
					end,
					Xmerl = row_to_xmerl({Uuid, Rev}, Keys, Columns),
					Xml = lists:flatten(xmerl:export_simple([Xmerl], xmerl_xml, [#xmlAttribute{name=prolog,value=""}])),
					ok = send_chunk(ModData, Xml),				
					{ok, {AccCount - 1, 0, header_sent}}
				end;
			false ->
				{stop, {0, 0, HeaderSent}} % we've done "count" rows, stop foldling
			end
		end
	end.
	
	
return_table(ModData, #uri_parts{doc=DocId, resource=TableId, querystr=QueryStr}=UriParts, Db) ->
	QueryArgs = table_parse_query(QueryStr),
	#query_args{
		start_keys=StartKeys,
		count=Count,
		skip=SkipCount,
		update=Update,
		direction=Dir,
		start_docid=StartDocId} = QueryArgs,
	case Update of
	true ->
		case couch_db:update_table_group_sync(Db, DocId) of
		ok ->		ok;
		Error ->	throw({error_response, Error})
		end;
	false ->
		ok
	end,
	case {Dir, StartDocId} of
	{fwd, StartDocId} ->
		StartTableKey = {StartKeys, StartDocId};
	{rev, nil} ->
		% since this is a reverse traversal, wa add an extra value to the keys
		% that sorts higher (fwd we want to sort lower, rev, we want higher)
		% and we want our StartDocId to sort HIGHER than all valid ids.
		StartTableKey = {StartKeys ++ [<<>>], <<>>}; % a binary sorts highest of all values
	{rev, StartDocId} ->
		StartTableKey = {StartKeys ++ [<<>>], StartDocId}
	end,
	FoldlFun = make_table_fold_fun(ModData, UriParts, QueryArgs),	
	case (catch couch_db:fold_table(Db, DocId, TableId, StartTableKey, Dir, FoldlFun, {Count, SkipCount, header_not_sent})) of
	{ok, TotalCount, {_, _, header_not_sent}} ->
		% nothing found in the table, nothing has been returned
		% send empty table	
		ok = send_header(ModData, 200, resp_xml_header(ModData)),
		EmptyXml = lists:flatten(io_lib:format(
			"<?xml version=\"1.0\"?><table id=\"~s:~s\" total_rows=\"~w\" result_start=\"~w\" xmlns=\"" ++ ?TABLE_NAMESPACE ++ "\"></table>",
				[xml_escape(DocId), xml_escape(TableId), TotalCount, 0])),
		ok = send_chunk(ModData, EmptyXml),
		ok = send_final_chunk(ModData),
		{ok, 200};
	{ok, _TableInfo, {_, _, header_sent}} ->
		% end the table
		ok = send_chunk(ModData, "</table>"),
		ok = send_final_chunk(ModData),
		{ok, 200};
	Error2 ->
		throw({error_response, Error2})
	end.


return_all_dbs(ModData, _Parts)->
	{ok, DbNames} = couch_server:all_databases(),
	Xmerl = [{dbs, [], [{db, [], [DbName++"/"]} || DbName <- DbNames]}],
	Xml = lists:flatten(xmerl:export_simple(Xmerl, xmerl_xml, [])),
	ok = send_header(ModData, 200, resp_xml_header(ModData)),
	ok = send_chunk(ModData, Xml),
	ok = send_final_chunk(ModData),
	{ok, 200}.
		
	

return_all_docs(ModData, #uri_parts{querystr=QueryStr}, Db) ->
	QueryArgs = table_parse_query(QueryStr),
	#query_args{
		start_keys=StartKeys,
		count=Count,
		direction=Dir} = QueryArgs,
	StartKey =
	case StartKeys of
	[] ->
		nil;
	[FirstKey|_Rest] ->
		FirstKey
	end,
	ok = send_header(ModData, 200, resp_xml_header(ModData)),
	ok = send_chunk(ModData,
		"<?xml version=\"1.0\"?><table id=\"$all_docs\" xmlns=\"" ++ ?TABLE_NAMESPACE ++ "\">"),	
	couch_db:enum_docs(Db, StartKey, Dir,
		fun(#doc_info{uuid=Id, revision=Rev, deleted=Deleted}, AccCount) ->	
			case {Deleted, AccCount > 0} of
			{false, true} ->
				Xmerl = {row, [{docid, Id}, {rev, couch_util:revid_to_list(Rev)}], []},
				Xml = lists:flatten(xmerl:export_simple([Xmerl], xmerl_xml, [#xmlAttribute{name=prolog,value=""}])),
				ok = send_chunk(ModData, "\n" ++ Xml),
				{ok, AccCount - 1};
			{true, true} ->
				{ok, AccCount}; % skip
			{_, false} ->
				{stop, 0} % we've sent "count" rows, stop folding
			end
		end, Count),
	ok = send_chunk(ModData, "\n</table>"),
	ok = send_final_chunk(ModData),
	{ok, 200}.
	


return_all_docs_by_seq(ModData, #uri_parts{querystr=QueryStr}, Db) ->
	QueryArgs = table_parse_query(QueryStr),
	#query_args{
		start_keys=StartKeys,
		count=Count,
		direction=Dir} = QueryArgs,
	StartKey =
	case StartKeys of
	[] ->
		0;
	[FirstKey|_Rest] ->
		list_to_integer(FirstKey)
	end,
	ok = send_header(ModData, 200, resp_xml_header(ModData)),
	ok = send_chunk(ModData,
		"<?xml version=\"1.0\"?><docs_by_update_seq xmlns=\"" ++ ?TABLE_NAMESPACE ++ "\">"),	
	couch_db:enum_docs_since(Db, StartKey, Dir,
		fun(#doc_info{uuid=Id, revision=Rev, update_seq=UpdateSeq, deleted=Deleted, conflict_revs=ConflictRevs}, AccCount) ->
			case AccCount of			
			0 ->
				{stop, 0}; % we've sent "count" rows, stop folding
			_ ->
				ConflictRevsXmerl = 
				case ConflictRevs of
				[] ->
					[];
				_ ->
					RevXmerl = [{rev, [{rev, couch_util:revid_to_list(ConflictRev)}], []} || ConflictRev <- ConflictRevs],
					[{revisions, [{conflicts, "true"}], RevXmerl}]
				end,
		
				Attribs = case Deleted of true -> [{deleted,"true"}]; false -> [] end ++
					[{id, Id}, {rev, couch_util:revid_to_list(Rev)}, {update_seq, UpdateSeq}],
					
				Xmerl = {doc_info, Attribs, ConflictRevsXmerl},
				
				Xml = lists:flatten(xmerl:export_simple([Xmerl], xmerl_xml, [#xmlAttribute{name=prolog,value=""}])),
				ok = send_chunk(ModData, "\n" ++ Xml),
				{ok, AccCount - 1}
			end
		end, Count),
	ok = send_chunk(ModData, "\n</docs_by_update_seq>"),
	ok = send_final_chunk(ModData),
	{ok, 200}.
	
	
return_error(ModData, Error) ->
	case Error of
	not_found ->
		return_error(ModData, 404, not_found, "missing");
	{not_found, deleted} ->
		return_error(ModData, 404, deleted, "deleted");
	{not_found, Reason} ->
		return_error(ModData, 404, not_found, Reason);
	{database_already_exists, Reason} ->
		return_error(ModData, 409, database_already_exists, Reason); % 409, conflict error	
	{update_error, conflict} ->
		return_error(ModData, 409, conflict, conflict); % 409, conflict error	
	{Id, Reason} when is_atom(Id) ->
		return_error(ModData, 500, Id, Reason);
	_Else ->
		return_error(ModData, 500, unknown_error, Error)
	end.
		
	
return_error(ModData, RespCode, ErrorAtom, ErrorString) when is_list(ErrorString) ->
	FlatErrorString = lists:flatten(ErrorString),
	Xmerl = [{error, [{id, ErrorAtom}], [FlatErrorString]}],
	Xml = lists:flatten(xmerl:export_simple(Xmerl, xmerl_xml, [])),
	couch_log:info("-HTTP Error-~n\tResponse Code: ~p.~n\tError id: ~p~n\tError message: ~s", [RespCode,  ErrorAtom, ErrorString]),
	ok = send_header(ModData, RespCode, resp_xml_header(ModData)),
	ok = send_chunk(ModData, Xml),
	ok = send_final_chunk(ModData),
	{ok, RespCode};
return_error(ModData, RespCode, ErrorAtom, Error) ->
	ErrorString = io_lib:format("~p", [Error]),
	return_error(ModData, RespCode, ErrorAtom, ErrorString).
	
	
parse_response([]) ->
	{ok, []};
parse_response(RawXml) ->
	{ok, Xmerl} = parse_xml(RawXml),
	case Xmerl of
	#xmlElement{name=doc} ->
		{Doc, _} = xmerl_to_doc(Xmerl),
		{doc, Doc};
	#xmlElement{name=docs_by_update_seq,content=Content} ->
		DocInfoList =
		[#doc_info{
				uuid=get_req_attrib(id, Element),
				revision=couch_util:list_to_revid(get_req_attrib(rev, Element)),
				update_seq=list_to_integer(get_req_attrib(update_seq, Element)),
				conflict_revs=get_revisions(SubContent, [])
				}
			||
				#xmlElement{name=doc_info,content=SubContent}=Element <- Content],
		{docs_by_update_seq, DocInfoList};
	#xmlElement{name=doc_revs, content=Content} ->
		Docs =
		lists:foldl(fun(XmerlDoc, DocsAcc) ->
				case XmerlDoc of
				#xmlElement{name=doc} ->
					case get_attrib(missing, XmerlDoc, "") == "true" of
					true ->
						Rev = couch_util:list_to_revid(get_req_attrib(rev, XmerlDoc)),
						[{{not_found, missing}, Rev} | DocsAcc];
					false ->
						{Doc, _} = xmerl_to_doc(XmerlDoc),
						[{ok, Doc} | DocsAcc]
					end;
				_ ->
					DocsAcc
				end
			end, [], Content),
		{doc_revs, Docs};
	#xmlElement{name=missing_revs, content=Content} ->
		DocsRevsList =
		[{get_req_attrib(id, XermlDoc), get_revisions(MissingRevsXmerl, [])}
			|| #xmlElement{name=doc, content=MissingRevsXmerl}=XermlDoc <- Content],
		{missing_revs, DocsRevsList};
		
	#xmlElement{name=update_results, content=Content} ->
		RevsList =
		[begin
			get_req_attrib(success, XermlDoc) == "true",
			{ok, couch_util:list_to_revid(get_req_attrib(rev, XermlDoc))}
		end
			|| #xmlElement{name=updateinfo}=XermlDoc <- Content],			
		{update_results, RevsList};
	#xmlElement{name=error,content=Content} ->
		ErrorId = get_req_attrib(id, Xmerl),
		ErrorText = lists:flatten( [Value || #xmlText{value=Value} <- Content]),		
		{error,
			case {ErrorId, ErrorText} of
			{"deleted", _} ->
				{not_found, deleted};
			{"not_found", "missing"} ->
				{not_found, missing};
			{"not_found", Reason} ->
				{not_found, Reason};
			{"conflict", _} ->
				conflict;
			{"error", ErrorText} ->
				{error, ErrorText};
			_ ->
				{ErrorId, ErrorText}
			end};
	#xmlElement{name=success, content=[]} ->
		{success, []};
	#xmlElement{name=success, content=[#xmlElement{name=updateinfo}=Element|_]} ->
		{rev, couch_util:list_to_revid(get_req_attrib(rev,Element))};
	#xmlElement{name=success, content=Content} ->
		{success, lists:flatten( [Value || #xmlText{value=Value} <- Content])}
	end.
	
%%
%% Configuration
%%

%% load

load("Foo Bar", []) ->
	{ok, [], {script_alias, {"foo", "bar"}}}.
	
	


	
test() ->
	%check_formula(),
	delete_db("test/db"),
	create_db("test/db"),
	loop(1000, fun() -> create_doc("test/db", "foo") end),
	ok.
	
loop(0, _Fun) ->
	ok;
loop(N, Fun) ->
	Fun(),
	loop(N-1, Fun).

check_formula() ->
	Request = {"http://localhost:8888/$fabric_check", [], "text", "SELECT foo = = \"bar\"; COLUMN X; COLUMN Y := Text(Bar)"},
	_Response = http:request(post, Request, [], []),
	io:format(".................~n~p~n....................", [_Response]),
	ok.


create_doc(DbName, DocId) ->
	XmlDoc =
	"<?xml version='1.0'?>" ++
	"<doc>"
		"<field name=\"$table_foo\">" ++
			"<text>SELECT Foo != \"\"; COLUMN Foo; COLUMN BAR</text>" ++
		"</field>" ++ 
		"<field name=\"bar\">" ++
			"<text>BAR!!!</text>" ++
		"</field>" ++
		"<blah name=\"bar\">" ++
			"<text>BAR!!!</text>" ++
		"</blah>" ++
	"</doc>",
	
	% Request = {"http://localhost:8888/foo/WebCreatedDoc", [], "application/xml; charset=utf-8", XmlDoc},
	Request = {"http://localhost:8888/" ++ DbName ++ "/" ++ DocId, [], "application/xml; charset=utf-8", XmlDoc},
	_Response = http:request(put, Request, [], []),
		
	%io:format(".................~n~p~n....................", [_Response]),
	
	ok.
	

delete_doc(DbName, DocId) ->
	Request = {"http://localhost:8888/" ++ DbName ++ "/" ++ DocId, []},
	_Response = http:request(delete, Request, [], []),
		
	%io:format(".................~n~p~n....................", [Response]),
	ok.
	
	
	
create_db(DbName) ->
	Request = {"http://localhost:8888/" ++ DbName ++ "/", [], "application/xml; charset=utf-8", "<blank/>"},
	_Response = http:request(put, Request, [], []),
	%io:format(".................~n~p~n....................", [_Response]),
	ok.
	
	
delete_db(DbName) ->	
	Request = {"http://localhost:8888/" ++ DbName ++ "/", []},
	_Response = http:request(delete, Request, [], []),
	io:format(".................~n~p~n....................", [_Response]),
	ok.