/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef COUCH_DOCUMENT_H
#define COUCH_DOCUMENT_H


//
// Document - Interface for a document context object for the running formula
//
class Document
{
public:
	// Gets the named value from the document.
	// If the document doesn't have the value, it must return NULL.
	virtual const List* GetValue(MA* ma, const TextElement& fieldName) = 0;

	// Set the named value in the document.
	virtual void SetValue(const TextElement& fieldName, const List* pList) = 0;

	// Get the values of all fields wheose name starts with fieldName, and whose suffix is an integer 
	// with a value >= startoffset and <= endoffset
	//
	// Example, if a fieldName of "Foo" is passed with a startOffset of 2 and endOffset of 5,
	// the fields Foo2, Foo3, Foo4, and Foo5 would be concatenated and returned, but Foo1 and Foo6
	// would be ignored.
	virtual const List* CombineFields(MA* ma, const TextElement& fieldName, long startOffset, long endOffset) = 0;

	//Return all non-special data fields on a document.
	virtual const List* FieldNames(MA* ma) = 0;
};


// This is a utility class to make document field names sort
// lexigraphically in the name, and if there is a number suffix,
// the remaining digits are sort as a single number. For example
// this is how these two strings sort lexigraphically in
// std containers:
// "a10" < "a2"
// But when using FieldName , they sort like this:
// "a2" < "a10"
// Because the prefixes are the same ("a"), it  then compares the
// suffixes as integers (rather than strings).


class FieldName
{
public:
	FieldName()
		: fieldSuffix(-1)
	{
	}

	FieldName(const TextElement& fieldNameIn)
	{
		fieldName = fieldNameIn;
		fieldPrefix = fieldNameIn.ExtractTrailingNumber(&fieldSuffix);		
	}

	FieldName(const TextElement& fieldPrefixIn, long fieldSuffixIn)
		: fieldPrefix(fieldPrefixIn), fieldSuffix(fieldSuffixIn)
	{
		fieldName = fieldPrefixIn;
	}

	TextElement fieldName;
	TextElement fieldPrefix;
	long fieldSuffix;

	bool operator < (const FieldName& right) const
	{
		int cmp =  TextElement::CompareI(fieldPrefix, right.fieldPrefix);

		if (cmp == 0) {
			if (fieldSuffix < right.fieldSuffix) {
				cmp = -1;
			}
		}

		if (cmp < 0) {
			return true;
		}
		else {
			return false;
		}

	}
};


class DefaultDocument : public Document
{
	typedef fmap<const FieldName, const List*> FieldNameValueMap;

	FieldNameValueMap m_fieldValueMap;

public:
	DefaultDocument(MA* ma)
		: m_fieldValueMap(ma)
	{
	}

	const List* GetValue(MA* ma, const TextElement& fieldName)
	{
		FieldNameValueMap::const_iterator itor;
		itor = m_fieldValueMap.find(fieldName);

		if (itor != m_fieldValueMap.end()) {
			return itor->second;
		}
		return NULL;
	}

	void SetValue(const TextElement& fieldName,const List* pValue)
	{
		if (pValue && pValue->size()) {
			m_fieldValueMap[fieldName] = pValue;
		}
		else {
			m_fieldValueMap.erase(fieldName);
		}
	}

	const List* CombineFields(MA* ma, const TextElement& fieldNamePrefix, long startOffset, long endOffset)
	{
		if (endOffset < 0) {
			endOffset = 0x7FFFFFFF; //huge positive number
		}
		if (startOffset > endOffset) {
			return NULL;
		}

		FieldName beginField(fieldNamePrefix, (long)startOffset);
		FieldName endField(fieldNamePrefix, (long)endOffset);

		FieldNameValueMap::const_iterator itor;
		FieldNameValueMap::const_iterator itorEnd;

		itor = m_fieldValueMap.lower_bound(beginField);
		itorEnd = m_fieldValueMap.upper_bound(endField);

		List* pResultValue = new(ma) List(ma);

		while (itor != itorEnd) {
			pResultValue->Append(*itor->second);
			itor++;
		}

		return pResultValue;
	}

	//Get a list of field names present on the document
	const List* FieldNames(MA* ma)
	{
		List* pResultValue = new(ma) List(ma);

		const TextElement dollar("$");

		FieldNameValueMap::const_iterator itor = m_fieldValueMap.begin();
		FieldNameValueMap::const_iterator itorEnd = m_fieldValueMap.end();
		for(; itor != itorEnd; itor++) {
			const TextElement& fieldName = itor->first.fieldName;
			if (!fieldName.Begins(dollar)) {
				pResultValue->Add(fieldName);
			}
		}
		return pResultValue;
	}
};




#endif // COUCH_DOCUMENT_H
