/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Element.h"

#include "unicode/ucol.h"
#include "unicode/ucasemap.h"

#define NUM_CMP(a,b) ((a)<(b) ? -1 : ((a)>(b) ? 1 : 0))

unsigned char TextElement::DefaultNullChar = 0;

const Element Element::DefaultElement;

UCollator* g_collCaseInsensitive;
UCollator* g_collCaseSensitive;

bool InitCollate()
{
    UErrorCode status = U_ZERO_ERROR;

	g_collCaseSensitive = ucol_open("", &status);
	
	if (U_FAILURE(status)) return false;

	g_collCaseInsensitive = ucol_open("", &status);

	if (U_FAILURE(status)) return false;

    ucol_setAttribute(g_collCaseInsensitive, UCOL_STRENGTH, UCOL_PRIMARY, &status);
	
	if (U_FAILURE(status)) return false;

	return true;
}

UCaseMap* g_caseMap;

bool InitCaseMap()
{
    UErrorCode status = U_ZERO_ERROR;
	
	g_caseMap = ucasemap_open("", U_FOLD_CASE_DEFAULT, &status);

	if (U_FAILURE(status)) return false;

	return true;
}

int TextElement::Compare(const TextElement& left, const TextElement& right)
{
	UCharIterator leftIter, rightIter;
    uiter_setUTF8(&leftIter, left.Ptr(), left.LenInt());
    uiter_setUTF8(&rightIter, right.Ptr(), right.LenInt());
	UErrorCode status = U_ZERO_ERROR;
    return ucol_strcollIter(g_collCaseSensitive, &leftIter, &rightIter, &status); // ignore any error for now
}

int TextElement::CompareI(const TextElement& left, const TextElement& right)
{
	UCharIterator leftIter, rightIter;
    uiter_setUTF8(&leftIter, left.Ptr(), left.LenInt());
    uiter_setUTF8(&rightIter, right.Ptr(), right.LenInt());
	UErrorCode status = U_ZERO_ERROR;
    return ucol_strcollIter(g_collCaseInsensitive, &leftIter, &rightIter, &status); // ignore any error for now
}

int TextElement::Characters() const
{
	int i=0;
	const char* p = Ptr();
	int count = 0;
	int len = LenInt();
	while (i<len) {
		UChar32 c;
		count++;
		U8_NEXT(p, i, len, c);
	}

	return count;
}

// Used to seperate out the leading text and trailing number in a string.
// If the text string has trailing digits, it removes them and returns
// shortened string. It then assigns the number the the value at 
// pRetTrailingNumber.
//
// If there are no trailing digits, returns the TextElement unchanged and
// pRetTrailingNumber is assigned -1.
TextElement TextElement::ExtractTrailingNumber(long* pRetTrailingNumber) const
{
	const char* pName = Ptr();
	const char* pEnd = pName + m_len;
	const char* pTrailingNumberText = pEnd -1;
	//start at the last character and keep backing up until we hit a non-digit character
	while (pTrailingNumberText > pName && (*pTrailingNumberText >= '0' && *pTrailingNumberText <= '9')) {
		pTrailingNumberText--;
	}

	// put the character back onto the first number in the trailing digits
	// or past last char if no digits
	pTrailingNumberText++;
	
	char buffer[10];
	size_t numTrailingDigits = pEnd - pTrailingNumberText;

	if (numTrailingDigits == 0 || numTrailingDigits > sizeof(buffer) - 1) {
		// no trailing digits or too many digits
		*pRetTrailingNumber = -1;
	}
	else {
		// we have trailing digits, copy into null terminated buff
		memcpy(buffer, pTrailingNumberText, numTrailingDigits);
		buffer[numTrailingDigits] = 0; // null terminate
		*pRetTrailingNumber = atol(buffer);
	}

	return TextElement(pName, pTrailingNumberText - pName);
}

