/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef FABRIC_ELEMENT_H
#define FABRIC_ELEMENT_H
#include "Fabric.h"
#include <string.h>
#include <time.h>
#include "FabricSTL.h"
#include <math.h> 

#ifdef WIN32
#define snprintf _snprintf
#endif

#pragma pack(push, r1, 4)

bool InitCollate();
bool InitCaseMap();

// The order of the elements in the Enum determines their collation sequence.
enum ElementType
{	
	UNKNOWN,
	NUMBER,
	TIME,
	TEXT,
	RUNTIME_ERROR,
	KEYWORD,
};

class Element;
class TextElement;
class NumberElement;
class KeywordElement;
class ErrorElement;

class ElementBase
{
public:

	virtual const TextElement ToText(MA* ma) const = 0;
	virtual bool ToBool() const = 0;	
	virtual ElementType GetType() const = 0;

	virtual const NumberElement ToNumber() const;
	virtual const ErrorElement ToError() const;
	virtual const KeywordElement ToKeyword() const;

	virtual long ToLong() const;

	
	const Element ConvertTo(MA* ma, ElementType) const;

};

class TextElement : public ElementBase
{
	const unsigned char* m_p;
	size_t	m_len;

	static unsigned char DefaultNullChar;

public:

	TextElement()
		: m_p(&DefaultNullChar), m_len(0)
	{}

	TextElement(const char* pStr, size_t len)
		: m_p((const unsigned char*)pStr), m_len(len)
	{}

	TextElement(const char* pStr)
		: m_p((const unsigned char*)pStr), m_len(strlen(pStr))
	{}

	virtual bool ToBool() const
	{
		// non null strings are true
		return m_len > 0;
	}

	virtual ElementType GetType() const
	{
		return TEXT;
	}

	const char* Ptr() const
	{
		return (const char*)m_p;
	}

	size_t Length() const
	{
		return m_len;
	}

	const TextElement ToText(MA* ma) const
	{
		return *this;
	}

	TextElement Concat(MA* ma, const TextElement& other)
	{
		char* pNewBuffer = new(ma) char[m_len + other.m_len];

		// copy in our string
		memcpy(pNewBuffer, m_p, m_len);

		// copy in the other string
		memcpy(pNewBuffer + m_len, other.m_p, other.m_len);

		return TextElement(pNewBuffer, m_len + other.m_len);
	}

	virtual const NumberElement ToNumber() const;

	int Compare(const TextElement& right) const
	{
		return Compare(*this, right);
	}

	int CompareI(const TextElement& right) const
	{
		return CompareI(*this, right);
	}

	int Equals(const TextElement& right) const
	{
		return Compare(*this, right) == 0;
	}

	int EqualsI(const TextElement& right) const
	{
		return CompareI(*this, right) == 0;
	}

	static int Compare(const TextElement& left, const TextElement& right);
	static int CompareI(const TextElement& left, const TextElement& right);

	int Characters() const;

	bool operator < (const TextElement& right) const
	{
		return Compare(*this, right) == 1;
	}

	char* ToNullTermStr(MA* ma) const
	{
		char* pReturnValue = new(ma) char[m_len + 1];
		memcpy(pReturnValue, m_p, m_len);
		pReturnValue[m_len] = 0;

		return pReturnValue;
	}

	TextElement ExtractTrailingNumber(long* pTrailingNumber) const;

	static TextElement MakeCopy(MA* ma, const char* pString)
	{
		return MakeCopy(ma, pString, strlen(pString));
	}

	static TextElement MakeCopy(MA* ma, const char* pString, size_t len)
	{
		char* p = new(ma) char[len];
		memcpy(p, pString, len);

		return TextElement(p, len);
	}

	static TextElement MakeCopy(MA* ma, const TextElement& text)
	{
		char* p = new(ma) char[text.m_len];
		memcpy(p, text.m_p, text.m_len);

		return TextElement(p, text.m_len);
	}

	int LenInt() const
	{
		return (int)m_len;
	}

	bool Begins(const TextElement& t) const
	{
		if (t.m_len > m_len) {
			return false;
		}
		if (memcmp(t.m_p, m_p, t.m_len) == 0) {
			return true;
		}
		return false;
	}


};

#define TEXT_ELEMENT(stringLiteral) TextElement(stringLiteral, sizeof(stringLiteral)-1)

class NumberElement : public ElementBase
{
	double m_value;
public:
	NumberElement(double number)
		: m_value(number) {}
	NumberElement()
		: m_value(0.0) {}

	operator double() const
	{
		return m_value;
	}

	virtual bool ToBool() const
	{
		return m_value != 0;
	}

	virtual ElementType GetType() const
	{
		return NUMBER;
	}

	const TextElement ToText(MA* ma) const
	{
		char buffer[200];
		snprintf(buffer, sizeof(buffer)-1, "%.f", m_value);

		return TextElement::MakeCopy(ma, buffer);
	}

	const NumberElement ToNumber() const
	{
		return *this;
	}

	
};


class StringBuilder
{
	char* m_pBegin;
	char* m_pCurrent;
	size_t m_size;
	MA* ma;
public:

	StringBuilder(MA* maIn, size_t initialSize)
		: ma(maIn)
	{
		m_pBegin = m_pCurrent = new(ma) char[initialSize];
		*m_pBegin = 0;
		m_size = initialSize;
	}

	void Append(const char* str, size_t appendLen)
	{
		size_t currentLen = m_pCurrent - m_pBegin;
		size_t spaceLeft = m_size - currentLen;
		if (appendLen + 1 <= spaceLeft) {
			memcpy(m_pCurrent, str, appendLen);
			m_pCurrent += appendLen;
			*m_pCurrent = 0;
		}
		else {
			size_t newSize = (currentLen + appendLen)*2;
			char* pNew = new(ma) char[newSize];
			memcpy(pNew, m_pBegin, currentLen);
			memcpy(pNew + currentLen, str, appendLen);
			m_size = newSize;
			m_pBegin = pNew;
			m_pCurrent = pNew + currentLen + appendLen;
			*m_pCurrent = 0;
		}
	}

	StringBuilder& operator += (const char* str)
	{
		Append(str,strlen(str));
		return *this;
	}

	StringBuilder& operator += (const TextElement& t)
	{
		Append(t.Ptr(), t.Length());
		return *this;
	}
	
	StringBuilder& operator += (int ival)
	{
		char buffer[200];
		snprintf(buffer, sizeof(buffer)-1, "%i", ival);
		Append(buffer, strlen(buffer));
		return *this;
	}

	// always null terminated
	operator const char* ()
	{
		return m_pBegin;
	}

	size_t Length() const
	{
		return m_pCurrent - m_pBegin;
	}
};


class TimeElement : public ElementBase
{
public:
	double  time;
};


class ErrorElement : public ElementBase
{
	const char* m_typeId;
	const char* m_message;
public:
	ErrorElement(RuntimeException& e)
		: m_typeId(e.typeId()), m_message(e.what())
	{}

	ErrorElement(const char* typeId, const char* message)
		: m_typeId(typeId), m_message(message)
	{}

	virtual bool ToBool() const
	{
		return false;
	}

	virtual ElementType GetType() const
	{
		return RUNTIME_ERROR;
	}

	virtual const TextElement ToText(MA* ma) const
	{
		StringBuilder str(ma,100);
		str += m_typeId;
		str += ": ";
		str += m_message;
		return TextElement(str, str.Length());
	}

	const TextElement TypeId() const
	{
		return TextElement(m_typeId);
	}

	const ErrorElement ToError() const
	{
		return *this;
	}
};



class KeywordElement : public ElementBase
{
	// store the keyword text as a null terminated string
	// with its atom. Using the atom makes keyword comparisons
	// fast and using a null terminated string saves Element
	// space by not having a length member.
	const char* m_pKeyword;
	IdAtom m_keywordID;

public:
	
	KeywordElement(IdAtom keywordID, const char* pKeyword)
		: m_keywordID(keywordID), m_pKeyword(pKeyword)
	{
	}

	bool ToBool() const
	{
		return true;
	}

	ElementType GetType() const
	{
		return KEYWORD;
	}

	const TextElement ToText(MA* ma) const
	{
		return TextElement(m_pKeyword, strlen(m_pKeyword));
	}

	const KeywordElement ToKeyword() const
	{
		return *this;
	}

	IdAtom GetIdAtom()
	{
		return m_keywordID;
	}
};

#define MAXSIZE(A,B) (A > B ? A : B)

#define ELEMENT_UNION_SIZE MAXSIZE(MAXSIZE(MAXSIZE(\
	sizeof(NumberElement), \
	sizeof(TextElement)), \
	sizeof(ErrorElement)), \
	sizeof(KeywordElement))

class Element
{
	unsigned char m_element[ELEMENT_UNION_SIZE];

public:

	static const Element DefaultElement;

	Element(const Element& elementIn)
	{
		memcpy(m_element, elementIn.m_element, sizeof(m_element));
	}

	Element(const TextElement& text)
	{
		memcpy(m_element, &text, sizeof(text));
	}

	Element(const char* pValue,	size_t len)
	{
		TextElement* p = new((void*)m_element) TextElement(pValue, len);
	}

	Element(const NumberElement& number)
	{
		memcpy(m_element, &number, sizeof(number));
	}

	Element(const ErrorElement& element)
	{
		memcpy(m_element, &element, sizeof(element));
	}

	Element(const KeywordElement& element)
	{
		memcpy(m_element, &element, sizeof(element));
	}

	operator const ElementBase*() const
	{
		return (const ElementBase*)m_element; 
	}

	const ElementBase* operator->() const
	{
		return (const ElementBase*)m_element; 
	}

	ElementBase* operator->()
	{
		return (ElementBase*)m_element; 
	}

	Element(RuntimeException& e)
	{
		ErrorElement* p = new((void*)m_element) ErrorElement(e);
	}

	Element(KeywordElement& keyword)
	{
		memcpy(m_element, &keyword, sizeof(keyword));
	}


	Element()
	{
#ifdef WIN32
#ifdef _DEBUG
		//This is guard to make sure we don't accidently expand the size of
		// an element. 12 bytes, The first DWORD is vtable Id for the class,
		// and each element then has 8 bytes to work with.
		// we only check on 32 bit windows for dev purposes.
		int size = sizeof(Element);
		ASSERT(size == 12);
#endif
#endif
		TextElement* p = new((void*)m_element) TextElement();
	}
	

	int Equals(const Element& other) const
	{
		return Collate(*this, other) == 0;
	}

	// are the values equal, case insenstitive for text
	int EqualsI(const Element& other) const
	{
		return Collate(*this, other, true) == 0;
	}

	static int Collate(const Element& a, const Element& b, bool insensitive=false);

	class ConversionException : public RuntimeException
	{
	public:
		ConversionException(const char* message)
			: RuntimeException("conversion", message)
		{
		}
	};

private:
	Element(int);
	Element(long);
	Element(size_t);
};

inline const NumberElement TextElement::ToNumber() const
{
	char buffer[200];
	memcpy(buffer, m_p, std::min(m_len, sizeof(buffer)-1));
	buffer[std::min(m_len, sizeof(buffer)-1)] = 0; // null terminate
	return NumberElement(atof(buffer));
}


inline
long ElementBase::ToLong() const
{
	double d = ToNumber();
	return (long)d;
}

inline
const KeywordElement ElementBase::ToKeyword() const
{
	throw Element::ConversionException("Cannot convert value to a keyword");
}

inline
const ErrorElement ElementBase::ToError() const
{
	throw Element::ConversionException("Cannot convert value to a error");
}

inline
const NumberElement ElementBase::ToNumber() const
{
	throw Element::ConversionException("Cannot convert value to a number");
}

inline
const Element ElementBase::ConvertTo(MA* ma, ElementType type) const
{
	switch (type) {
	case TEXT:
		return ToText(ma);
	case NUMBER:
		return ToNumber();
	case RUNTIME_ERROR:
		return ToError();
	case KEYWORD:
		return ToKeyword();
	default:
		throw Element::ConversionException("Cannot convert value.");
	}
}

inline
int Element::Collate(const Element& aIn, const Element& bIn, bool bInsensitive)
{
	const ElementBase& a = *(ElementBase*)aIn.m_element;
	const ElementBase& b = *(ElementBase*)bIn.m_element;

	ElementType aType = a.GetType();
	ElementType bType = b.GetType();

	if (aType < bType) {
		return -1;
	} else if (aType > bType) {
		return 1;
	}

	switch (aType) {
		case NUMBER:
			{
			double dbl_a = ((NumberElement&)a);
			double dbl_b = ((NumberElement&)b);
			return cmp(dbl_a, dbl_b);
			}
		case TEXT:
			if (bInsensitive) {
				return TextElement::CompareI((TextElement&)a, (TextElement&)b);
			}
			else {
				return TextElement::Compare((TextElement&)a, (TextElement&)b);
			}
		case TIME:
			return cmp(((TimeElement&)a).time, ((TimeElement&)b).time);
		case RUNTIME_ERROR:
			return TextElement::CompareI(((ErrorElement&)a).TypeId(), ((ErrorElement&)b).TypeId());
		case KEYWORD:
			return cmp(((KeywordElement&)a).GetIdAtom(), ((KeywordElement&)b).GetIdAtom());
		default:
			return 0;
	}
}


#pragma pack(pop, r1)

#endif // FABRIC_ELEMENT_H
