/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Fabric.h"
#include "antlr/CharInputBuffer.hpp"
#include "antlr/CommonAST.hpp"
#include "FabricParser.hpp"
#include "FabricLexer.hpp"
#include "RuntimeTree.h"
#include "List.h"
#include "FabricSTL.h"
#include "Document.h"

using namespace antlr;
using namespace std;


bool TextElementILess::operator () (const TextElement& left, const TextElement& right) const
{
	return TextElement::CompareI(left, right) == -1;
}


void Formula::ProcessInit()
{
	if (!InitCollate()) {
		throw std::runtime_error("Error initializing collation.");
	}
	if (!InitCaseMap()) {
		throw std::runtime_error("Error initializing case map.");
	}
}

Formula::Formula()
	: indefMa(), m_destructables(&indefMa), m_idAtoms(&indefMa), m_columnLists(&indefMa)
{
	m_pRootNode = NULL;
	ma = &transMa;
	m_nextIdAtom = 1;
	m_pTrueList = NULL;
	m_pNilList = NULL;
	m_pFalseList = NULL;
	m_pCompileError = NULL;
	m_pDoc = NULL;
	m_pDefaultDoc = NULL;
	m_pFailedAssertHandler = NULL;
	m_pEarlyReturnList = NULL;
	m_selectedFlag = false;
}

Formula::~Formula()
{
	// destruct all the destructables
	size_t size = m_destructables.size();
	for (size_t i = 0; i < size; i++) {
		m_destructables[i]->Destruct();
	}
}

void Formula::Reset()
{
	transMa.Reset();
}


bool Formula::IsReady()
{
	return m_pRootNode != NULL;
}

bool Formula::SelectCompute(Document& doc)
{
    SetDocument(&doc);
	Execute();
	return GetSelected();
}

const fvector<const List*>& Formula::GetColumns()
{
	return m_columnLists;
}

void Formula::AppendColumnList(const List* pList)
{
	if (pList == NULL) {
		pList = NilList();
	}
	m_columnLists.push_back(pList);
}

fvector<TextElement> Formula::GetColumnNames()
{
	fvector<TextElement> result(ma);
	const RuntimeNode* pNode = m_pRootNode->GetFirstChild();
	
	while (pNode) {
		if (pNode->GetNodeType() == NodeType_ColumnAssignment) {
			result.push_back(pNode->GetTokenText());
		}
		pNode = pNode->GetNextSibling();
	}
	return result;
}

const char* Formula::CompileError()
{
	return m_pCompileError != NULL ? m_pCompileError : "";
}

void Formula::SetSelected(bool flag)
{
	m_selectedFlag = flag;
}

bool Formula::GetSelected()
{
	return m_selectedFlag;
}


// Compiles a formula and sets up the internal machinery for execution
void Formula::Compile(const char* pInputFormula, FormulaType formulaType )
{
	m_pCompileError = NULL;
	UnicodeCharBuffer inputBuffer(pInputFormula, strlen(pInputFormula));
	FabricLexer lexer(inputBuffer);
	FabricParser parser(lexer);
	ASTFactory ast_factory;

	parser.initializeASTFactory(ast_factory);
	parser.setASTFactory(&ast_factory);
	
	RefAST ast;

	try {
		if (formulaType == kTable) {
			parser.table_spec();
		}
		else {
			// simple formula
			parser.statements();
		}

		ast = parser.getAST();
		
	} catch(ANTLRException& e) {
		string str = e.toString();
		m_pCompileError = new(ma) char[str.size() + 1];
		memcpy(m_pCompileError, &str[0], str.size());
		m_pCompileError[str.size()] = 0;
		throw SyntaxException(m_pCompileError);
	}

	m_pRootNode = BuildTree(ast);
}

// Build a runtime tree from the Abstract Syntax Tree. Calls itself recursively
RuntimeNode* Formula::BuildTree(RefAST pAST)
{
	// create a runtime node that matches this AST node
	RuntimeNode* pRuntime = NewRuntimeNode(pAST->getType(), pAST->getText().c_str());

	RuntimeNode* pRuntimeChild;
	RefAST pChildAST = pAST->getFirstChild();
	int numChildren = 0;

	if (pChildAST) {
		// recurse down, building runtime subtrees for each syntax subtree
		pRuntimeChild = BuildTree(pChildAST);
		pRuntime->SetFirstChild(pRuntimeChild);
		numChildren++;

		pChildAST = pChildAST->getNextSibling();
		while (pChildAST) {
			RuntimeNode* pNextChild = BuildTree(pChildAST);
			pRuntimeChild->SetNextSibling(pNextChild);
			pRuntimeChild = pNextChild;
			numChildren++;

			pChildAST = pChildAST->getNextSibling();
		}
	}

	pRuntime->m_numChildren = numChildren;
	pRuntime->CompilerValidateArgsAndInit();

	return pRuntime;
}



const List* Formula::GetVariableValue(const TextElement& varId)
{
	return GetVariableValue(GetIdAtom(varId));
}

const List* Formula::GetVariableValue(IdAtom varIdAtom)
{
	fmap<const IdAtom, const List*>::iterator itor;
	itor = m_pCurrentVarScope->map.find(varIdAtom);
	if (itor == m_pCurrentVarScope->map.end()) {
		return NULL;
	}
	else {
		return itor->second;
	}
}

const List* Formula::GetValue(const TextElement& varId, IdAtom varIdAtom)
{
	const List* pRet = GetVariableValue(varIdAtom);
	if (pRet) {
		return pRet;
	}
	else {
		pRet = GetDocument().GetValue(ma, varId);
		if (pRet && pRet->size() == 0) {
			// make sure nil lists are returned as null pointers
			pRet = NULL;
		}
		return pRet;
	}
}


void Formula::SetVariableValue(const TextElement& varId, const List* pList)
{
	SetVariableValue(GetIdAtom(varId), pList);
}

void Formula::SetVariableValue(IdAtom varIdAtom, const List* pList)
{
	if (pList && pList->size() == 0) {
		// NIL list (zero elements), set to null
		pList = NULL;
	}

	m_pCurrentVarScope->map[varIdAtom] = pList;
}

const List* Formula::Execute()
{
	m_pCurrentVarScope = new(ma) VariableScope(ma);
	m_pUserFunctions = new(ma) AtomUserFunctionMap(ma);
	m_columnLists.clear();
	m_selectedFlag = false;

	const List* pList = Execute(m_pRootNode);

	if (m_selectedFlag) {
		// true selection statement always returns 1
		// (pList is result of last expression, not selection statement)
		return TrueList();
	}
	else {
		return pList;
	}
}

VariableScope& Formula::GetCurrentScope()
{
	return *m_pCurrentVarScope;
}


const List* Formula::Execute(const RuntimeNode* pRoot)
{
	// This is the main execution loop. It essentially does a
	// post order traversal of the execution tree, and the results
	// of each subtree's execution is fed as arguments to the
	// parent node.


	const RuntimeNode* pNode = pRoot;
	fvector<StackFrame> runtimeStack(ma); // a stack of stack frames
	fvector<const List*> argsStack(ma); //a stack of of all non-consumed arguments
	runtimeStack.reserve(100); // 100 arbitrarily chosen.
	argsStack.reserve(100);

	const List* pList = NULL;

	while (true) {	
		while (pNode) {
			// First traverse down the execution tree, creating stackframes
			// for each node.
			runtimeStack.push_back(StackFrame(argsStack));
			StackFrame& frame = runtimeStack.back();
			frame.pNode = pNode;
			pNode = pNode->FirstNodeToExecute(frame);
		}

		// execute the last node on the stack
		StackFrame& frame = runtimeStack.back();
		pNode = frame.pNode;

		try {
			pList = pNode->Execute(frame);

			if (pList == NULL) {
				pList = NilList();
			}
		} catch (RuntimeException& e) {
			// convert runtime exception into a value.
			pList = new(ma) List(ma, Element(e));
		}

		// now clear its frame and its args off the stack
		argsStack.resize(argsStack.size() - frame.m_argsStored);
		runtimeStack.pop_back();

		if (m_pEarlyReturnList) {
			// The execution set the early return value. This means we abort 
			// the execution and go up the stack popping frames until
			// we reach a NodeType_UserFunctionCall.

			pList = m_pEarlyReturnList;
			m_pEarlyReturnList = NULL;
			
			while (runtimeStack.size() &&
					runtimeStack.back().pNode->GetNodeType() != NodeType_UserFunctionCall) {
				pNode = runtimeStack.back().pNode;
				argsStack.resize(argsStack.size() - runtimeStack.back().m_argsStored);
				runtimeStack.pop_back();
			}
		}

		if (runtimeStack.size() == 0) {
			break; // no more nodes. all done
		}

		StackFrame& parentFrame = runtimeStack.back();

		pNode = parentFrame.pNode->NextNodeToExecute(parentFrame, pNode, pList);
	}

	return pList;
}


IdAtom Formula::GetIdAtom(const TextElement& ident)
{
	fmap<const TextElement, IdAtom, TextElementILess>::iterator itor;
	itor = m_idAtoms.find(ident);
	if (itor != m_idAtoms.end()) {
		// atom was found in the map, return it
		return itor->second;
	}
	else {
		// no match found in the map. Create a new mapping
		TextElement indefIdent = TextElement::MakeCopy(&indefMa, ident);
		IdAtom atom = m_nextIdAtom++;
		m_idAtoms.insert(std::pair<const TextElement, IdAtom>(indefIdent, atom));
		return atom;
	}
}

void Formula::PushVariableScope()
{
	//initializes a new variable scope for when calling functions
	ASSERT(m_pCurrentVarScope);
	if (m_pCurrentVarScope->pNext) {
		m_pCurrentVarScope = m_pCurrentVarScope->pNext;
		// clear all mapped values. we do this instead of calling the clear
		// method so the container doesn't have to realloc the nodes (since mem
		// doesn't get recovered its wasteful)
		fmap<const IdAtom, const List*>& map = m_pCurrentVarScope->map;
		fmap<const IdAtom, const List*>::iterator itor;
		for (itor = map.begin(); itor != map.end(); itor++) {
			itor->second = NULL;
		}
		m_pCurrentVarScope->inputArgs.clear();
	}
	else {
		VariableScope* pNew = new(ma) VariableScope(ma);
		m_pCurrentVarScope->pNext = pNew;
		pNew->pPrev = m_pCurrentVarScope;
		m_pCurrentVarScope = pNew;
	}
}

void Formula::PopVariableScope()
{
	ASSERT(m_pCurrentVarScope->pPrev);
	m_pCurrentVarScope = m_pCurrentVarScope->pPrev;
}

void Formula::SetEarlyReturnList(const List* pList)
{
	m_pEarlyReturnList = pList;
}

const UserFunctionDescriptor* Formula::GetUserFunction(IdAtom idAtom)
{
	AtomUserFunctionMap::const_iterator itor = m_pUserFunctions->find(idAtom);

	if (itor == m_pUserFunctions->end()) {
		return NULL;
	}
	else {
		return itor->second;
	}
}

void Formula::SetUserFunction(IdAtom idAtom, const UserFunctionDescriptor* pUserFunc)
{	
	(*m_pUserFunctions)[idAtom] = pUserFunc;
}

const List* Formula::TrueList()
{
	if (m_pTrueList == NULL) {
		m_pTrueList = new(&indefMa) List(&indefMa, NumberElement(1.0));
	}
	return  m_pTrueList;
}


const List* Formula::FalseList()
{	
	if (m_pFalseList == NULL) {
		m_pFalseList = new(&indefMa) List(&indefMa, NumberElement(0.0));
	}
	return  m_pFalseList;
}

const List* Formula::NilList()
{	
	if (m_pNilList == NULL) {
		m_pNilList = new(&indefMa) List(&indefMa);
	}
	return  m_pNilList;
}


const List* Formula::BoolList(bool trueOrFalse)
{
	return trueOrFalse ? TrueList() : FalseList();
}


void Formula::SetDocument(Document* pDoc)
{
	m_pDoc = pDoc;
}

Document& Formula::GetDocument()
{
	if (m_pDoc == NULL) {
		if (m_pDefaultDoc == NULL) {
			m_pDefaultDoc = new(ma) DefaultDocument(ma);
		}
		return *m_pDefaultDoc;
	}
	else {
		return *m_pDoc;
	}
}

FailedAssertHandler* Formula::GetFailedAssertHandler()
{
	return m_pFailedAssertHandler;
}

void Formula::SetFailedAssertHandler(FailedAssertHandler* pHandler)
{
	m_pFailedAssertHandler = pHandler;
}

void Formula::AddToDestructionQueue(Destructable* d)
{
	m_destructables.push_back(d);
}

