/*
Fabric formula engine

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef FABRIC_STL_H
#define FABRIC_STL_H

// This is a version of the STL that works with the fabric throw-away memory
// model. If defines an allocator thingy and some basic container classes
// and utilities.

// The code is this file is a modified version of code from article
// by Gabriel Fleseriu and Andreas Masur.
// http://www.codeguru.com/Cpp/Cpp/cpp_mfc/stl/article.php/c4079/


#include <vector>
#include <map>

class MA // Mem Alloc
{
public:
	virtual void* Alloc(size_t size) = 0;
	virtual void Free(void*) = 0;
	MA(){}
private:
	MA(const MA& maOther);
	void operator = (const MA& maOther);
};

template <typename T> struct fabric_allocator_base;

template<class _Ty>
	struct fabric_allocator_base
	{	// base class for generic allocators
	typedef _Ty value_type;
	};

		// TEMPLATE CLASS _Allocator_base<const _Ty>
template<class _Ty>
	struct fabric_allocator_base<const _Ty>
	{	// base class for generic allocators for const _Ty
	typedef _Ty value_type;
	};



inline void* operator new(size_t num_bytes, MA* ma)
{
	return ma->Alloc(num_bytes);
}

inline void operator delete(void*, MA* ma)
{
	ASSERT(false);
}

inline void* operator new[](size_t num_bytes, MA* ma)
{
	return ma->Alloc(num_bytes);
}

inline void operator delete[](void*, MA* ma)
{
	ASSERT(false);
}


template <typename T> class fabric_allocator;

template <> class fabric_allocator<void>
{
public:
	typedef void* pointer;
	typedef const void* const_pointer;
	// reference to void members are impossible.
	typedef void value_type;
	template <class U> struct rebind {
		typedef fabric_allocator<U> other;
	};
};    

namespace fabric_alloc{
	inline void destruct(char *){}
	inline void destruct(wchar_t*){}
	template <typename T> 
	inline void destruct(T *t){t->~T();}
} // namespace
    
template <typename T>
class fabric_allocator
{
public:
	
	typedef fabric_allocator_base<T> _Mybase;
	typedef typename _Mybase::value_type value_type;

	typedef value_type  *pointer;
	typedef value_type & reference;
	typedef const value_type  *const_pointer;
	typedef const value_type & const_reference;

    typedef size_t size_type;
    typedef ptrdiff_t difference_type;

	fabric_allocator(MA* maIn)
		: ma(maIn)
	{}

    template <class U> 
        struct rebind { typedef fabric_allocator<U> other; };
		fabric_allocator() : ma(0) {}
    pointer address(reference x) const {return &x;}
    const_pointer address(const_reference x) const {return &x;}
    pointer allocate(size_type size, fabric_allocator<void>::const_pointer hint = 0)
    {
		return static_cast<pointer>(ma->Alloc(size*sizeof(T)));
    }
    //for Dinkumware:
    char *_Charalloc(size_type n){return static_cast<char*>(ma->Alloc(n));}
    // end Dinkumware

	template <class U> fabric_allocator(const fabric_allocator<U>& other){ma=other.ma;}

    void deallocate(pointer p, size_type n)
    {
		ma->Free(p);
    }
    void deallocate(void *p, size_type n)
    {
		ma->Free(p);
    }
    size_type max_size() const throw() {return size_t(-1) / sizeof(value_type);}
    void construct(pointer p, const T& val)
    {
        new(static_cast<void*>(p)) T(val);
    }
    void construct(pointer p)
    {
        new(static_cast<void*>(p)) T();
    }
    void destroy(pointer p)
	{
		fabric_alloc::destruct(p);
	}
    static void dump(){};

	MA* ma;
};

template <typename T, typename U>
inline bool operator==(const fabric_allocator<T>&, const fabric_allocator<U>){return true;}

template <typename T, typename U>
inline bool operator!=(const fabric_allocator<T>&, const fabric_allocator<U>){return false;}


// For VC6/STLPort 4-5-3 see /stl/_alloc.h, line 464
// "If custom allocators are being used without member template classes support :
// user (on purpose) is forced to define rebind/get operations !!!"
#ifdef _WIN32
#define FABRIC_ALLOC_CDECL __cdecl
#else
#define FABRIC_ALLOC_CDECL
#endif

namespace std{
template <class _Tp1, class _Tp2>
inline fabric_allocator<_Tp2>& FABRIC_ALLOC_CDECL
__stl_alloc_rebind(fabric_allocator<_Tp1>& __a, const _Tp2*) 
{  
    return (fabric_allocator<_Tp2>&)(__a); 
}


template <class _Tp1, class _Tp2>
inline fabric_allocator<_Tp2> FABRIC_ALLOC_CDECL
__stl_alloc_create(const fabric_allocator<_Tp1>&, const _Tp2*) 
{ 
    return fabric_allocator<_Tp2>(); 
}

} // namespace std
// end STLPort

// Needed to subclass rather typedef these types in order to make the passing of the
// memory allocator objects into the constructors mandatory.


template <typename T>
class fvector : public std::vector<T, fabric_allocator<T> >
{
public:
	fvector(MA* ma)
		: std::vector<T, fabric_allocator<T> >(fabric_allocator<T>(ma))
	{
	}

	fvector(MA* ma, const T& t)
		: std::vector<T, fabric_allocator<T> >(1, t, fabric_allocator<T>(ma))
	{
	}
};

template <typename TKey, class TType, class TTraits = std::less<TKey> >
class fmap : public std::map<TKey, TType, TTraits, fabric_allocator< std::pair<TKey,TType> > >
{
public:
	fmap(MA* ma, TTraits& comp)
		: std::map<TKey, TType, TTraits, fabric_allocator< std::pair<TKey,TType> > >(comp, fabric_allocator< std::pair<TKey,TType> >(ma))
	{
	}
	fmap(MA* ma)
		: std::map<TKey, TType, TTraits, fabric_allocator< std::pair<TKey,TType> > >(TTraits(), fabric_allocator< std::pair<TKey,TType> >(ma))
	{
	}
};


#endif // FABRIC_STL_H

