/* 
Fabric formula tests
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/


#define FABRIC_ALLOW_NEW 1

#include "antlr/CharInputBuffer.hpp"
#include "antlr/CommonAST.hpp"
#include "FabricParser.hpp"
#include "FabricLexer.hpp"
#include "RuntimeTree.h"
#include "Fabric.h"
#include "Document.h"
#include <fstream>
#include <string.h>

#define INITIAL_FIELDS_DELIM "\n#initialfields"
#define FORMULA_DELIM "\n#formula"
#define TABLE_FORMULA_DELIM "\n#tableformula"
#define RESULTS_DELIM "\n#results"
#define RETURN_VALUE "%ret"
#define COLUMN_VALUE "%column"
#define TEST_DELIM "\n--"

#define WHITESPACE " \n\r\t"

#define PARSEERROR(text) throw text

using namespace antlr;
using namespace std;

void RunTests(char* pTheTests);

typedef vector< pair<char*, List*> > NameValuePairs;

class BasicFailedAssertHandler : public FailedAssertHandler
{
	void HandleFailedAssert()
	{
		ASSERT(false);
	}
};

int main(int argc, char* argv[])
{


	Formula::ProcessInit();

	char** testFileNames;
	int testFileNameCount;

	if (argc > 1) {
		//use the input files
		testFileNames = &argv[1];
		testFileNameCount = argc - 1;
	}
	else {
		// Use the default test files. If you add new test files, 
		// must add it to this list be run be default.
		char* defaultFiles[] =
					{
					"coretests.fab",
					"textfunctions.fab",
					"listfunctions.fab",
					};
		testFileNames = defaultFiles;
		testFileNameCount = sizeof(defaultFiles)/sizeof(char*);
	}

	for(int i=0; i<testFileNameCount; i++) {
		char* testFileName = testFileNames[i];

		ifstream is;
		is.open (testFileName, ios::binary );

		if (!is.is_open()) {
			printf("***ERROR***\n Couldn't open formula test file \"%s\"", testFileName);
		}
		else {
			// get length of file:
			is.seekg (0, ios::end);
			std::streamsize length = is.tellg();
			is.seekg (0, ios::beg);

			// allocate memory:
			char* buffer = new char [length + 1 /*prepended newline char*/ + 1 /*null char*/];
			buffer[0] = '\n';

			// read data as a block:
			is.read (buffer + 1,length);

			is.close();

			buffer[length + 1] = 0;

			RunTests(buffer);
		}
	}

}

char* LiteralizeString(char* pInput);

void PreorderDumpTree(RefCommonAST ast);
void PreorderDumpCompNodeTree(const RuntimeNode* ast);
void StacklessDumpTree(const RuntimeNode* pNode);

NameValuePairs ParseDocumentState(Formula* f, char* pDocumentStateStr);

void RunTests(char* pTheTests)
{
	char* pTest = pTheTests;

	while (pTest) {
		char* p	= pTest;
		char* pTestEnd = strstr(pTest, TEST_DELIM);

		if (pTestEnd) {
			*pTestEnd = 0;
			pTest = pTestEnd + strlen(TEST_DELIM);
		}
		else {
			pTest = NULL;
		}

		char* pInitialFields = NULL;
		char* pFormula = NULL;
		char* pResults = NULL;

		char* pTemp = strstr(p, INITIAL_FIELDS_DELIM);

		if (pTemp) {
			p = pTemp + strlen(INITIAL_FIELDS_DELIM);
			pInitialFields = p;
		}

		bool tableFormula = false;

		pFormula = strstr(p, FORMULA_DELIM);

		if (pFormula) {
			*pFormula = 0; // NULL terminate previous section
			pFormula = pFormula + strlen(FORMULA_DELIM); // skip over the delimiter
		}
		else {
			pFormula = strstr(p, TABLE_FORMULA_DELIM);
			if (!pFormula) {
				// no formula found, nothing more to do.
				continue;
			}
			tableFormula = true;
			*pFormula = 0; // NULL terminate previous section
			pFormula = pFormula + strlen(TABLE_FORMULA_DELIM); // skip over the delimiter
		}
		

		p = strstr(pFormula, RESULTS_DELIM);

		if (p) {
			*p = 0; // null terminate pFormula
			p = p + strlen(RESULTS_DELIM);
			pResults = p;
		}

		if (pFormula[strspn(pFormula, WHITESPACE)] == 0) {
			// nothing but whitespace, skip
			continue;
		}

		Formula f;
		BasicFailedAssertHandler handler;
		f.SetFailedAssertHandler(&handler);
		
		printf("-- Formula --\n%s \n", pFormula);

		try {			
			f.Compile(pFormula, tableFormula ? kTable : kSimple);
		} catch(Formula::SyntaxException& e) {
			printf("Parse exception: %s \n", e.what());
			ASSERT(false);
			continue;
		}

		DefaultDocument doc(f.ma);

		NameValuePairs docInitialFields;
		NameValuePairs docResults;
		
		if (pInitialFields) {
			docInitialFields = ParseDocumentState(&f, pInitialFields);
		}
		
		if (pResults) {
			docResults = ParseDocumentState(&f, pResults);
		}

		for (size_t i=0; i < docInitialFields.size(); i++) {
			pair<char*, List*> nameValue = docInitialFields[i];
			
			if (nameValue.second != NULL) {
				doc.SetValue(TextElement(nameValue.first, strlen(nameValue.first)), nameValue.second);
			}
		}

		f.SetDocument(&doc);
		const List* pResultValue = f.Execute();

		if (docResults.size()) {
			const fvector<const List*>& columns = f.GetColumns();
			size_t nthColumn = 0;
			for (size_t i=0; i < docResults.size(); i++) {
				pair<char*, List*> nameValue = docResults[i];

				TextElement fieldName(nameValue.first, strlen(nameValue.first));

				if (fieldName.CompareI(TEXT_ELEMENT(RETURN_VALUE)) == 0) {
					// its the special %ret field. That means compare with the result,
					// not an document field.

					ASSERT(pResultValue->Equals(*nameValue.second));
				}
				else if (fieldName.CompareI(TEXT_ELEMENT(COLUMN_VALUE)) == 0) {
					// its the special %column field. That means compare with the next column result
					ASSERT(columns[nthColumn++]->Equals(*nameValue.second));
				}
				else {
					const List* pDocValue = doc.GetValue(f.ma, fieldName);

					if (pDocValue == NULL) {
						ASSERT(nameValue.second->size() == 0);
					}
					else {
						ASSERT(pDocValue->Equals(*nameValue.second));
					}
				}
			}
		}
		else {
			ASSERT(pResultValue->size() == 1);
			ASSERT(pResultValue->At(0)->GetType() == NUMBER);
			ASSERT(pResultValue->At(0)->ToNumber() == 1.0);
		}
	}

}

/* converts a string buffer a vector of name value pairs. The string buffer is this format:

valuename "literalvalue1" 123 123.345 work on [12/12/2006 12:03 pm]
valuename2 "literalvalue2" "foo1" "foo2" "she said \"bite me\""

valuename3 "literalvalue3"

*/

NameValuePairs ParseDocumentState(Formula* f, char* pDocumentStateStr)
{
	char* pNextLine = pDocumentStateStr;

	NameValuePairs results;
	
	while (pNextLine) {
		char* pCurrent = pNextLine;
		pNextLine = strchr(pNextLine, '\n');
		if (pNextLine) {
			*pNextLine++ = 0;
		}

		pCurrent += strspn(pCurrent, WHITESPACE); // skip any whitespace

		if (*pCurrent == 0) {
			// empty line
			continue;
		}

		char* pValueName = pCurrent;

		pCurrent += strcspn(pCurrent, WHITESPACE); // find next whitespace

		if (*pCurrent != 0) {
			*pCurrent++ = 0;
		}

		pCurrent += strspn(pCurrent, WHITESPACE); // skip any whitespace

		List* pValue = new(f->ma) List(f->ma);

		 while (*pCurrent != 0) {
			char c = *pCurrent;

			// sniff the first character to see what kind of element it is.

			if (c == '"') {
				// its a literal
				char* pLiteral = pCurrent;

				// pCurrentNow points the byte after the closing quote
				pCurrent = LiteralizeString(pLiteral);

				pValue->Add(TextElement(pLiteral, strlen(pLiteral)));
			}
			else if (c == '-' || (c >= '0' && c <= '9')) {
				// its a number
				char* pLiteral = pCurrent;
				size_t end = strcspn(pLiteral, WHITESPACE); // find the end
				if (pLiteral[end] == 0) {
					pCurrent = pLiteral + end;
				}
				else {
					pLiteral[end] = 0; // null out and go to next char
					pCurrent = pLiteral + end + 1;
				}

				pValue->Add(TextElement(pLiteral, strlen(pLiteral)).ToNumber());
			}
			else if (c == '[') {
				// its a keyword or date
				char* pLiteral = pCurrent + 1;
				char* pEnd = strchr(pLiteral, ']'); // find the end
				if (pEnd == NULL) {
					PARSEERROR("Unterminated keyword/datetime element");
				}

				*pEnd++ = 0; // null terminate and advance pointer

				pCurrent = pEnd;

				char c2 = *(pLiteral + 1);

				if (c2 >= '0' && c2 <= '9') {
					// its a date
					ASSERT(false); // not done yet

				}
				else {
					//its a keyword
					IdAtom id = f->GetIdAtom(pLiteral);
					KeywordElement keyword(id, pLiteral);
					pValue->Add(keyword);
				}				
			}
			else {
				PARSEERROR("Unknown element type");
			}

			pCurrent += strspn(pCurrent, WHITESPACE); // skip any whitespace

		}

		results.push_back( pair<char*,List*>(pValueName, pValue));
	}

	return results;
}

// Converts a quoted, escaped string into it's literal value, in-place. The function
// removes the surrounding quotes and escape characters from the string and converts
// the control characters to their proper value. It then null terminates the string.
//
// Returns: a pointer the character following the final quote in the original string.
char* LiteralizeString(char* pInput)
{
	ASSERT(*pInput == '"');

	char* pInsert = pInput;
	char* pRead = pInput + 1;
	while (true) {
		switch (*pRead) {
			case 0:
				PARSEERROR("Unterminated string");
				break;
			case '"':
				*pInsert = 0;
				return pRead + 1;

			case '\r':
				// strip these out (they can be explicitly added with escape chars)
				pRead++;
				break;

			case '\\':
				// its the escape ccharacter. Lets use the built in C++
				// escape system.
				{
				pRead++; //skip past the escape character
				char nextChar = *pRead; // get the character escaped
				pRead++; //skip past the character escaped

				// strip out 

				switch (nextChar) {
					case 'n' : // newline 
						nextChar = '\n';
						break;

					case 'r' : // carriage return 
						nextChar = '\r';
						break;

					case 't' : // tab
						nextChar = '\t';
						break;
				}

				*pInsert++ = nextChar;
				}
				break;

			default:
				*pInsert++ = *pRead++;
				break;
		}
	}
}




