/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

header "pre_include_hpp"
{
#include "antlr/CommonAST.hpp"
#include "antlr/UnicodeCharScanner.hpp"
#include "antlr/UnicodeCharBuffer.hpp"
}

options	{
	language="Cpp";
}

class FabricParser extends Parser;

options {
	ASTLabelType = "ANTLR_USE_NAMESPACE(antlr)RefCommonAST";
	buildAST=true;
	defaultErrorHandler=false;
}

tokens {
	FIELD="field";
	COLUMN="column";
	SELECT="select";
	IF="if";
	TRUELITERAL="true";
	FALSELITERAL="false";
	NILLITERAL="nil";
	IN="in";
}

anyformula
	:	(table_spec) => table_spec
	|	statements
	;

statements
	:	(statement)*
			{
			##=#(#[OUTERFORMULA,"OUTERFORMULA"], ##);
			}
	;
	
table_spec
	:	((toplevelstatement)* selectstatement (toplevelstatement)*)
			{
			##=#(#[OUTERFORMULA,"OUTERFORMULA"], ##);
			}
	;
	
terminator // statement terminator
	:	SEMIC! | ENDLINE!
	;

selectstatement
	// SELECT * means SELECT true
	:	SELECT^ (MULTIPLY! | expr) terminator
	;
	
toplevelstatement
	:	COLUMN! id:IDENT!
			(	(terminator) {##=#(#[COLUMN, id->getText()], ##);}
			|	(ASSIGN! exp1:expr  terminator) {##=#(#[COLUMNASSIGN,id->getText()], exp1);}
			|	(ASSIGNINIT! exp2:expr  terminator) {##=#(#[COLUMNINIT,id->getText()], exp2);}
			)
			
	|	statement
	;

statement
	:	expr terminator
	|	terminator
	;
	

expr
	:	FIELD! id:IDENT!
			(	(ASSIGN^ exp1:expr) {##=#(#[FIELDASSIGN, id->getText()], exp1);}
			|	(ASSIGNINIT^ exp2:expr) {##=#(#[FIELDINIT, id->getText()], exp2);}
			)
	|	(IDENT (ASSIGN | ASSIGNINIT)) =>  id2:IDENT!
		(	(ASSIGN! exp3:expr) {##=#(#[ASSIGN, id2->getText()], exp3);}
		|	(ASSIGNINIT! exp4:expr) {##=#(#[ASSIGNINIT, id2->getText()], exp4);}
		) 
	|	booleanexpr
	;
	
booleanexpr
	:	comparisonexpr ((AND^ | OR^) comparisonexpr)*
	;
	
comparisonexpr
	:	concatexpr
			((	
				(	EQUALEQUAL^
				|	LESS^
				|	GREATER^
				|	CASEINSENSITIVEMATCH^
				|	LESSEQUAL1^
				|	LESSEQUAL2^
				|	GREATEREQUAL1^
				|	GREATEREQUAL2^
				|	NOTEQUAL1^
				|	NOTEQUAL2^
				|	NOTEQUAL3^
				|	NOTEQUAL4^
				|	IN^
				)
				
				concatexpr)*
			|	REGEXP^
				)
	;

concatexpr
	:	addexpr (COLON^ addexpr (COLON! addexpr)*)?
	;
	
addexpr
	:	mexpr
		(	(PLUS^ mexpr (PLUS! mexpr)*)
		|	(MINUS^ mexpr (MINUS! mexpr)*)
		)?
	;

mexpr	   
	:	arrayexpr ((MULTIPLY^ | DIVIDE^) arrayexpr)*
	;
	
arrayexpr	   
	:	unaryopexpr
		(
			LBRACK! expr (COMMA! expr)? RBRACK!
				{
				##=#(#[ARRAYSUBSCRIPT,"ARRAYSUBSCRIPT"], ##);
				}
		)?
		
	;

unaryopexpr
	:	(NOT^ unaryopexpr)
	|	(MINUS! unaryopexpr)
			{
			##=#(#[NEGATE,"NEGATE"], ##);
			}
	|	PLUS! unaryopexpr
	|	atom
	;


atom
	:	REAL^
	|	INTEGER^
	|	STRINGLITERAL^
	|	DATETIME^
	|	userfunctionexpr
	|	ifexpr
	|	forallexpr
	|	corefunctions
	|	keywordconstantexpr
	|	id:IDENT^
			(	(PLUSPLUS^ | MINUSMINUS^)		// identifier++ identifier--
			|	FIELDCONCATENATION^ (atom | MULTIPLY!) //identier1..N
			|	(LPAREN! args:argsexpr RPAREN!)	// identifier(args)
					{
					#atom=#(#[FUNCTIONCALL, id->getText()], args);
					}
			)?
	|	nestedexprs
	|	LPAREN! expr RPAREN!
	|	USERFUNCTIONCALL^ LPAREN! argsexpr RPAREN!
	;
	
keywordconstantexpr
	:	TRUELITERAL^
	|	FALSELITERAL^
	|	NILLITERAL^
	;
	
forallexpr
	:	"forall"! (LPAREN! forallin (COMMA! forallin)*  RPAREN!) nestedexprs
			{
			##=#(#[FORALL,"FORALL"], ##);
			}
	;
	
	
userfunctionexpr
	: 	"function"! id:IDENT! LPAREN! (IDENT (COMMA! IDENT)* )? RPAREN! nestedexprs
			{
			##=#(#[USERFUNCTION, id->getText()], ##);
			}
	;
	
	
ifexpr
	:	IF^ LPAREN! expr	(	RPAREN! nestedexprs ("else"! ( nestedexprs | ifexpr2 ))?
							|	(COMMA! expr)+ RPAREN!
							)
	;

ifexpr2
	:	IF! LPAREN! expr RPAREN! nestedexprs ("else"! ( nestedexprs | ifexpr2 ))?
	;
	
nestedexprs
	:	(LCURLY! (statement)* RCURLY!)
			{
			##=#(#[NESTEDEXPRS,"NESTEDEXPRS"], ##);
			}
	;
	
argsexpr
	:	| expr (COMMA! expr)*
	;
	
multipleforallin
	: forallin (COMMA! forallin)*
	;
	
forallin
	:	id:IDENT! "in"!
			expr ("by"! INTEGER)?
			{
			##=#(#[FORALLIN,id->getText()], ##);
			}
	;
	
corefunctions
	:	(	"assert"^
		|	"assertfalse"^
		|	"repeat"^
		|	"return"^
		|	"text"^
		|	"error"^
		|	"uppercase"^
		|	"lowercase"^
		|	"left"^
		|	"right"^
		|	"leftback"^
		|	"rightback"^
		|	"split"^
		|	"explode"^
		|	"join"^
		|	"implode"^
		|	"length"^
		|	"count"^
		|	"list_trim"^
		|	"trim"^
		|	"fulltrim"^
		|	"matches"^
		|	"iserror"^
		|	"number"^
		|	"docfields"^
		|	"getfield"^
		)
		LPAREN! argsexpr RPAREN!
	;
 
class FabricLexer extends Lexer("UnicodeCharScanner");

options	{
	k=3;
	// Allow any char but \uFFFF (16 bit -1)
    charVocabulary='\u0000'..'\uFFFE';
    caseSensitiveLiterals=false;
	defaultErrorHandler=false;
}

tokens {
	FUNCTIONCALL;
	NESTEDEXPRS;
	ARRAYSUBSCRIPT;
	NEGATE;
	ASSIGNTARGET;
	OUTERFORMULA;
	FIELDASSIGN;
	FORALLIN;
	FORALL;
	USERFUNCTION;
	COLUMNASSIGN;
	COLUMNINIT;
	FIELDINIT;
	VARINIT;
}

COMMA
options { paraphrase = "a ','"; }
	: ',' ;

LPAREN
options { paraphrase = "a '('"; }
	: '(' ;

RPAREN
options { paraphrase = "a ')'"; }
	: ')' ;

LCURLY
options { paraphrase = "a '{'"; }
	: '{' ;

RCURLY
options { paraphrase = "a '}'"; }
	: '}' ;
	
PLUS
options { paraphrase = "a '+'"; }
	: '+' ;
	  
MINUS
options { paraphrase = "a '-'"; }
	: '-' ;
	
PLUSPLUS
options { paraphrase = "a '++'"; }
	: '+' '+' ;
	
MINUSMINUS
options { paraphrase = "a '--'"; }
	: '-' '-' ;
	
MULTIPLY
options { paraphrase = "a '*'"; }
	  : '*' ;

DIVIDE
options { paraphrase = "a '/'"; }
	  : '/' ;
	  
SEMIC
options { paraphrase = "a ';'"; }
	 : ';' ;

	 
ASSIGN
options { paraphrase = "a ':='"; }
	: ':' '=' ;
	 
ASSIGNINIT
options { paraphrase = "a '?='"; }
	: '?' '=' ;
	
COLON
options { paraphrase = "a ':'"; }
	 : ':' ;
	 
CASEINSENSITIVEMATCH
options { paraphrase = "a '=^'"; }
	 : '=' '^' ;
	 
EQUALEQUAL
options { paraphrase = "a '=='"; }
	 : '=' '=' ;
	 
LESS
options { paraphrase = "a '<'"; }
	: '<' ;
	
GREATER
options { paraphrase = "a '>'"; }
	: '>' ;
	
LESSEQUAL1
options { paraphrase = "a '<'"; }
	: '<' '=' ;
	
LESSEQUAL2
options { paraphrase = "a '>'"; }
	: '=' '<' ;
	
GREATEREQUAL1
options { paraphrase = "a '>='"; }
	: '>' '=' ;
	
GREATEREQUAL2
options { paraphrase = "a '=>'"; }
	: '=' '>' ;
	
NOTEQUAL1
options { paraphrase = "a '!='"; }
	: '!' '=' ;
	
NOTEQUAL2
options { paraphrase = "a '<>'"; }
	: '<' '>' ;
	
NOTEQUAL3
options { paraphrase = "a '><'"; }
	: '>' '<' ;
	
NOTEQUAL4
options { paraphrase = "a '=!'"; }
	: '=' '!' ;
	
NOT
options { paraphrase = "a '!'"; }
	: '!' ;
	
AND
options { paraphrase = "a '&&'"; }
	: '&' '&' ;
		
OR
options { paraphrase = "a '||'"; }
	: '|' '|' ;
	
FIELDCONCATENATION
options { paraphrase = "a '..'"; }
	 : '.' '.' ;

LBRACK
options { paraphrase = "a '['"; }
	: '[' ;
	
RBRACK
options { paraphrase = "a ']'"; }
	: ']' ;
		
IDENT
options { paraphrase = "an identifier"; }
	: ('$'|'A'..'Z'|'a'..'z')('$'|'_'|'A'..'Z'|'a'..'z'|'0'..'9')*;
	

USERFUNCTIONCALL
options { paraphrase = "a user function call"; } 
	: ('%') ('A'..'Z'|'a'..'z'|'0'..'9')+;


MULTI_LINE_COMMENTS
options { paraphrase = "a multi-line comment"; }
	:	
		'/' '*'   
		(	{LA(2) != '/'}? '*'
		|	~('*')
		)*
		'*' '/' {_ttype = ANTLR_USE_NAMESPACE(antlr)Token::SKIP;}
	;

SINGLE_LINE_COMMENTS
options { paraphrase = "a single line comment"; }
	:	
		('/' '/' (~('\n' | '\r'))*)
			{_ttype	= ANTLR_USE_NAMESPACE(antlr)Token::SKIP;}
	;
	

REGEXP
options { paraphrase = "a regular expression"; }
	: 
		'='! '~'! (WS!)* ('m')? '/'
		( ESCAPE
		| ~('\n' | '/' | '\\' )
		)*
		'/' ('i')?
		;
	
LINECONTINUE
	:	'_'  (' ' | '\t')* (SINGLE_LINE_COMMENTS)? ( '\r''\n' | '\n')
		{
			newline();
			_ttype	= ANTLR_USE_NAMESPACE(antlr)Token::SKIP;
		}
	;


ENDLINE
options { paraphrase = "the end of the line"; }	
	:  '\n'  { newline(); };

WS
options { paraphrase = "white space"; }	
	:	( ' '
		| '\t'
        | '\r'
        )
		{_ttype	= ANTLR_USE_NAMESPACE(antlr)Token::SKIP;}
	  ;
	  
protected
EXPONENT
options { paraphrase = "an exponent"; }	
	
	:	('e'|'E'|'d'|'D') ('+'|'-')? ('0'..'9')+
	;
	
protected
INTEGER
options { paraphrase = "an integer"; }	
	:   ('0'..'9')+
    ;

protected
REAL
options { paraphrase = "a number"; }
	:   INTEGER '.' INTEGER
    ;

REAL_OR_INT
    :   ( INTEGER '.' )  => REAL (EXPONENT)? { $setType(REAL); }
    |   INTEGER  (EXPONENT)?      { $setType(INTEGER); }
    ;

STRINGLITERAL
	:	('"'!
		( ESCAPE
		| ~( '"' | '\\' )
		)*
		'"'!
		)
		
	|	('\''!
		( ESCAPE
		| ~( '\'' | '\\' )
		)*
		'\''!
		)
	;

protected
ESCAPE  
        : '\\' .
        ;
