/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef FABRIC_LIST_H
#define FABRIC_LIST_H

#include "Element.h"
#include "FabricSTL.h"

// This is the main list type used in Fabric.
// All variables and values are lists, and
// consist of elements. 

class List
{
	MA* ma;
	fvector<Element> m_elements;
public:
	List(MA* maIn)
		: ma(maIn), m_elements(maIn)
	{}

	List(MA* maIn, const Element& e)
		: ma(maIn), m_elements(maIn, e)
	{}

	virtual bool ToBool() const
	{
		fvector<Element>::const_iterator itor = m_elements.begin();
		while (itor != m_elements.end()) {
			if ((*itor)->ToBool()) {
				return true;
			}
			itor++;
		}
		return false;
	}

	void Append(const List& value)
	{
		m_elements.insert(m_elements.end(), value.m_elements.begin(), value.m_elements.end());
	}

	template<typename Vector> void Append(const Vector& vector)
	{
		m_elements.insert(m_elements.end(), vector.begin(), vector.end());
	}

	const Element& At(size_t index) const
	{
		size_t size = m_elements.size();
		if (size == 0) {
			// if empty, just get the default element
			return Element::DefaultElement;
		}
		if(index >= size) {
			// adjust the index downward
			index = size - 1;
		}
		return m_elements.at(index);
	}

	

	void Add(const Element& e)
	{
		m_elements.push_back(e);
	}

	size_t size() const
	{
		return m_elements.size();
	}

	void Reserve(size_t size)
	{
		m_elements.reserve(size);
	}

	const Element& operator [] (size_t i) const
	{
		return At(i);
	}

	bool Equals(const List& val) const
	{
		const List& a = *this;
		const List& b = val;

		if (a.size() != b.size()) {
			return false;
		}

		for (size_t i=0; i < a.size(); i++) {
			if (!a[i].Equals(b[i])) {
				return false;
			}
		}

		return true;
	}

	bool Contains(const List& sublist) const
	{		
		const List& list = *this;
		
		size_t listSize = list.size();
		size_t sublistSize = sublist.size();

		if (sublistSize == 0) {
			return true;
		}
		
		if (sublistSize > listSize) {
			return false;
		}

		size_t end = listSize - sublistSize + 1;

		for(size_t i = 0; i < end; i++) {
			size_t j;
			for(j = 0; j < sublistSize; j++) {
				if (!sublist[j].Equals(list[i+j])) {
					break;
				}
			}
			if (j == sublistSize) {
				return true;
			}
		}

		return false;
	}
};

#endif // FABRIC_LIST_H
