/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

// This is the header file for the Fabric memory allocation facility.
// Its damn simple. Its damn quick. It's damn wasteful. Oh well, you can't have everything....

#ifndef MEMORYPOOL_H
#define MEMORYPOOL_H

#include "FabricSTL.h"

#define ALIGNMENTSIZE 4
#define UPALIGN(end)  (((end + ALIGNMENTSIZE - 1) / ALIGNMENTSIZE) * ALIGNMENTSIZE)


#ifdef DEBUG_MEM
#define DEBUG_BUFFER_SIG 0xCACACACA

struct BeginSig {
	long len;
	long sig;
};
#endif


// A one way memory allocation class. Memory buffers are allocated individually
// like normal allocation, but memory is freed "all at once" when the Reset()
// method is called.
// The template parameter MinBlockSize is used to specify the allocation size of
// the internal memory blocks.


template <size_t MinBlockSize>
class AllocOnlyMemoryPool : public MA
{
	// Header struct for the internally allocated memory blocks.
	struct BlockHeader {
		char* pCurrent;		// Pointer to the current place in the block to allocate from
		char* pEnd;			// Pointer the first byte following the block.
		BlockHeader* pPreviousBlock; // The previous block
#ifdef DEBUG_MEM
		long allocations;
#endif
	};


	// The current block to sub-allocate from.
	BlockHeader* m_pCurrentBlock;	

	// default buffer. This is to avoid an unnecessary
	// memory allocation for the first block.
	char m_initialBuffer[sizeof(BlockHeader) + MinBlockSize];

public:

	AllocOnlyMemoryPool()
	{
		init();
	}

	~AllocOnlyMemoryPool()
	{
		Reset();
	}

	void* Alloc(size_t len)
	{
		ASSERT(sizeof(size_t) == sizeof(long));	
	
#ifdef DEBUG_MEM
		// account for debug signatures before and after the buffer
		BeginSig beginSig;
		beginSig.sig = DEBUG_BUFFER_SIG;
		beginSig.len = (long)len;
		long endSig = DEBUG_BUFFER_SIG;
		len += sizeof(beginSig) + sizeof(endSig);
#endif

		// we convert the unsigned length, to a signed value, so that we can do proper
		// comparisions with the buffer below
		long signedLen = (long)len;

		if (signedLen < 0) {
			//This means the len value is *huge*. Throw an error.
			throw std::bad_alloc();
		}

		// If there isn't enough room in the current block, allocate a new block.
		// (Note: we do the signed comparison here for a reason, because pCurrent
		// can actually be past pEnd, which would return a negative buffer size
		// when subtracted(which means a huge positive integer if using unsigned
		// values.)
		if (m_pCurrentBlock->pEnd - m_pCurrentBlock->pCurrent < signedLen) {

			// Allocate the new block
			size_t rawAllocLen = len + MinBlockSize + sizeof(BlockHeader);
			char* pNewBlock = new char[rawAllocLen];

			// if a alloc failure occurs above, it's supposed to throw an
			// exception, not return NULL
			ASSERT(pNewBlock != NULL);

			// the beginning of the buffer is the block header, initialize it
			BlockHeader* pHeader = (BlockHeader*)pNewBlock;
			pHeader->pEnd = pNewBlock + rawAllocLen;
			pHeader->pCurrent = pNewBlock + sizeof(BlockHeader);

			// Now set the previous value and point the current block at our
			// newly allocated block.
			pHeader->pPreviousBlock = m_pCurrentBlock;
			m_pCurrentBlock = pHeader;

#ifdef DEBUG_MEM
			m_pCurrentBlock->allocations = 0;
#endif
		}
		
		// the pointer to the beginning of the buffer we're returning
		char* pRet = m_pCurrentBlock->pCurrent;

#ifdef DEBUG_MEM
		// add signatures before and after the buffer
		memcpy(pRet, &beginSig, sizeof(beginSig));
		pRet += sizeof(beginSig);
		memset(pRet, 0xAA, beginSig.len); // shred memory
		memcpy(pRet + beginSig.len, &endSig, sizeof(endSig));
		
		m_pCurrentBlock->allocations++;
#endif

		// now increment the current buffer pointer so we know
		// where to allocate on the next call
		len = UPALIGN(len);		// Up-align the value to the necessary alignment for the next allocation
		m_pCurrentBlock->pCurrent += len;	// NOTE: this upaligned value can push the pCurrent past pEnd

		return pRet;
	}

	virtual void Free(void*)
	{
		// do nothing
	}

	void Reset()
	{
		FreeAll();
		init();
	}

private:

	void init()
	{
		// initialize the first block, which actually uses
		// the memory in m_initialBuffer
		m_pCurrentBlock = (BlockHeader*)m_initialBuffer;
		m_pCurrentBlock->pPreviousBlock = 0;
		m_pCurrentBlock->pCurrent = (char*)(m_pCurrentBlock + 1);
		m_pCurrentBlock->pEnd = m_pCurrentBlock->pCurrent + MinBlockSize;
#ifdef DEBUG_MEM
		m_pCurrentBlock->allocations = 0;
#endif
	}
	
	void FreeAll()
	{
		BlockHeader* pBlock = m_pCurrentBlock;

		// Free the blocks until we get the last block (the one that has no previous value)
		// The last block is allocated from m_initialBuffer, and cannot be freed.
		while (pBlock->pPreviousBlock != 0) {
			checkBlock(pBlock);
			BlockHeader* pPrevBlock = pBlock->pPreviousBlock;
			delete[] (char*)pBlock; // cast to original allocated type
			pBlock = pPrevBlock;
		}
		checkBlock(pBlock);
		// Reinit the block header for reuse
		pBlock->pCurrent = (char*)(pBlock + 1);
		pBlock->pEnd = pBlock->pCurrent + MinBlockSize;
	}

	void checkBlock(BlockHeader* pBlock)
	{
		
#ifdef DEBUG_MEM
		// Check all the signatures before and after the allocated
		// buffers. This is a validation that occurs to help find
		// buffer overflows and random overwrites.
		char* p = ((char*)pBlock) + sizeof(BlockHeader);
		for(long i=0; i < pBlock->allocations; i++) {			
			BeginSig beginSig;
			memcpy(&beginSig, p, sizeof(beginSig));
			p += sizeof(beginSig);
			ASSERT(beginSig.sig == DEBUG_BUFFER_SIG);
			long endSig;
			memcpy(&endSig, p + beginSig.len, sizeof(endSig));			
			ASSERT(endSig == DEBUG_BUFFER_SIG);
			p += UPALIGN(beginSig.len + sizeof(endSig));
		}
#endif
	}
};


#endif // MEMORYPOOL_H
