/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/


class Fun_error : public RuntimeNode
{	
public:
	
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(1,2);

		List* pReturn = new(ma) List(ma);
		if (frame.ArgCount() == 1) {
			const List& messages = frame.GetArg(0);
			pReturn->Reserve(messages.size());

			for (size_t i=0; i < messages.size(); i++) {
				const Element message = messages[i];
				
				if (message->GetType() == RUNTIME_ERROR) {
					pReturn->Add(message);
				}
				else {		
					const char* messageStr = message->ToText(ma).ToNullTermStr(ma);
					pReturn->Add(ErrorElement("error", messageStr));
				}
			}
		} else {
			
			const List& types = frame.GetArg(0);
			const List& messages = frame.GetArg(1);	

			size_t elements = std::max(types.size(), messages.size());
			pReturn->Reserve(elements);

			for (size_t i=0; i < elements; i++) {
				const char* typeStr = types[i]->ToText(ma).ToNullTermStr(ma);
				const char* messageStr = messages[i]->ToText(ma).ToNullTermStr(ma);
				pReturn->Add(ErrorElement(typeStr, messageStr));
			}
		}

		return pReturn;
	}
};


class Fun_repeat : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(2,2);

		const List& list = frame.GetArg(0);
		long count = frame.GetArg(1)[0]->ToLong();

		if (count < 0) {
			count = 0;
		}
		
		List* pReturn = new(ma) List(ma);
		
		pReturn->Reserve(count * list.size());

		while (count--) {
			pReturn->Append(list);
		}

		return pReturn;
	}
};


class Fun_count : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(1,1);
		return new(ma) List(ma, NumberElement(frame.GetArg(0).size()));
	}
};


class Fun_list_trim : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(1,1);

		const List& list = frame.GetArg(0);	
		size_t size = list.size();

		for (size_t i = 0; i < size; i++) {
			Element element = list[i];
			if (element->GetType() == TEXT && element->ToText(ma).Length() == 0) {
				// we've hit our first NULL text element
				List* pReturn = new(ma) List(ma);
				pReturn->Reserve(size - 1);
				// add all the previous elements to a new list
				for (size_t j = 0; j < i; j++) {
					pReturn->Add(list[j]);
				}
				
				// now we scan the remaining elements, adding in all non-null elements
				for(i++; i < size; i++) {
					element = list[i];
					if (element->GetType() == TEXT) {
						if (element->ToText(ma).Length() == 0) {
							continue; // don't add this element
						}
					}
					pReturn->Add(list[i]);
				}
				return pReturn;
			}
		}
		// no nulls found to have reached here, return the original input list
		return &list;
	}
};


class Fun_iserror : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{
		return NumberElement(element->GetType() == RUNTIME_ERROR ? 1 : 0);
	}
};

class Fun_number : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{
		return element->ToNumber();
	}
};


class Fun_docfields : public RuntimeNode
{
public:

	const List* Execute(StackFrame& frame) const
	{
		return f->GetDocument().FieldNames(ma);
	}
};

class Fun_getfield : public RuntimeNode
{
public:

	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(1,1);

		const List& names = frame.GetArg(0);

		size_t namesSize = names.size();

		if (namesSize <= 1) {
			return f->GetDocument().GetValue(ma, names[0]->ToText(ma));
		}
		else {
			Document& doc = f->GetDocument();
			List* pReturn = new(ma) List(ma);
			pReturn->Reserve(namesSize);
			for(size_t i = 0; i < namesSize; i++) {
				const List* pValue = doc.GetValue(ma, names[i]->ToText(ma));
				if (pValue) {
					pReturn->Append(*pValue);
				}
			}
			return pReturn;
		}
	}
};

