/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "Fabric.h"
#include "RuntimeTree.h"
#include "Document.h"
#include "FabricParserTokenTypes.hpp"
#include "stdio.h"
#include "ICU.h"
#include <algorithm>
#include <limits>

// RuntimeNode definitions
#include "SpecialNodes.h"
#include "OperatorNodes.h"
#include "MiscFunNodes.h"
#include "TextFunNodes.h"

using namespace antlr;
using namespace std;



RuntimeNode* Formula::NewRuntimeNode(int type, const char* pTokenStr)
{
#define NODEMATCH(type, classid) case FabricParserTokenTypes::type: p = new(&indefMa) classid(); break;
#define NODEFUNMATCH(Funname) NODEMATCH(LITERAL_##Funname, Fun_##Funname);
#define NODEFUNMATCHALIAS(Funname, AliasFunname) NODEMATCH(LITERAL_##Funname, Fun_##AliasFunname);
	
	RuntimeNode* p;
	switch (type) {
		NODEMATCH(PLUS, OpPlus);
		NODEMATCH(MINUS, OpMinus);
		NODEMATCH(IDENT, VariableReference);
		NODEMATCH(ASSIGN, VariableAssignment);
		NODEMATCH(ASSIGNINIT, VariableInit);
		NODEMATCH(REAL, NumberLiteral);
		NODEMATCH(INTEGER, NumberLiteral);
		NODEMATCH(STRINGLITERAL, StringLiteral);
		NODEMATCH(OUTERFORMULA, OuterFormula);
		NODEMATCH(USERFUNCTION, UserFunction);
		NODEMATCH(USERFUNCTIONCALL, UserFunctionCall);
		NODEMATCH(NESTEDEXPRS, NestedExprs);
		NODEMATCH(IF, If);
		NODEMATCH(COLON, OpListConcatenate);
		NODEMATCH(FORALL, ForAll);
		NODEMATCH(FORALLIN, ForAllIn);
		NODEMATCH(EQUALEQUAL, OpEqualEqual);
		NODEMATCH(LESS, OpLess);
		NODEMATCH(GREATER, OpGreater);
		NODEMATCH(CASEINSENSITIVEMATCH, OpCaseInsensitiveMatch);
		NODEMATCH(LESSEQUAL1, OpLessEqual);
		NODEMATCH(LESSEQUAL2, OpLessEqual);
		NODEMATCH(GREATEREQUAL1, OpGreaterEqual);
		NODEMATCH(GREATEREQUAL2, OpGreaterEqual);
		NODEMATCH(NOTEQUAL1, OpNotEqual);
		NODEMATCH(NOTEQUAL2, OpNotEqual);
		NODEMATCH(NOTEQUAL3, OpNotEqual);
		NODEMATCH(NOTEQUAL4, OpNotEqual);
		NODEMATCH(IN, OpIn);
		NODEMATCH(REGEXP, OpRegExp);
		NODEMATCH(NEGATE, OpNegate);
		NODEMATCH(MULTIPLY, OpMultiply);
		NODEMATCH(DIVIDE, OpDivide);
		NODEMATCH(ARRAYSUBSCRIPT, OpArraySubscript);
		NODEMATCH(FIELDCONCATENATION, OpFieldConcatenation);
		NODEMATCH(PLUSPLUS, OpPlusPlus);
		NODEMATCH(MINUSMINUS, OpMinusMinus);
		NODEMATCH(NOT, OpNot);
		NODEMATCH(AND, OpAnd);
		NODEMATCH(OR, OpOr);
		NODEMATCH(FIELDASSIGN, FieldAssignment);
		NODEMATCH(FIELDINIT, FieldInit);
		NODEMATCH(TRUELITERAL, TrueLiteral);
		NODEMATCH(FALSELITERAL, FalseLiteral);
		NODEMATCH(NILLITERAL, NilLiteral);
		NODEMATCH(COLUMN, Column);
		NODEMATCH(COLUMNASSIGN, ColumnAssignment);
		NODEMATCH(COLUMNINIT, ColumnInit);
		NODEMATCH(SELECT, SelectStatement);
		NODEFUNMATCH(assert);
		NODEFUNMATCH(assertfalse);
		NODEFUNMATCH(repeat);
		NODEFUNMATCH(return);
		NODEFUNMATCH(uppercase);
		NODEFUNMATCH(lowercase);
		NODEFUNMATCH(text);
		NODEFUNMATCH(error);
		NODEFUNMATCH(left);
		NODEFUNMATCH(right);
		NODEFUNMATCH(leftback);
		NODEFUNMATCH(rightback);
		NODEFUNMATCH(split);
		NODEFUNMATCHALIAS(explode,split);
		NODEFUNMATCH(join);
		NODEFUNMATCHALIAS(implode,join);
		NODEFUNMATCH(trim);
		NODEFUNMATCH(fulltrim);
		NODEFUNMATCH(list_trim);
		NODEFUNMATCH(count);
		NODEFUNMATCH(length);
		NODEFUNMATCH(matches);
		NODEFUNMATCH(iserror);
		NODEFUNMATCH(number);
		NODEFUNMATCH(docfields);
		NODEFUNMATCH(getfield);

		case FabricParserTokenTypes::FUNCTIONCALL:
			// This means a core function is is being called but no 
			// match for it exists.
			// We make this return a runtime error instead of compile time
			// so a future script written for a future Fabric engine
			// can still run if properly written and degrade gracefully by
			// avoiding actually making the call.

			p = new(&indefMa) UnknownFunction();
			break;

		default:
			ASSERT(0); // shouldn't ever be possible hit here
			p = new(&indefMa) UnknownFunction(); // but just in case we somehow do...
	}

	p->SetFabric(this);
	p->SetTokenText(TextElement::MakeCopy(&indefMa, pTokenStr));

	return p;
}

