/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/



class OpPlus : public RuntimeNode
{	
public:

	
	const List* Execute(StackFrame& frame) const
	{
		size_t argCount = frame.ArgCount();
		const List** args = frame.ArgArray();
		fvector<Element> processElements(ma);		
		processElements.reserve(argCount);

		size_t maxElements = 0;
		size_t nonNilArgs = 0;

		for (size_t i = 0; i < argCount; i++) {
			size_t elements = args[i]->size();
			if (elements != 0) {
				nonNilArgs++;
			}
			if (elements > maxElements) {
				maxElements = args[i]->size();
			}
		}
		
		if (nonNilArgs == 0) {
			return NULL;
		}
		else if (nonNilArgs == 1) {
			// only a single argument is nonnil, so we can't parallel process
			// find and return it
			for (size_t i = 0; i < argCount; i++) {
				if (args[i]->size() != 0) {
					return args[i];
				}
			}
			ASSERT(false); //shouldn't be able to hit here
			return NULL;
		}
		else {
			List* pReturnList = new(f->ma) List(f->ma);
			pReturnList->Reserve(maxElements);

			for (size_t i=0; i<maxElements; i++) {
				processElements.clear();

				for (size_t j = 0; j < argCount; j++) {
					if (args[j]->size()) {
						processElements.push_back(args[j]->At(i));
					}
				}
				
				try {
					pReturnList->Add( ParallelProcessElements(processElements) );
				} catch(RuntimeException& e) {
					pReturnList->Add( Element(e));
				}
			}
			
			return pReturnList;
		}
	}

	virtual const Element ParallelProcessElements(fvector<Element>& elements) const
	{
		
		fvector<Element>::iterator itor;
		bool hasText = false;
		for (itor = elements.begin(); itor != elements.end(); itor++) {
			if ((*itor)->GetType() == TEXT) {
				hasText = true;
				break;
			}
		}
		
		if (hasText) {
			size_t newLen=0;
			// convert all elements to text
			for (itor = elements.begin(); itor != elements.end(); itor++) {
				// this just returns the same TextElement if already a TextElement
				TextElement element = (*itor)->ToText(ma);
				*itor = element; // assign back to vector element

				newLen += element.Length(); // add length
			}
			
			char* pBuffer = new(ma) char[newLen];
			char* p = pBuffer;
			
			for (itor = elements.begin(); itor != elements.end(); itor++) {
				TextElement element = (*itor)->ToText(ma);
				memcpy(p, element.Ptr(), element.Length());
				p += element.Length();
			}

			ASSERT(p == (pBuffer + newLen));

			return Element(pBuffer, newLen);
		}
		else {
			double result = 0;
			// convert all elements to a number
			for (itor = elements.begin(); itor != elements.end(); itor++) {
				// this just returns the same NumberElement if already a NumberElement
				result += (*itor)->ToNumber();
			}
			return NumberElement(result);
		}	
	}
};

class OpMinus : public DualListProcess
{	
public:

	virtual const Element ProcessElements(const Element& a, const Element& b) const
	{
		double a_dbl = a->ToNumber();
		double b_dbl = b->ToNumber();
		return NumberElement(a_dbl - b_dbl);
	}

	//override to provide special different return value on nil list
	virtual const List* NilListEvent(const List& listA, const List& listB) const
	{
		
		if (listA.size() > 0) {
			return &listA;
		}
		
		return NULL;
	}
};

class OpListConcatenate : public RuntimeNode
{		
public:
	const List* Execute(StackFrame& frame) const
	{
		size_t argCount = frame.ArgCount();
		const List** args = frame.ArgArray();
		size_t newSize = 0;
		size_t nonNilLists = 0;
		for (size_t i=0; i<argCount; i++) {
			size_t argSize = args[i]->size();
			if (argSize) {
				nonNilLists++;
				newSize += argSize;
			}
		}

		if (nonNilLists == 1) {
			// only one non-nil list, find it and return it
			for (size_t i=0; i<argCount; i++) {
				if (args[i]->size()) {
					return args[i];
				}
			}
		}

		List* pNewValue = new(ma) List(ma);
		pNewValue->Reserve(newSize);

		for (size_t i=0; i<argCount; i++) {
			const List* pList = args[i];
			pNewValue->Append(*pList);
		}
		
		return pNewValue;
	}
};



class OpEqualEqual : public RuntimeNode
{		
public:

	const List* Execute(StackFrame& frame) const
	{
		const List& a = frame.GetArg(0);
		const List& b = frame.GetArg(1);
		
		return f->BoolList(a.Equals(b));
	}
};

class OpNotEqual : public RuntimeNode
{		
public:

	const List* Execute(StackFrame& frame) const
	{
		const List& a = frame.GetArg(0);
		const List& b = frame.GetArg(1);

		return f->BoolList(!a.Equals(b));
	}
};

class OpCaseInsensitiveMatch : public RuntimeNode
{		
public:

	const List* Execute(StackFrame& frame) const
	{
		const List& a = frame.GetArg(0);
		size_t sizeA = a.size();

		const List& b = frame.GetArg(1);
		size_t sizeB = b.size();

		// nil handling
		if (sizeA == 0 || sizeB == 0) {
			return f->FalseList(); // false is zero
		}
		for (size_t i=0; i < sizeA; i++) {
			const Element elementA = a[i];
			ElementType typeA = elementA->GetType();
			for (size_t j=0; j < sizeB; j++) {
				const Element elementB = b[j];
				ElementType typeB = elementB->GetType();
				if (typeA == typeB) {
					if (elementA.EqualsI(elementB)) {
						return new(ma) List(ma, NumberElement(1.0 + i));
					}
				}
				else {
					try {
						if (elementB.EqualsI(elementA->ConvertTo(ma, typeB))) {
							return new(ma) List(ma, NumberElement(1.0 + i));
						}
					}
					catch (Element::ConversionException& ) {
						//ignore conversion exceptions
					}
				}
				
			}
		}
		// no matches
		return f->FalseList();
	}
};

// this abstract class is for operators which take
// two lists and pairwise compares them and return
// a single boolean
class BinaryComparisonOperator : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		const List& a = frame.GetArg(0);
		const List& b = frame.GetArg(1);

		size_t maxElements = std::min(a.size(), b.size());

		int result = 0;
		for (size_t i=0; i < maxElements; i++) {
			result = Element::Collate(a[i], b[i]);
			if (result != 0) {
				break;
			}
		}

		if (result == 0) {
			result = cmp(a.size(), b.size());
		}

		if (ToBool(result)) {
			return f->TrueList();
		}
		else {
			return f->FalseList();
		}
	}

	virtual bool ToBool(int result) const = 0;
};


class OpLess : public BinaryComparisonOperator
{		
public:
	bool ToBool(int result) const
	{		
		return result < 0;
	}
};

class OpLessEqual : public BinaryComparisonOperator
{		
public:
	bool ToBool(int result) const
	{		
		return result <= 0;
	}
};

class OpGreater : public BinaryComparisonOperator
{		
public:
	bool ToBool(int result) const
	{		
		return result > 0;
	}
};

class OpGreaterEqual : public BinaryComparisonOperator
{		
public:
	bool ToBool(int result) const
	{		
		return result >= 0;
	}
};

class OpNegate : public SingleListProcess
{
public:
	const Element ProcessElement(const Element& element) const
	{
		double anumber = element->ToNumber();
		return NumberElement(-anumber);
	}
};

class OpNot : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		const List& list = frame.GetArg(0);

		if (list.ToBool()) {
			return f->FalseList();
		}
		else {
			return f->TrueList();
		}
	}	
};


class OpMultiply : public DualListProcess
{	
public:

	virtual const Element ProcessElements(const Element& a, const Element& b) const
	{
		double a_dbl = a->ToNumber();
		double b_dbl = b->ToNumber();
		return NumberElement(a_dbl * b_dbl);
	}

	
	virtual const List* NilListEvent(const List& listA, const List& listB) const
	{
		if (listA.size() > 0) {
			return &listA;
		}
		else {		
			return &listB;
		}
	}

};

class OpDivide : public DualListProcess
{	
public:

	virtual const Element ProcessElements(const Element& a, const Element& b) const
	{
		double a_dbl = a->ToNumber();
		double b_dbl = b->ToNumber();
		if (b_dbl != 0) {
			return NumberElement(a_dbl / b_dbl);
		}
		else {
			return NumberElement(std::numeric_limits<float>::quiet_NaN());
		}
	}
	
	virtual const List* NilListEvent(const List& listA, const List& listB) const
	{
		return &listA;
	}

};



class OpArraySubscript : public RuntimeNode
{

	// convert an 1 based signed index to positive, 0 based index
	long FixIndex(long index, const List& list) const
	{
		if (abs(index) > 0 && abs(index) <= (long)list.size()) {
			if (index > 0) {
				return index - 1;
			}
			else  {
				return (long)list.size() + index;
			}
		}
		else {
			return -1;
		}
	}

	const List* Execute(StackFrame& frame) const
	{
		const List& list = frame.GetArg(0);

		if (list.size() == 0) {
			// if nil list, all index operations result in nil regardless
			return &list;
		}

		if (frame.ArgCount() == 2) {
			const List& indexList = frame.GetArg(1);

			List* pReturnList = new(ma) List(ma);
			pReturnList->Reserve(indexList.size());

			for (size_t i=0; i < indexList.size(); i++)  {
				try {
					long index = FixIndex(indexList[i]->ToLong(), list);
					if (index != -1) {
						pReturnList->Add(list[index]);
					}
				} catch (RuntimeException& e) {
					pReturnList->Add(e);
				}
			}

			return pReturnList;
		}
		else {
			ASSERT(frame.ArgCount() == 3);			

			// we only use the first elements of the 2 start and stop indexes
			long startIndex = frame.GetArg(1)[0]->ToLong();
			long endIndex = frame.GetArg(2)[0]->ToLong();
			
			if (startIndex <= 0) {
				startIndex = 1;
			}

			if (endIndex < startIndex) {
				return NULL;
			}

			startIndex = FixIndex(startIndex, list);

			if (startIndex == -1 ) {
				// too big
				return NULL;
			}
			
			endIndex = FixIndex(endIndex, list);

			if (endIndex == -1) {
				// too big, shrink to fit
				endIndex = (long)list.size() - 1;
			}

			List* pReturnList = new(ma) List(ma);

			if (startIndex <= endIndex) {
				pReturnList->Reserve(endIndex - startIndex + 1);

				for (long i = startIndex; i <= endIndex; i++) {
					pReturnList->Add(list[i]);
				}
			}

			return pReturnList;
		}
	}
};

class OpRegExp : public RuntimeNode, Destructable
{
public:
	URegularExpression* re;

	void CompilerValidateArgsAndInit()
	{
		std::vector<UChar> patternUni;
		TextElement patternText = GetTokenText();
		const char* pPattern = patternText.Ptr();
		int patternLen = patternText.LenInt();
		int regexpFlags = 0;

		ASSERT(*pPattern == '/'); // must start with a slash
		
		pPattern++;
		patternLen--;

		if (pPattern[patternLen - 1] == '/') {
			patternLen -= 1;
		}
		else if (pPattern[patternLen - 1] == 'i') {
			ASSERT(pPattern[patternLen - 2] == '/');
			regexpFlags = UREGEX_CASE_INSENSITIVE;
			patternLen -= 2;
		}
		else {
			// the token is supposed to end in either "/" or "/i"
			ASSERT(FALSE);
		}
		
		utf8_convert(pPattern, patternLen, patternUni);

		UParseError pe;
		UErrorCode status = U_ZERO_ERROR;
		re = uregex_open(&patternUni[0], (int)patternUni.size(), regexpFlags, &pe, &status);
		if (U_FAILURE(status)) {
			char* pErrMrg = new(ma) char[200];
			snprintf(pErrMrg, 200, "Pattern error at position %d: %s \n",
				pe.offset, u_errorName(status));
			throw Formula::SyntaxException(pErrMrg);
		}

		f->AddToDestructionQueue(this); // this causes out destruct method to be called
										//once the whole formula is destroyed
	}
	
	const List* Execute(StackFrame& frame) const
	{
		const List& srcList = frame.GetArg(0);
		size_t srcListSize = srcList.size();
		if (srcListSize == 0) {
			return f->FalseList();
		}
		for (size_t i=0; i<srcListSize; i++) {
			TextElement src = srcList[i]->ToText(ma);
			std::vector<UChar> srcUni;
			utf8_convert(src, srcUni);
			UErrorCode status = U_ZERO_ERROR;

			uregex_setText(re, &srcUni[0], (int)srcUni.size(), &status);
			UBool didMatch =  uregex_matches(re, 0, &status);
			CHECK_STATUS(status);

			if (didMatch) {				
				// return the 1 based offest as the result
				return new(ma) List(ma, NumberElement(1.0 + i));
			}
		}
		return f->FalseList();
	}

	void Destruct()
	{		
		uregex_close(re);
	}
};

class OpFieldConcatenation : public RuntimeNode
{
public:
	TextElement m_id;
	long m_firstFieldIndex;

	void CompilerValidateArgsAndInit()
	{
		TextElement id = m_pFirstChild->GetTokenText();

		m_id = id.ExtractTrailingNumber(&m_firstFieldIndex);
		if (m_firstFieldIndex < 0) {
			throw Formula::SyntaxException("Identifier must end in digit (0-9)");
		}
	}

	
	virtual const RuntimeNode* FirstNodeToExecute(StackFrame& frame) const
	{
		// we never the execute the first arg
		
		if (m_numChildren == 1) {
			return NULL; // its the * node
		}
		else {
			return m_pFirstChild->GetNextSibling();
		}

	}

	const List* Execute(StackFrame& frame) const
	{		
		if (m_numChildren == 1) {
			return f->GetDocument().CombineFields(ma, m_id, m_firstFieldIndex, -1);
		}
		else {
			long lastFieldIndex = frame.GetArg(0)[0]->ToLong();

			if (lastFieldIndex < 0 || lastFieldIndex < m_firstFieldIndex) {
				return NULL;
			}

			return f->GetDocument().CombineFields(ma, m_id, m_firstFieldIndex, lastFieldIndex);
		}
	}
};


class OpPlusPlus : public SingleListProcess
{
public:

	const List* Execute(StackFrame& frame) const
	{		
		ASSERT(frame.ArgCount() == 1);

		const List* pReturnList = SingleListProcess::Execute(frame);

		// Now assign pReturnList value back to the variable from where it came
		
		f->SetVariableValue( m_pFirstChild->GetTokenIdAtom(), pReturnList);

		return pReturnList;
	}

	const Element ProcessElement(const Element& element) const
	{
		return NumberElement(element->ToNumber() + 1.0);
	}
};


class OpMinusMinus : public OpPlusPlus
{
public:
	// override OpPlusPlus's ProcessElement method
	const Element ProcessElement(const Element& element) const
	{
		return NumberElement(element->ToNumber() - 1.0);
	}
};



class OpAnd : public RuntimeNode
{
	public:

	const RuntimeNode* NextNodeToExecute(StackFrame& frame, const RuntimeNode* pPreviousExecuteNode, const List* pPreviousValue) const
	{
		if (pPreviousValue->ToBool()) {
			if (pPreviousExecuteNode == m_pFirstChild) {
				// this is the first arg. Tell it to execute the second arg
				return m_pFirstChild->GetNextSibling();
			}
			else {
				// the is the second arg. its true. flag that we hit true and the 
				frame.boolVal = true;
				return NULL;
			}
		}
		else {
			frame.boolVal = false;
			return NULL;
		}
	}

	const List* Execute(StackFrame& frame) const
	{		
		return (frame.boolVal ? f->TrueList() : f->FalseList());
	}
};

class OpOr : public RuntimeNode
{
public:
	const RuntimeNode* NextNodeToExecute(StackFrame& frame, const RuntimeNode* pPreviousExecuteNode, const List* pPreviousValue) const
	{
		if (pPreviousValue->ToBool()) {
			frame.boolVal = true;
			return NULL;
		}
		else {
			frame.boolVal = false;
			return pPreviousExecuteNode->GetNextSibling();
		}
	}

	const List* Execute(StackFrame& frame) const
	{		
		return (frame.boolVal ? f->TrueList() : f->FalseList());
	}
};

class OpIn : public RuntimeNode
{
public:

	const List* Execute(StackFrame& frame) const
	{	
		const List& sublist = frame.GetArg(0);
		const List& list = frame.GetArg(1);
		
		return (list.Contains(sublist) ? f->TrueList() : f->FalseList());
	}
};


