/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#ifndef RUNTIMETREE_H
#define RUNTIMETREE_H

#include <string>
#include "List.h"
#include "FabricSTL.h"

// This file contains the core definitions used by runtime trees


enum NodeType {
	NodeType_NotSpecified,
	NodeType_VariableReferenceNode,
	NodeType_ForAllInNode,
	NodeType_OpMultiply,
	NodeType_UserFunctionCall,
	NodeType_ColumnAssignment
};

// This is the stackframe used by each function during runtime.
class StackFrame
{
	friend class Formula;
	friend class RuntimeStack;

	const RuntimeNode* pNode;

	fvector<const List*>* m_pArgsStack; // a pointer to the parent runtime argument stack
	size_t	m_argsStored; // the number of arguments we are storing in the stack
public:

	StackFrame(fvector<const List*>& argsStack)
		: m_pArgsStack(&argsStack), pNode(NULL), m_argsStored(0)
	{
	}

	// get a pointer the beginning of our arguments
	const List** ArgArray()
	{
		return &m_pArgsStack->at(m_pArgsStack->size() - m_argsStored);
	}

	// how many arguments
	size_t ArgCount()
	{
		return m_argsStored;
	}

	// The out nth argument. Unchecked
	const List& GetArg(size_t n)
	{
		ASSERT(n < m_argsStored);
		return *m_pArgsStack->at(m_pArgsStack->size() - m_argsStored + n);
	}

	const List& LastArg()
	{
		ASSERT(m_argsStored > 0);
		return *m_pArgsStack->back();
	}

	void PushArg(const List* pList)
	{
		m_pArgsStack->push_back(pList);
		m_argsStored++;
	}

	const List* PopArg()
	{
		ASSERT(m_argsStored > 0);
		const List* pReturn = m_pArgsStack->back();
		m_pArgsStack->pop_back();
		m_argsStored--;
		return pReturn;
	}

	void ClearArgs()
	{
		m_pArgsStack->resize(m_pArgsStack->size() - m_argsStored);
		m_argsStored = 0;
	}

	// this is the public union, initialized and used
	// by any node that needs it. You can add new items
	// to it as needed.
	union {
		struct {
			List* pReturnList;
			size_t currentIteration;
			size_t totalIterations;
		} forAllLoopData;
		bool boolVal;
		const List* pReturnList;
	};
};

class RuntimeNode
{
	friend class Formula;

protected:
	Formula*		f;
	MA*				ma;
	RuntimeNode*	m_pFirstChild;
	RuntimeNode*	m_pNextSibling;
	int				m_numChildren;
	TextElement		m_tokenText;
	IdAtom			m_tokenIdAtom;

public:

	RuntimeNode()
		: m_pFirstChild(NULL), m_pNextSibling(NULL), m_tokenIdAtom(0), m_numChildren(0)
	{
	}

	void ThrowArgException(int minArgsAllowed, int maxArgsAllowed) const
	{	
		StringBuilder str(ma,100);
		str += "Incorrect number of arguments for function \"";
		str += GetTokenText();
		str += "\". ";

		if (minArgsAllowed != maxArgsAllowed) {
			str += " It requires ";
			str += minArgsAllowed;
			str += " to ";
			str += maxArgsAllowed;
			str += " arguments, but ";
		}
		else {
			str += " It requires ";
			str += minArgsAllowed;
			str += " arguments, but ";					
		}
		str += m_numChildren;
		str += "were passed.";

		throw RuntimeException("InvalidArguments", str);
	}

	inline void CheckArgCount(int minArgsAllowed, int maxArgsAllowed) const
	{
		if (m_numChildren < minArgsAllowed || m_numChildren > maxArgsAllowed) {
			ThrowArgException(minArgsAllowed, maxArgsAllowed);
		}
	}

	virtual NodeType GetNodeType() const
	{
		return NodeType_NotSpecified;
	}

	void SetFabric(Formula* f)
	{
		this->f = f;
		this->ma = &f->transMa;
	}

	virtual const RuntimeNode* FirstNodeToExecute(StackFrame& frame) const
	{
		return m_pFirstChild;
	}

	virtual const RuntimeNode* NextNodeToExecute(StackFrame& frame, const RuntimeNode* pPreviousExecuteNode, const List* pPreviousList) const
	{
		frame.PushArg(pPreviousList);
		return pPreviousExecuteNode->GetNextSibling();
	}

	virtual const List* Execute(StackFrame& frame) const = 0;

	const RuntimeNode* GetFirstChild() const
	{
		return m_pFirstChild;
	}

	const RuntimeNode* GetNextSibling() const
	{
		return m_pNextSibling;
	}
	
	void SetFirstChild(RuntimeNode* pNode)
	{
		ASSERT(m_pFirstChild == NULL);
		m_pFirstChild = pNode;
	}
	
	void SetNextSibling(RuntimeNode* pNode)
	{
		ASSERT(m_pNextSibling == NULL);
		m_pNextSibling = pNode;
	}

	void SetTokenText(const TextElement& text)
	{
		m_tokenText = text;
	}

	const TextElement& GetTokenText() const
	{
		return m_tokenText;
	}

	IdAtom GetTokenIdAtom() const
	{
		if (m_tokenIdAtom == 0) {
			RuntimeNode* pThis = const_cast<RuntimeNode*>(this);
			pThis->m_tokenIdAtom = f->GetIdAtom(m_tokenText);
			return pThis->m_tokenIdAtom;
		}

		return m_tokenIdAtom;
	}

	virtual void CompilerValidateArgsAndInit()
	{
	}
};


class SingleListProcess : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(1,1);

		const List& list = frame.GetArg(0);
		List* pReturnList = new(ma) List(ma);
		size_t elements = list.size();

		pReturnList->Reserve(elements);

		for (size_t i = 0; i < elements; i++) {
			try {
				pReturnList->Add(ProcessElement(list[i]));
			} catch(RuntimeException& e) {
				pReturnList->Add(Element(e));
			}
		}
		
		return pReturnList;
	}

	virtual const Element ProcessElement(const Element& element) const = 0;
};


class DualListProcess : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(2,2);

		const List& listA = frame.GetArg(0);
		size_t sizeA = listA.size();

		const List& listB = frame.GetArg(1);
		size_t sizeB = listB.size();

		if (sizeA == 0 || sizeB == 0 ){
			return NilListEvent(listA, listB);
		}

		List* pReturnList = new(ma) List(ma);
		size_t elements = std::max(sizeA, sizeB);
		pReturnList->Reserve(elements);

		for (size_t i = 0; i < elements; i++) {
			try {
				pReturnList->Add(ProcessElements(listA[i], listB[i]));
			} catch(RuntimeException& e) {
				pReturnList->Add(Element(e));
			}
		}
		
		return pReturnList;
	}

	virtual const Element ProcessElements(const Element& a, const Element& b) const = 0;

	//override to provide special different return value on nil list
	virtual const List* NilListEvent(const List& listA, const List& listB) const
	{
		// if either is nil
		// then return nill
		return f->NilList();
	}
};



#endif // RUNTIMETREE_H
