/*
Fabric formula engine
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/


extern UCaseMap* g_caseMap;

class Fun_text : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{
		return element->ToText(ma);
	}

};


class Fun_lowercase : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{
		UErrorCode status = U_ZERO_ERROR;

		TextElement text = element->ToText(ma);
		char* pNew = new(ma) char[text.Length()];
		int newLen =
			ucasemap_utf8ToLower(g_caseMap,
				pNew, text.LenInt(),
				text.Ptr(), text.LenInt(),
				&status);

		// when newLen is bigger than the input, that means it needed more room
		if (newLen > text.LenInt()) {			
			pNew = new(ma) char[newLen];
			newLen =
				ucasemap_utf8ToLower(g_caseMap,
					pNew, newLen,
					text.Ptr(),  text.LenInt(),
					&status);
		}
		
		CHECK_STATUS(status);

		return TextElement(pNew, newLen);
	}
};

class Fun_uppercase : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{
		UErrorCode status = U_ZERO_ERROR;

		TextElement text = element->ToText(ma);
		char* pNew = new(ma) char[text.Length()];
		int newLen =
			ucasemap_utf8ToUpper( g_caseMap,
				pNew, text.LenInt(),
				text.Ptr(), text.LenInt(),
				&status);

		// when newLen is bigger than the input, that means it needed more room
		if (newLen > text.LenInt()) {			
			pNew = new(ma) char[newLen];
			newLen =
				ucasemap_utf8ToUpper( g_caseMap,
					pNew, newLen,
					text.Ptr(),  text.LenInt(),
					&status);
		}
		
		CHECK_STATUS(status);

		return TextElement(pNew, newLen);
	}
};


class Fun_left : public DualListProcess
{
public:
	virtual const Element ProcessElements(const Element& in, const Element& arg) const
	{
		const TextElement text = in->ToText(ma);
		if (arg->GetType() == NUMBER) {
			// left(str, n)
			// return the first n characters
			long count = arg->ToLong();
			
			int i=0;
			const char* p = text.Ptr();
			int len = text.LenInt();
			while (count-- > 0 && i < len) {
				UChar32 c;
				U8_NEXT(p, i, len, c);
			}
			return TextElement(p, std::min(len, i));
		}
		else {
			// left(str, substr)
			// return everything to the left of substring
			const TextElement substr = arg->ToText(ma);
			const char* pSubstr = substr.Ptr();
			int substrLen = substr.LenInt();

			if (substrLen == 0) {
				// when substring is null, return the whole thing
				return text;
			}

			int i=0;
			const char* p = text.Ptr();

			// set the endpoint of where to quit scanning.
			int end = text.LenInt() - substrLen + 1;

			// advancing character by one, check for the substring
			while (i < end) {
				if (memcmp(p + i, pSubstr, substrLen) == 0) {
					return TextElement(p, i);
				}
				UChar32 buff;
				U8_NEXT(p, i, end, buff);
			}
			return text;
		}
	}
};

class Fun_right : public DualListProcess
{
public:
	virtual const Element ProcessElements(const Element& in, const Element& arg) const
	{		
		const TextElement text = in->ToText(ma);
		if (arg->GetType() == NUMBER) {
			// right(str, n)
			// Returns the last n characters in a string

			// get the total number of characters
			int characters = text.Characters();
			long count = arg->ToLong();
			
			count = characters - count; // this is how many to advance
			
			int i=0;
			const char* p = text.Ptr();
			int len = text.LenInt();
			while (count-- > 0 && i < len) {
				UChar32 c;
				U8_NEXT(p, i, len, c);
			}
			if (i > len) {
				i = len;
			}
			return TextElement(p + i, len - i);
		}
		else {
			// right(str, substr)
			// return everything to the right of substring
			const TextElement substr = arg->ToText(ma);
			const char* pSubstr = substr.Ptr();
			int substrLen = substr.LenInt();

			if (substrLen == 0) {
				// when substring is null, return the whole thing
				return text;
			}

			int i=0;
			const char* p = text.Ptr();

			// set the endpoint of where to quit scanning.
			int end = text.LenInt() - substrLen + 1;

			// advancing character by one, check for the substring
			while (i < end) {
				if (memcmp(p + i, pSubstr, substrLen) == 0) {
					// found a match, return everything to right of it
					return TextElement(p + i + substrLen, text.Length() - (i + substrLen));
				}
				UChar32 c;
				U8_NEXT(p, i, end, c);
			}
			return text;
		}
	}
};


class Fun_leftback : public DualListProcess
{
public:
	virtual const Element ProcessElements(const Element& in, const Element& arg) const
	{
		// leftback(str, substr)
		// return everything to the left of the last occurance of substr.
		const TextElement text = in->ToText(ma);
		const TextElement substr = arg->ToText(ma);
		const char* pSubstr = substr.Ptr();
		int substrLen = substr.LenInt();

		if (substrLen == 0) {
			// when substring is null, return the whole thing
			return text;
		}

		int i=0;
		const char* p = text.Ptr();
		
		// set the endpoint of where to quit scanning.
		int end = text.LenInt() - substrLen + 1;
		
		// keep track of the last substring match
		int lastMatch = text.LenInt();

		while (i < end) {
			if (memcmp(p + i, pSubstr, substrLen) == 0) {
				lastMatch = i;
				i += substrLen;
			}
			else {
				UChar32 buff;
				U8_NEXT(p, i, end, buff);
			}
		}
		return TextElement(p, lastMatch);
	}
};


class Fun_rightback : public DualListProcess
{
public:
	virtual const Element ProcessElements(const Element& in, const Element& arg) const
	{
		// rightback(str, substr)
		// return everything to the right of the last occurance of substr.
		const TextElement text = in->ToText(ma);
		const TextElement substr = arg->ToText(ma);
		const char* pSubstr = substr.Ptr();
		int substrLen = substr.LenInt();

		if (substrLen == 0) {
			// when substring is null, return the whole thing
			return text;
		}

		int i=0;
		const char* p = text.Ptr();

		// set the endpoint of where to quit scanning.
		int end = text.LenInt() - substrLen + 1;
		
		int lastMatch = 0;

		while (i < end) {
			if (memcmp(p + i, pSubstr, substrLen) == 0) {
				lastMatch = i + substrLen;
				i += substrLen;
			}
			else {
				UChar32 buff;
				U8_NEXT(p, i, end, buff);
			}
		}
		return TextElement(p + lastMatch, text.LenInt() - lastMatch);
	}
};


class Fun_split : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(2,2);

		const List& strs = frame.GetArg(0);
		size_t strElements = strs.size();

		const List& substrs = frame.GetArg(1);
		size_t substrElements = substrs.size();

		if (substrElements == 0) {
			return &strs;
		}

		List* pReturnList = new(ma) List(ma);
		pReturnList->Reserve(strElements);

		// loop through all the strings in the first list
		for (size_t i = 0; i < strElements; i++) {
			const TextElement text = strs[i]->ToText(ma);

			int str_index = 0; //current string index

			const char* pStr = text.Ptr();
			int str_len = text.LenInt();
			
			int lastMatch = 0; // indicates the last index past where we matched a delim

			while (str_index < str_len) {
				size_t j;
				for (j = 0; j < substrElements; j++) {
					const TextElement substr = substrs[j]->ToText(ma);
					const char* pSubstr = substr.Ptr();
					int substrLen = substr.LenInt();
					if (substrLen == 0 || str_index + substrLen > str_len ) {
						// null delim or delim too long
						continue;
					}
					if (memcmp(pStr + str_index, pSubstr, substrLen) == 0) {
						// got a match, add the substring since lastmatch to the results
						pReturnList->Add(TextElement(pStr + lastMatch, str_index - lastMatch));
						lastMatch = str_index + substrLen;
						str_index += substrLen; // advance the pointer
						break;
					}
				}
				if (j == substrElements) { // we didn't match anything, advance index
					UChar32 c;
					U8_NEXT(pStr, str_index, str_len, c);
				}
			}
			pReturnList->Add(TextElement(pStr + lastMatch, str_index - lastMatch));
		}		
		return pReturnList;
	}
};


class Fun_join : public RuntimeNode
{
public:
	const List* Execute(StackFrame& frame) const
	{
		CheckArgCount(1,2);

		const List& list = frame.GetArg(0);

		if (list.size() <= 1) {
			// one or less elements, nothing to join
			return &list;
		}

		// if no delim argument, so we use a null string ""
		const List& delims = (frame.ArgCount() == 2 ? frame.GetArg(1) : *f->NilList());

		size_t listCount = list.size();
		size_t buffersize = 0;

		ASSERT(listCount > 0);
		for (size_t i = 0; i < listCount - 1; i++) {
			// figure out how big the new buffer needs to be
			buffersize += list[i]->ToText(ma).Length() + delims[i]->ToText(ma).Length();
		}
		buffersize += list[listCount]->ToText(ma).Length();

		// now join all the strings together into the new buffer
		char* pNew = new(ma) char[buffersize];
		size_t insert=0;
		for (size_t i = 0; i < listCount - 1; i++) {
			const TextElement value = list[i]->ToText(ma);
			memcpy(pNew + insert, value.Ptr(), value.Length());
			insert += value.Length();
			const TextElement delim = delims[i]->ToText(ma);
			memcpy(pNew + insert, delim.Ptr(), delim.Length());
			insert += delim.Length();
		}
		
		TextElement value = list[listCount]->ToText(ma);

		memcpy(pNew + insert, value.Ptr(), value.Length());

		return new(ma) List(ma,TextElement(pNew, buffersize));
	}
};

// returns the length of each element as a string (even if its not a string)
class Fun_length : public SingleListProcess
{
	virtual const Element ProcessElement(const Element& element) const
	{
		return NumberElement(element->ToText(ma).Characters());
	}
};

const TextElement Trim(const TextElement text)
{
	int i=0;
	const char* p = text.Ptr();
	int len = text.LenInt();
	int firstNonWhite = 0;
	int afterLastNonWhite = 0;
	
	// scan the string until we hit the first non white character
	while (i < len) {
		// hit it, we've already recorded it
		// now set lastNonWhite to i,
		// we will contine in the loop below
		UChar32 c;
		firstNonWhite = i;
		U8_NEXT(p, i, len, c);
		if (!u_isUWhiteSpace(c)) {
			afterLastNonWhite = i;
			break;
		}
	}
	// scan the string until done, recording the last non-white space character
	while (i < len) {
		UChar32 c;
		U8_NEXT(p, i, len, c);
		if (!u_isUWhiteSpace(c)) {
			afterLastNonWhite = i;
		}
	}

	return TextElement(p + firstNonWhite, (afterLastNonWhite - firstNonWhite));
}

class Fun_trim : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{			
		return Trim(element->ToText(ma));
	}
};


class Fun_fulltrim : public SingleListProcess
{
public:
	virtual const Element ProcessElement(const Element& element) const
	{			
		int i=0;
		const TextElement text = Trim(element->ToText(ma));
		const char* p = text.Ptr();
		int len = text.LenInt();

		bool wasPrevWhite = false;
		while (i < len) {
			UChar32 c;
			U8_NEXT(p, i, len, c);
			bool isThisWhite = (u_isUWhiteSpace(c) != 0);

			if (wasPrevWhite && isThisWhite) {
				// we've a internal duplicate, which means we now
				// need to make a new string (as opposed to just a substring)
				int newLen = len - sizeof(char);
				char* pNew = new(ma) char[newLen];
				// copy in all the chars up til now
				// i indexes the character after the second whitespace character
				memcpy(pNew, p, i - sizeof(char));
				int insertionPoint =  i - sizeof(char);
				while (i < len) {
					U8_NEXT(p, i, len, c);
					isThisWhite = (u_isUWhiteSpace(c) != 0);
					if (wasPrevWhite && isThisWhite) {
						continue; // skip
					}
					UBool isError = FALSE;
					U8_APPEND((unsigned char*)pNew, insertionPoint, newLen, c, isError);
					wasPrevWhite = isThisWhite;
				}

				return TextElement(pNew, insertionPoint);

			}
			wasPrevWhite = isThisWhite;
		}

		// no duplicates found
		return text;
	}
};


class Fun_matches : public DualListProcess
{
public:
	virtual const Element ProcessElements(const Element& src, const Element& pattern) const
	{
		TextElement srcText = src->ToText(ma);
		TextElement patternText = pattern->ToText(ma);

		std::vector<UChar> srcUni;
		std::vector<UChar> patternUni;

		utf8_convert(srcText, srcUni);
		utf8_convert(patternText, patternUni);

		UParseError pe;
		UErrorCode status = U_ZERO_ERROR;
		URegularExpression* re = uregex_open(&patternUni[0], (int)patternUni.size(), 0, &pe, &status);
		if (U_FAILURE(status)) {
			char* pErrMrg = new(ma) char[200];
			snprintf(pErrMrg, 200, "Pattern error at position %d: %s \n",
				pe.offset, u_errorName(status));
			throw RuntimeException("regex", pErrMrg);
		}
		uregex_setText(re, &srcUni[0], (int)srcUni.size(), &status);
		UBool didMatch =  uregex_matches(re, 0, &status);
		uregex_close(re);
		CHECK_STATUS(status);

		return NumberElement(didMatch ? 1 : 0);
	}
};
