#ifndef INC_UnicodeCharBuffer_hpp__
#define INC_UnicodeCharBuffer_hpp__

#include <istream>
#include <cassert>
#include <antlr/config.hpp>
#include <antlr/InputBuffer.hpp>
#include <antlr/CharStreamIOException.hpp>

class ANTLR_API UnicodeCharBuffer : public antlr::InputBuffer {
public:
	typedef unsigned int char_type;	// should be 32 bits!

	/// Create a character buffer
	UnicodeCharBuffer(const char* pIn, size_t len)
		: p(pIn), pEnd(pIn + len)
	{
	}

	/// Get the next character from the stream
	int getChar()
	{
		
		if (p == pEnd)  {
			// this makes it like every stream has a newline at the end
			p++;
			return '\n';
		}
		else if (p > pEnd)  {
			return -1;
		}

		char_type ch = 0;

		int inchar = *p++;

// This is how UTF8 is encoded
// +---------------------------+----------+----------+----------+----------+
// | Unicode scalar            | 1st      | 2nd      | 3th      | 4th      |
// +---------------------------+----------+----------+----------+----------+
// |00000000 0xxxxxxx          | 0xxxxxxx |          |          |          |
// |00000yyy yyxxxxxx          | 110yyyyy | 10xxxxxx |          |          |
// |zzzzyyyy yyxxxxxx          | 1110zzzz | 10yyyyyy | 10xxxxxx |          |
// |000uuuuu zzzzyyyy yyxxxxxx | 11110uuu | 10uuzzzz | 10yyyyyy | 10xxxxxx |
// +---------------------------+----------+----------+----------+----------+

		if( (inchar & 0x80) == 0 )
			return inchar;

		unsigned int need = 0;
		if( (inchar & 0xF8) == 0xF8 )
		{
			ch = inchar & 7;
			need = 3;
		}
		else if( (inchar & 0xE0) == 0xE0 )
		{
			ch = inchar & 0xF;
			need = 2;
		}
		else if( (inchar & 0xC0) == 0xC0 )
		{
			ch = inchar & 0x1F;
			need = 1;
		}
		else
		{
			std::runtime_error e("Malformed UTF8");
			throw antlr::CharStreamIOException(e);
		}

		while( need )
		{
			if( p >= pEnd ) {
				std::runtime_error e("Malformed UTF8");
				throw antlr::CharStreamIOException(e);
			}
			inchar = *p++;
			ch <<= 6;
			ch += inchar & 0x3F;
			need--;
		}
		return ch;
	}
private:

	const char* p;
	const char* pEnd;

	// NOTE: Unimplemented
	UnicodeCharBuffer(const UnicodeCharBuffer& other);
	UnicodeCharBuffer& operator=(const UnicodeCharBuffer& other);
};

#endif //INC_UnicodeCharBuffer_hpp__
