/*
Fabric formula engine server. For use with Erlang and CouchDb
Copyright (C) 2006  Damien Katz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

// This file is an executable that reads requests in binary format from stdin
// (fabric formulas, documents, etc) and performs the requested processing and 
// outputs the binary results to stdout. The protocol it uses is adhoc
// and not meant to be a documented generic inteface.

#include <stdio.h>
#include <vector>
#include <string>
#include "fcntl.h"

#ifdef WIN32

#include "Windows.h"
#undef min
#include "io.h"

#else

#include <unistd.h>

typedef unsigned char BYTE;
typedef unsigned int DWORD;

#endif

using namespace std;

#include "Fabric.h"
#include "Element.h"
#include "List.h"
#include "Document.h"


// forward declarations
bool setup_io();

DWORD endian_swap(DWORD nLongNumber);

void read_exact(void *buf, DWORD len);
void write_exact(const void *buf, DWORD len);

void send_packet(const void *buf, DWORD len);
void get_packet(void *buf, DWORD len);
DWORD packet_len();

DWORD get_dword();
void send_dword(DWORD dword);
string get_string();
char* get_str();
void send_string(const TextElement& text);
void send_string(const char* buffer, size_t len);
void get_list(List& list);
void send_list(const List& list);
double get_double();
void send_double(double dbl);
void get_doc(Document& document);
void free_formulas(vector<Formula*> formulas);

class io_exception : public runtime_error
{
public:
	io_exception() : runtime_error("IO error")
	{}
};

class closed_pipe_exception : public runtime_error
{
public:
	closed_pipe_exception()  : runtime_error("Pipe closed")
	{}
};



// global var that hold the duplicated stdout port id
int stdout_dup;

// global memory allocator
AllocOnlyMemoryPool<0xFFFF>	maPool;
MA*	ma;


int main(int argc, char* argv[])
{
	if (argc < 2 || strcmp(argv[1], "launched_from_erlang")) {
		// this isn't in any way a generic utility
		printf("Sorry, %s, should only be used from the couch_fabric Erlang module.", argv[0]);
		return 1;
	}

	//DebugBreak();

	if (!setup_io()) {
		return 1;
	}

	ma = &maPool;

	vector<Formula*> formulas;

	try {
		// initalize the formula library (once per process)
		Formula::ProcessInit();

		while (true) {
			maPool.Reset(); // reset memory pool, brand new request

			// get_dword waits until a result is returned or throws a closed_pipe_exception
			// This value is a integer that indicates the operation to perform.
			DWORD operation;
			switch (operation = get_dword()) {
				case 1: //check syntax
				{
					// get the formula type
					FormulaType formulaType;
					switch (get_dword()) {
						case 0: //simple
							formulaType = kSimple;
							break;
						case 1: //table
							formulaType = kTable;
							break;
						default:
							throw io_exception();
					}
					string formulaStr = get_string();
					bool ok = false;
					try {
						Formula f;
						f.Compile(&formulaStr[0],formulaType);
						ok = true;
					} catch (exception& e) {
						// formula didn't compile, return an error
						send_dword(1);
						send_string("compile_error");
						send_string(e.what());
					}
					if (ok) {
						send_dword(0); // success
					}
					break;
				}
				case 2: // table_compute start
				{
					// first thing we get is a list of formulas, followed
					// by a list of documents. Each formula will be computed
					// on each document and results returned on a document
					// by document basis
					DWORD numFormulas = get_dword();
					free_formulas(formulas);
					formulas.resize(numFormulas);

					for(DWORD i=0; i<numFormulas; i++) {
						Formula* pFormula = new Formula();
						formulas[i] = pFormula;
						char* formulaStr = get_str();
						try {
							pFormula->Compile(formulaStr, kTable);
							send_dword(0);
						} catch (exception& e) {
							send_dword(1); // failure
							send_string("compile_error");
							send_string(e.what());
						}
					}
					break;
				}
				case 3: // table compute document
				{
					DefaultDocument doc(ma);
					get_doc(doc);
					// We have a document, now compute all the formulas on it
					
					//send the number of table formulas we'll be computing
					send_dword((DWORD)formulas.size());

					// now send the results
					for(DWORD i=0; i<formulas.size(); i++) {
						Formula* pFormula = formulas[i];
						if (pFormula->IsReady()) {
							pFormula->Reset();
							if (pFormula->SelectCompute(doc)) {
								send_dword(1); // selected
								//now send the column results
								const fvector<const List*>& values = pFormula->GetColumns();
								send_dword((DWORD)values.size()); // number of columns returned
								for (size_t j=0; j<values.size(); j++) {
									const List* pValue = values[j];
									send_list(*pValue);
								}
							}
							else {
								send_dword(0); // not selected
							}
						}
						else {
							// formula not ready, compile error earlier.
							// (this table will have no documents in it)
							send_dword(0);
						}
					}

					break;
				}

				case 4: // get column names
				{
					DWORD numFormulas = get_dword();
					vector<Formula>::iterator pFormula;
					for(DWORD i=0; i<numFormulas; i++) {
						Formula f;
						char* formulaStr = get_str();
						try {
							f.Compile(formulaStr, kTable);
							
							send_dword(1); // success
							// send the columns as a list
							fvector<TextElement> columns = f.GetColumnNames();
							// send the number of columns 
							send_dword((DWORD)columns.size());
							for(size_t j=0; j<columns.size(); j++) {
								send_dword(0); // indicate string
								send_string(columns[j]);
							}
						} catch (Formula::SyntaxException& e) {
							send_dword(0); // failure
							send_string("compile_error");
							send_string(e.what());
						}

					}
					break;
				}
				default:
					printf("Unknown operation: %i", operation);
					throw io_exception();	
			}
		}
	} catch(closed_pipe_exception&) {
		printf("Normal termination\r\n");
		// normal termination
		return 0;
	} catch(exception& e) {
		printf("Fatal exception: %s\r\n", e.what());
		return 1;
	} catch(...) {		
		printf("Fatal unknown exception\r\n");
		throw; // rethrow to allow OS crash debugging/tracing
	}
}

void free_formulas(vector<Formula*> formulas)
{
	for(size_t i=0; i<formulas.size(); i++) {
		delete formulas[i];
	}
}

void get_doc(Document& document)
{
	DWORD itemCount = get_dword();
	while(itemCount--) {
		char* fieldnameStr = get_str();
		List* pValue = new(ma) List(ma);
		get_list(*pValue);
		document.SetValue(TextElement(fieldnameStr), pValue);
	}
}


void get_list(List& list)
{
	DWORD count = get_dword();
	list.Reserve(count);
	while(count--) {
		DWORD type = get_dword();
		switch (type) {
			case 0:
			{
				char* str = get_str();
				list.Add(TextElement(str));
				break;
			}
			case 1:
			{
				list.Add(NumberElement(get_double()));
				break;
			}
			default:
				printf("Unknown element type: %u", type);
				throw io_exception();
		}
	}
}

void send_list(const List& list)
{
	DWORD count = (DWORD)list.size();
	send_dword(count);
	
	for(DWORD i=0; i<count; i++) {
		Element element = list[i];
		switch (element->GetType()) {			
			case NUMBER:
			{
				// preserve numbers
				send_dword(1);
				send_double(element->ToNumber());
				break;
			}
			default:
			{
				// everything else convert to text
				send_dword(0);
				send_string(element->ToText(ma));
				break;
			}
		}
	}
}


bool setup_io()
{
#ifdef WIN32
	// set io streams to binary
	if (setmode(0, _O_BINARY) == -1)
		return false;
	if (setmode(1, _O_BINARY) == -1)
		return false;
#endif

	// copy and save off stdout to a global var
	stdout_dup = dup(1);

	if (stdout_dup == -1)
		return false;

#ifdef _DEBUG
	// redirect stdout and stderr to a file
	FILE* fd;
	if (fd = fopen("fabric_debug.log","a")) {
		if (dup2(fileno(fd), 1) == -1)
			return false;
		if (dup2(fileno(fd), 2) == -1)
			return false;
	} else {
		// couldn't open the file, just redirect stdout to stderr
		if (dup2(2, 1) == -1)
			return false;
	}
#else
	// redirect stdout into stderr
    if (dup2(2, 1) == -1)
		return false;
#endif
	return true;
}

void get_packet(void* buffer, size_t len)
{
	size_t plen = packet_len();
	if (len != plen) {
		printf("incorrect packet length: %u", (DWORD)plen);
		throw io_exception();
	}
	read_exact(buffer, (DWORD)len);
}

string get_string()
{
	DWORD len = packet_len();
	string str;
	str.resize(len + 1);
	read_exact(&str[0], len);
	if (len == 1 && str[0] == 0) {
		str.clear();
	}
	return str;
}

char* get_str()
{
	DWORD len = packet_len();
	char* pBuffer = new(ma) char[len + 1];
	read_exact(pBuffer, len);
	pBuffer[len] = 0; // Null terminate
	return pBuffer;
}

void send_string(const char* buffer)
{
	send_string(buffer, strlen(buffer));
}

void send_string(const TextElement& text)
{
	send_string(text.Ptr(), text.Length());
}

void send_string(const char* buffer, size_t len)
{
	if (len == 0) {
		// Never send 0 bytes. Erlang barfs. Instead
		// send a single zeroed byte. the other side will intepret it
		// as a null string.
		send_packet("", 1);
	}
	else {
		send_packet(buffer, (DWORD)len);
	}
	
}

void read_exact(void *buf, DWORD len)
{
	DWORD i;
	DWORD got=0;

	do {
		if ((i = read(0, ((BYTE*)buf)+got, len-got)) <= 0) {
			throw closed_pipe_exception();
		}
		got += i;
	} while (got<len);  
}

void write_exact(const void *buf, DWORD len)
{
	DWORD i;
	DWORD wrote = 0;

	do {
		if ((i = write(stdout_dup, ((BYTE*)buf)+wrote, len-wrote)) <= 0) {
			throw closed_pipe_exception();
		}
		wrote += i;
	} while (wrote<len);
}

DWORD packet_len()
{
	DWORD len;
	read_exact(&len, sizeof(DWORD));
	len = endian_swap(len);
	return len;
}

void send_packet(const void *buf, DWORD len)
{
	DWORD bigEndianLen = endian_swap(len);
	
	write_exact(&bigEndianLen, sizeof(bigEndianLen));
	write_exact(buf, len);
}


DWORD endian_swap(DWORD nLongNumber)
{
#ifdef COUCH_BIG_ENDIAN 
	return nLongNumber;
#else
	return (((nLongNumber&0x000000FF)<<24)+((nLongNumber&0x0000FF00)<<8)+
		((nLongNumber&0x00FF0000)>>8)+((nLongNumber&0xFF000000)>>24));
#endif
}

#define DOUBLE_INDEX(x) (((unsigned char *)&internalEndianMagic)[7-(x)])

static const double internalEndianMagic = 7.949928895127363e-275;

double endian_net_double(double native_number)
{
#ifdef COUCH_BIG_ENDIAN
	return native_number;
#else
  double result;
  int i;

  ASSERT(sizeof(double) == 8);

  for (i = 0; i < (int)sizeof(native_number); i++) {
    ((unsigned char *)&result)[DOUBLE_INDEX(i)] = ((unsigned char *)&native_number)[i];
  }
  return result;
#endif
}


double endian_native_double(double net_number)
{
#ifdef COUCH_BIG_ENDIAN
	return net_number;
#else
  double result;
  int i;

  ASSERT(sizeof(double) == 8);

  for (i = 0; i < (int)sizeof(net_number); i++) {
    ((unsigned char *)&result)[i] = ((unsigned char *)&net_number)[DOUBLE_INDEX(i)];
  }
  return result;
#endif
}

double get_double()
{
	double dbl;
	get_packet(&dbl, sizeof(dbl));
	return endian_native_double(dbl);
}

void send_double(double dbl)
{
	dbl = endian_net_double(dbl);
	send_packet(&dbl, sizeof(dbl));
}

DWORD get_dword()
{
	DWORD dword;
	get_packet(&dword, sizeof(DWORD));
	return dword;
}

void send_dword(DWORD dword)
{
	send_packet(&dword, sizeof(DWORD));
}
