/* Double-linked List prototypes - original designed for
 * Airwave IRC Services by Arnd Behring and Dirk Schalge
 *
 * CouchDB C API (libcouchdb) - Copyright (c) 2006 Dirk Schalge
 *
 * Thanks to Damien Katz and all other involved developers for this great
 * database!
 *
 * See doc/LICENSE for licensing details.
 */

/* airwave services
 * list manager
 * 
 * CAUTION: if you use these functions while using threads, make sure you don't
 * 	    use them on global lists or at least protect the list with mutex
 * 	    locks before.
 * 
 * window pointers (after function call):
 * unchanged - window is not changed 
 * undefined - don't rely on the window pointer after the function call
 * <explanation> - see explanation
 */

#ifndef AW_HEADER_LIST
#define AW_HEADER_LIST

#ifndef boolean
#define boolean int
#endif

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef struct listelement_ ListElement;

struct listelement_
{
	ListElement *next, *prev;
	void *data;
};

typedef struct list_
{
	ListElement *start;
	ListElement *end;
	ListElement *window;
	int count;
} List;

/* list_init()
 * initializes a list
 * CAUTION: - modifies **list!
 * returns: - TRUE on success
 * 	    - FALSE if (*list) is not NULL (either already initialized or
 * 	      not set to NULL after declaration) (errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: undefined
 */
extern boolean list_init	( List **list );

/* list_cleanup()
 * removes all elements and frees the list. it doesn't free the data elements
 * returns: - TRUE on success
 * 	    - FALSE if (*list) is NULL (not initialized) (errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: undefined
 */
extern boolean list_cleanup	( List **list );

/* list_append()
 * appends new data to the end of a list
 * returns: - TRUE on success
 *          - FALSE on failure (errno is set accordingly)
 * window pointer: - on empty list: points to the start
 * 		   - on non-empty list: unchanged
 * 		   - failure: unchanged
 */
extern boolean list_append	( List *list, void *data );

/* list_prepend()
 * adds a new element to the beginning of a list
 * returns: - TRUE on success
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - on empty list: points to the start
 * 		   - on non-emtpy list: unchanged
 * 		   - failure: unchanged
 */
extern boolean list_prepend	( List *list, void *data );

/* list_insert()
 * inserts a new element at the given position. if the position is 0, the
 * entry will be prepended and appended if pos is < 0. if pos is greater than
 * the number of list elements, the element will also be appended.
 * returns: - TRUE on success
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - success: points to the element indicated by pos
 * 		   - failure: undefined
 */
extern boolean list_insert	( List *list, void *data, int pos );

/* list_prepend_window()
 * inserts a new element right before the current window position
 * returns: - TRUE on success
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - on emtpy list: points to the start
 * 		   - on non-empty list: unchanged (points right after the new
 * 		     element)
 * 		   - failure: unchanged
 */
extern boolean list_prepend_window	( List *list, void *data );

/* list_append_window()
 * inserts a new element right after the current window position
 * returns: - TRUE on success
 *          - FALSE on failure (errno is set accordingly)
 * window pointer: - on empty list: points to the start
 * 		   - on non-empty list: unchanged (points right before the new
 * 		     element)
 * 		   - failure: unchanged
 */
extern boolean list_append_window	( List *list, void *data );

/* list_remove_by_data()
 * removes an entry from the list. doesn't free the data element
 * returns: - TRUE on success
 * 	    - FALSE if no element carrying *data was found (errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - on multiple element list: points to the element after the
 * 					       remove one
 * 		   - on one element list: undefined
 * 		   - failure: undefined
 */
extern boolean list_remove_by_data	( List *list, void *data );

/* list_remove_by_pos()
 * removes an entry from the list. doesn't free the data element
 * returns: - TRUE on success
 * 	    - FALSE if the list is empty (errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - on multiple element list: points to the element after the
 * 					       removed one
 * 		   - on one element list: undefined
 * 		   - failure: undefined
 */
extern boolean list_remove_by_pos	( List *list, int pos );

/* list_remove_window()
 * removes the list element indicated by window. doesn't free the data element
 * returns: - TRUE on success
 * 	    - FALSE if the list is empty (errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - on multiple element list: points to the element after the
 * 					       removed one
 * 		   - on one element list: undefined (NULL)
 * 		   - failure: unchanged
 */
extern boolean list_remove_window	( List *list );

/* list_remove_all()
 * removes all elements. doesn't free the data elements
 * returns: - TRUE on success
 * 	    - FALSE if the list is empty (errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: undefined
 */
extern boolean list_remove_all	( List *list );

/* list_cat()
 * concatenate two lists (append all elements of list2 to list and leave list2
 * empty). list and list2 must not be identical
 * returns: - TRUE on success
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - success: undefined
 * 		   - failure: unchanged
 */
extern boolean list_cat		( List *list, List *list2 );

/* list_length()
 * returns: the length of a list or -1 if list is NULL
 * window pointer: unchanged
 */
extern int list_length	( List *list );

/* list_pos()
 * returns: - the position of an element in the given list
 * 	    - -1 if the element is not present (errno is set to 0)
 * 	    - -1 if an error occured (errno is set accordingly).
 * window pointer: - success: points to the found element
 * 		   - failure: undefined
 */
extern int list_pos		( List *list, void *data );

/* list_pos_window()
 * returns: - window position
 * 	    - -1 if the list is empty (errno is set to 0)
 * 	    - -1 on failure (errno is set accordingly)
 * window pointer: - success: unchanged
 * 		   - failure: undefined
 */
extern int list_pos_window	( List *list );

/* list_get()
 * returns: - the element at the given position in the list. if the position is
 * 	      greater than the list length the last element will be returned.
 *          - NULL if list is NULL or the list is empty
 * window pointer: - success: points to the element indicated by pos 
 * 		   - failure: undefined
 */
extern void *list_get		( List *list, int pos );

/* list_get_window()
 * returns: - the element at the current window position in the list.
 *	    - NULL if list is NULL or the list is empty
 * window pointer: unchanged
 */
extern void *list_get_window	( List *list );

/* list_foreach()
 * executes a given function on each list entry.
 * the first parameter of the callback function is the current data element,
 * *data_user may be user specified data (i.e. the pointer (*data) supplied to
 * list_foreach).
 * if the callback function returns 1, list_set_window_next() called and the
 * function proceeds to the next element.
 * if the callback fucntion returns 0, list_set_window_next() is _not_ called
 * so return this, if you called for example list_remove_window() in the
 * callback function.
 * if the return value of the callback function is negative the loop is
 * terminated and this value is returned by list_foreach().
 * returns: - 1 on success
 * 	    - 0 on failure (errno is set accordingly)
 * 	    - <0 if the callback function returned a negative value (the loop
 * 	         has been terminated)
 * window pointer: undefined
 */
extern int list_foreach		( List *list, int (*func) ( void *data,
				void *data_user ), void *data );

/* list_search()
 * executes a callback function on each list element, until the function 
 * returns TRUE. the first parameter of the callback function is the current
 * data element, *data_user may be user specified data (i.e. the pointer 
 * (*data) supplied to list_search).
 * returns: - the data element indicated by the search function
 * 	    - NULL if the list is empty or the element was not found (errno is
 * 	      set to 0)
 * 	    - NULL on failure (errno is set accordingly)
 * window pointer: - success: points to the found element
 * 		   - failure: undefined
 */
extern void *list_search	( List *list, boolean (*func) ( void *data,
				void *data_user ), void *data );

/* list_set_window()
 * sets the window to the element indicated by pos
 * if pos is < 1 or greater than the number of list elements the window is set
 * to the last element
 * returns: - TRUE on success
 * 	    - FALSE if the list is empty (errno is set to 0)
 * 	    - FALSE the list is NULL (errno is set accordingly)
 */
extern boolean list_set_window	( List *list, int pos );

/* list_set_window_start()
 * sets the window to the start of the list
 * returns: - TRUE on success
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - success: points to the start
 * 		   - failure: undefined
 */
extern boolean list_set_window_start	( List *list );

/* list_set_window_end()
 * set the window to the end of the list
 * returns: - TRUE on success
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - empty list: undefined
 * 		   - non-empty list: points to the end
 * 		   - failure: unchanged
 */
extern boolean list_set_window_end	( List *list );

/* list_set_window_next()
 * shifts the window to the next element
 * returns: - TRUE on success
 * 	    - FALSE if window already points to the last element in the list
 * 		(errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - success: points to the next element
 * 		   - failure: unchanged
 */
extern boolean list_set_window_next	( List *list );

/* list_set_window_prev()
 * shifts the window to the previous element
 * returns: - TRUE on success
 * 	    - FALSE if window already points to the last element in the list
 * 		(errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - success: points to the previous element
 * 		   - failure: unchanged
 */
extern boolean list_set_window_prev	( List *list );

/* list_set_window_by_data()
 * finds the list element carrying the indicated data element and lets the
 * window pointer point to it
 * returns: - TRUE on success
 * 	    - FALSE if no element could be found or if the list is empty
 * 		(errno is set to 0)
 * 	    - FALSE on failure (errno is set accordingly)
 * window pointer: - success: points to the found element
 * 		   - failure: undefined
 */
extern boolean list_set_window_by_data	( List *list, void *data );

#endif
