/* CouchDB libcurl functions
 *
 * CouchDB C API (libcouchdb) - Copyright (c) 2006 Dirk Schalge
 *
 * Thanks to Damien Katz and all other involved developers for this great
 * database!
 *
 * See doc/LICENSE for licensing details.
 */

#include "libcouchdb.h"
#include "couch.h"
#include "curl.h"
#include "func.h"

static void couch_curl_perform ( Couch *c, char *url, int type, char *upload );
static size_t couch_curl_readfunc ( void *ptr, size_t size, size_t nmemb, void *stream );
static size_t couch_curl_store ( void *ptr, size_t size, size_t nmemb, void *data );


void couch_curl_doc_post ( Couch *c, char *xml )
{
	char url[1024];

	if ( !c || !xml || !c->server || !c->db )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/" );
	strcat ( url, c->db );

	couch_curl_perform ( c, url, HTTP_POST, xml );
}

void couch_curl_doc_delete ( Couch *c, char *docid )
{
	char url[1024];

	if ( !c || !docid || !c->server || !c->db )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/" );
	strcat ( url, c->db );
	strcat ( url, "/" );
	strcat ( url, docid );

	couch_curl_perform ( c, url, HTTP_DELETE, NULL );
}

void couch_curl_doc_get ( Couch *c, char *docid )
{
	char url[1024];

	if ( !c || !docid || !c->server || !c->db )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/" );
	strcat ( url, c->db );
	strcat ( url, "/" );
	strcat ( url, docid );

	couch_curl_perform ( c, url, HTTP_GET, NULL );
}

void couch_curl_doc_get_all ( Couch *c )
{
	char url[1024];

	if ( !c || !c->server || !c->db )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/" );
	strcat ( url, c->db );
	strcat ( url, "/$all_docs" );

	couch_curl_perform ( c, url, HTTP_GET, NULL );
}

void couch_curl_db_put ( Couch *c, char *name )
{
	char url[1024];

	if ( !c || !name || !c->server )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/" );
	strcat ( url, name );
	
	couch_curl_perform ( c, url, HTTP_PUT, "." );
}

void couch_curl_db_delete ( Couch *c, char *name )
{
	char url[1024];

	if ( !c || !name || !c->server )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/" );
	strcat ( url, name );

	couch_curl_perform ( c, url, HTTP_DELETE, NULL );
}

void couch_curl_db_get_all ( Couch *c )
{
	char url[1024];

	if ( !c || !c->server )
	{
		errno = EINVAL;
		return;
	}

	/* create url */
	memset ( url, 0, 1024 );
	strcpy ( url, "http://" );
	strcat ( url, c->server );
	strcat ( url, "/$all_dbs" );

	couch_curl_perform ( c, url, HTTP_GET, NULL );
}

static void couch_curl_perform ( Couch *c, char *url, int type, char *upload )
{
	struct curl_slist *head = NULL;
	struct MemoryStruct bodybuf;
	int len = 0;

	if ( !c || !url || !c->curl )
	{
		errno = EINVAL;
		return;
	}

	/* cleanup old stuff */
	couch_result_clear ( c );
	memset ( &bodybuf, 0, sizeof(bodybuf) );

	/* upload len? */
	if ( upload )
		len = strlen ( upload );

	/* create http header */
	head = curl_slist_append ( head, "Content-Type: application/xml" );

	/* server adress */
	curl_easy_setopt ( c->curl, CURLOPT_URL, url );

	/* set http header */
	curl_easy_setopt ( c->curl, CURLOPT_HTTPHEADER, head );

	/* send all body-data to this function  */
	curl_easy_setopt ( c->curl, CURLOPT_WRITEFUNCTION, couch_curl_store );
	curl_easy_setopt ( c->curl, CURLOPT_WRITEDATA, (void *)&bodybuf );

	/* type specific settings */
	if ( type == HTTP_POST )
	{
		curl_easy_setopt ( c->curl, CURLOPT_POST, 1 );
		curl_easy_setopt ( c->curl, CURLOPT_POSTFIELDSIZE, len );

		curl_easy_setopt ( c->curl, CURLOPT_READFUNCTION, couch_curl_readfunc );
		curl_easy_setopt ( c->curl, CURLOPT_READDATA, upload );

	}
	else if ( type == HTTP_PUT )
	{
		curl_easy_setopt ( c->curl, CURLOPT_UPLOAD, TRUE );
		curl_easy_setopt ( c->curl, CURLOPT_PUT, 1 );

		curl_easy_setopt ( c->curl, CURLOPT_READFUNCTION, couch_curl_readfunc );
		curl_easy_setopt ( c->curl, CURLOPT_READDATA, upload );
	}
	else if ( type == HTTP_DELETE )
	{
		curl_easy_setopt ( c->curl, CURLOPT_CUSTOMREQUEST, "DELETE" );
	}

	/* extreme verbosity */
	curl_easy_setopt ( c->curl, CURLOPT_VERBOSE, 1 );

	/* send query now */
	curl_easy_perform ( c->curl );

	/* cleanup */
	curl_slist_free_all ( head );

	c->xml = bodybuf.memory;
}



static size_t couch_curl_readfunc ( void *ptr, size_t size, size_t nmemb, void *stream )
{
	int i = 0;
	static int wrote = 0, done = 0;
	char *dest = (char *)ptr, *src = (char *)stream;

	if ( !stream || done == 1 )
	{
		wrote = 0;
		done = 0;

		return 0;
	}

	while ( sizeof(char)*i < size*nmemb )
	{
		if ( src[wrote] == 0 )
		{
			done = 1;
			break;
		}

		dest[i] = src[wrote];

		wrote++;
		i++;
	}

	return i;
}

/* taken from libcurl c api examples:
 * http://curl.haxx.se/lxr/source/docs/examples/getinmemory.c
 */
static size_t couch_curl_store ( void *ptr, size_t size, size_t nmemb, void *data )
{
	size_t realsize = size * nmemb;
	struct MemoryStruct *mem = (struct MemoryStruct *)data;

	mem->memory = (char *)couch_realloc ( mem->memory, mem->size + realsize+1 );
	if ( mem->memory )
	{
		memcpy ( &(mem->memory[mem->size]), ptr, realsize );
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}

	return realsize;
}
