/* Double-linked List functions - orgiginal designed for
 * Airwave IRC Services by Arnd Behring and Dirk Schalge
 *
 * CouchDB C API (libcouchdb) - Copyright (c) 2006 Dirk Schalge
 *
 * Thanks to Damien Katz and all other involved developers for this great
 * database!
 *
 * See doc/LICENSE for licensing details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "list.h"

static boolean list_init_element ( ListElement **element, void *data );

boolean list_init ( List **list )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}

	if ( (*list) )
	{
		/* already initialized list */
		errno = 0;
		return TRUE;
	}

	/* allocate a new list */
	(*list) = calloc ( sizeof ( List ), 1 );
	if ( !(*list) )
		return FALSE;

	memset ( (*list), 0, sizeof ( List ) );
	
	return TRUE;
}

boolean list_cleanup ( List **list )
{
	if( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	if ( !(*list) )
	{
		/* uninitialized list */
		errno = 0;
		return FALSE;
	}

	list_set_window_start ( (*list) );
	
	while ( (*list)->window )
	{
		(*list)->start = (*list)->window;
		(*list)->window = (*list)->window->next;
		free ( (*list)->start );
	}

	free ( (*list) );
	(*list) = NULL;

	return TRUE;
}

boolean list_append ( List *list, void *data )
{
	ListElement *element = NULL;

	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}

	/* allocate new element */
	if ( !list_init_element ( &element, data ) )
		return FALSE;

	if ( !list->start )
	{
		/* empty list */
		list->start  = element;
		list->end    = element;
		list->window = element;
	}
	else
	{
		/* non-empty list */
		list->end->next = element;
		element->prev = list->end;
		list->end = element;
	}

	list->count++;

	return TRUE;
}

boolean list_prepend ( List *list, void *data )
{
	ListElement *element = NULL;

	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}

	/* allocate new element */
	if ( !list_init_element ( &element, data ) )
		return FALSE;

	if ( !list->start )
	{
		/* empty list */
		list->start  = element;
		list->end    = element;
		list->window = element;
	}
	else
	{
		/* non-empty list */
		list->start->prev = element;
		element->next = list->start;
		list->start = element;
	}

	list->count++;

	return TRUE;
}

boolean list_insert ( List *list, void *data, int pos )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	/* set the window to the given position and prepend the new element */
	list_set_window ( list, pos );
	if ( pos < 0 || pos >= list->count )
		return list_append_window ( list, data );
	else
		return list_prepend_window ( list, data );
}

boolean list_prepend_window ( List *list, void *data )
{
	ListElement *element = NULL;
	
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	/* allocate new element */
	if ( !list_init_element ( &element, data ) )
		return FALSE;

	if ( !list->start )
	{
		/* empty list */
		list->start  = element;
		list->end    = element;
		list->window = element;
	}
	else
	{
		if ( list->window->prev )
		{
			/* not first element in the list */
			element->prev = list->window->prev;
			element->prev->next = element;
		}
		else
		{
			/* first element in the list */
			list->start = element;
		}
		
		element->next = list->window;
		list->window->prev = element;
	}

	list->count++;
	
	return TRUE;
}

boolean list_append_window ( List *list, void *data )
{
	ListElement *element = NULL;
	
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	/* allocate new element */
	if ( !list_init_element ( &element, data ) )
		return FALSE;
	
	if ( !list->start )
	{
		/* empty list */
		list->start  = element;
		list->end    = element;
		list->window = element;
	}
	else
	{
		if ( list->window->next )
		{
			/* not last element in the list */
			element->next = list->window->next;
			element->next->prev = element;
		}
		else
		{
			/* last element in the list */
			list->end = element;
		}
		
		element->prev = list->window;
		list->window->next = element;
	}
	
	list->count++;
	
	return TRUE;
}

boolean list_remove_by_data ( List *list, void *data )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	if ( !list_set_window_by_data ( list, data ) )
		return FALSE;
	
	return ( list_remove_window ( list ) );
}

boolean list_remove_by_pos ( List *list, int pos )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	/* set the window to the given position and delete the element */
 	list_set_window ( list, pos );
	return list_remove_window ( list );
}

boolean list_remove_window ( List *list )
{
	ListElement *window = NULL;
	
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	if ( !list->start )
	{
		/* empty list */
		errno = 0;
		return FALSE;
	}

	window = list->window;

	if ( window->next )
	{
		/* normal case: the list window pointer will be set to the next
		 * element in the list.
		 */
		list->window = list->window->next;
	
		if ( window->prev )
		{
			/* the element is somewhere in the middle of the list */
			window->next->prev = window->prev;
			window->prev->next = window->next;
		}
		else
		{
			/* the element was the first in the list */
			window->next->prev = NULL;
			list->start = window->next;
		}
	}
	else
	{
		/* last-element-case: the list window pointer will be set
		 * to the last element in the list.
		 */
		
		if ( window->prev )
		{
			/* the element is at the end of the list but there
			 * are still other elements.
			 */
			list->window = window->prev;
			list->window->next = NULL;
			list->end = list->window;
		}
		else
		{
			/* the element was the only one in the list. */
			list->start = NULL;
			list->end = NULL;
			list->window = NULL;
		}
	}
	
	list->count--;
	free ( window );
	
	return TRUE;
}

boolean list_remove_all ( List *list )
{
	int x = 0, count = 0;
	
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	/* empty list */
	if ( !list->window )
	{
		errno = 0;
		return FALSE;
	}
	
	list_set_window_start ( list );
	
	for ( count = list->count, x = 0; x < count; x++ )
		list_remove_window ( list );
	
	return TRUE;
}

boolean list_cat ( List *list, List *list2 )
{
	if ( !list || !list2 || list == list2 )
	{
		errno = EINVAL;
		return FALSE;
	}

	/* list2 is empty -> we're done*/
	if ( !list2->start )
		return TRUE;

	if( !list->start )
	{
		/* list is empty */
		list->start = list2->start;
		list->end = list2->end;
		list->window = list2->window;
	}
	else
	{
		/* non-emtpy list -> append elements */
		list->end->next = list2->start;
		list2->start->prev = list->end;
		list->end = list2->end;
	}

	/* update list->count */
	list->count += list2->count;

	/* clear list2 */
	memset ( list2, 0, sizeof ( List ) );
	
	return TRUE;
}

int list_length ( List *list )
{
	if ( !list )
	{
		errno = EINVAL;
		return -1;
	}
	
	return list->count;
}

int list_pos ( List *list, void *data )
{
	int x = 0;
	
	if ( !list )
	{
		errno = EINVAL;
		return -1;
	}
	
	/* set to the beginning */
	list_set_window_start ( list );
	
	for ( x = 0; x < list->count; x++ )
	{
		if ( list->window->data == data )
			return x;
		
		list_set_window_next ( list );
	}

	/* we didn't find the element */
	errno = 0;
	return -1;
}

int list_pos_window ( List *list )
{
	int x = 0;
	
	ListElement *tmp = NULL;

	if ( !list )
	{
		errno = EINVAL;
		return -1;
	}
	
	tmp = list->window;
	
	/* set to the beginning */
	list_set_window_start ( list );
	
	for ( x = 0; x < list->count; x++ )
	{
		if ( list->window == tmp )
			return x;
		
		list_set_window_next ( list );
	}
	
	/* nothing found - should only happen on empty lists */
	errno = 0;
	return -1;
}

void *list_get ( List *list, int pos )
{
	if ( !list )
	{
		errno = EINVAL;
		return NULL;
	}
	
	if ( !list_set_window ( list, pos ) )
		return NULL;

	return list->window->data;
}

void *list_get_window ( List *list )
{
	if ( !list )
	{
		errno = EINVAL;
		return NULL;
	}
	
	if ( !list->window )
		return NULL;

	return list->window->data;
}

int list_foreach ( List *list, int (*func) ( void *data, void *data_user ),
			void *data )
{
	int retval = 0;
	ListElement *cur = NULL, *next = NULL;
	
	if ( !list || !func )
	{
		errno = EINVAL;
		return 0;
	}
	
	/* empty list */
	if ( !list->start )
		return 1;

	for ( cur = list->start; cur; cur = next )
	{
		next = cur->next;
		list->window = cur;

		retval = (*func) ( list->window->data, data );
		if ( retval < 0 )
			return retval;
	}

	return 1;
}

void *list_search ( List *list, boolean (*func) ( void *data,
			void *data_user ), void *data )
{
	int x = 0;
	
	if ( !list || !func )
	{
		errno = EINVAL;
		return NULL;
	}
	
	/* empty list */
	if ( !list->start )
	{
		errno = 0;
		return NULL;
	}
	
	/* set to the beginning */
	list_set_window_start ( list );
	
	for ( x = 1; x <= list->count; x++ )
	{
		if ( (*func) ( list->window->data, data ) )
			return list->window->data;
		
		list_set_window_next ( list );
	}
	
	/* nothing found :/ */
	errno = 0;
	return NULL;
}

boolean list_set_window ( List *list, int pos )
{
	int x = 0;

	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}

	if ( !list->start )
	{
		/* empty list */
		errno = 0;
		return FALSE;
	}
	
	/* for negative values and values greater than the length of the list
	 * we set the pointer to the end; count - 1 is a short cut, so that we
	 * don't have to run through the whole list...
	 */
	if ( pos < 0 || pos >= list->count - 1 )
	{
		list_set_window_end ( list );
		return TRUE;
	}
	
	/* set to the beginning */
	list_set_window_start ( list );

	/* FIXME/SUGGESTION: if pos is greater than count/2 start at the end */
	for ( x = 0; x < pos; x++ )
		list_set_window_next ( list );
	
	return TRUE;
}

boolean list_set_window_start ( List *list )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	list->window = list->start;
	
	return TRUE;
}

boolean list_set_window_end ( List *list )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	list->window = list->end;
	
	return TRUE;
}

boolean list_set_window_next ( List *list )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}

	/* we are already at the end of the list or the list is empty */
	if ( (list->window == list->end) )
	{
		errno = 0;
		return FALSE;
	}

	list->window = list->window->next;
	
	return TRUE;
}

boolean list_set_window_prev ( List *list )
{
	if ( !list )
	{
		errno = EINVAL;
		return FALSE;
	}
	
	/* we are already at the beginning of the list or the list is empty */
	if ( list->window == list->start )
	{
		errno = 0;
		return FALSE;
	}
	
	list->window = list->window->prev;
	
	return TRUE;
}

boolean list_set_window_by_data ( List *list, void *data )
{
	if ( !list || !data )
	{
		errno = EINVAL;
		return FALSE;
	}

	list_set_window_start ( list );
	while ( list->window )
	{
		if ( list->window->data == data )
			return TRUE;
		if ( !list_set_window_next ( list ) )
		{
			/* we reached the end and didn't find anything */
			errno = 0;
			return FALSE;
		}
	}

	/* the list was empty */
	errno = 0;
	return FALSE;
}

/* ------------------------------------------------------------------------- */

static boolean list_init_element ( ListElement **element, void *data )
{
	/* nothing to add */
	if ( !data || !element )
	{
		errno = EINVAL;
		return FALSE;
	}

	if ( (*element) )
	{
		errno = 0;
		return FALSE;
	}

	/* allocate new element */
	(*element) = calloc ( sizeof ( ListElement ), 1 );
	if ( !(*element) )
		return FALSE;

	memset ( (*element), 0, sizeof ( ListElement ) );

	(*element)->data = data;

	return TRUE;	
}
