/* CouchDB libxml2 functions
 *
 * CouchDB C API (libcouchdb) - Copyright (c) 2006 Dirk Schalge
 *
 * Thanks to Damien Katz and all other involved developers for this great
 * database!
 *
 * See doc/LICENSE for licensing details.
 */

#include "libcouchdb.h"
#include "couch.h"
#include "xml.h"
#include "func.h"

char *couch_xml_doc_save ( CouchDocument *d )
{
	xmlDocPtr doc = NULL;
	xmlNodePtr root = NULL;
	xmlChar *buf = NULL;
	int bufsize = 0;
	char *result = NULL;

	if ( !d )
	{
		errno = EINVAL;
		return NULL;
	}

	/* build the xml tree */
	doc = xmlNewDoc ( "1.0" );
	root = xmlNewNode ( NULL, "doc" );
	xmlDocSetRootElement ( doc, root );

	/* do we want to update an existing or create a named document? */
	if ( d->docid  )
	{
		xmlNewProp ( root, "id", d->docid );

		if ( d->revid )
			xmlNewProp ( root, "rev", d->revid );
	}

	list_foreach ( d->content, couch_xml_doc_save_field, root );

	/* xml tree complete. dump to buffer */
	xmlDocDumpFormatMemory ( doc, &buf, &bufsize, 1 );

	/* do some cleanups */
	xmlFreeDoc ( doc );

	/* copy buffer for further use */
	result = strndup ( buf, bufsize+1 );
	result[bufsize-1] = 0;
	puts ( result );

	return result;
}

int couch_xml_doc_save_field ( void *data, void *data_user )
{
	CouchDocumentField *field = NULL;
	xmlNodePtr root = NULL, node = NULL;

	if ( !data || !data_user )
		return 1;

	field = (CouchDocumentField *)data;
	root = (xmlNodePtr)data_user;

	node = xmlNewChild ( root, NULL, "field", NULL );
	xmlNewProp ( node, "name", field->name );

	list_foreach ( field->values, couch_xml_doc_save_field_content, node );

	return 1;
}

int couch_xml_doc_save_field_content ( void *data, void *data_user )
{
	CouchDocumentFieldContent *c = NULL;
	xmlNodePtr node = NULL;

	if ( !data || !data_user )
		return 1;

	c = (CouchDocumentFieldContent *)data;
	node = (xmlNodePtr)data_user;

	xmlNewChild ( node, NULL, c->type ? "number" : "text", c->value );

	return 1;
}

CouchDocument *couch_xml_parse_doc_get ( char *xml )
{
	CouchDocument *d = NULL;
	CouchDocumentField *field = NULL;
	xmlDoc *doc = NULL;
	xmlNode *root = NULL, *node1 = NULL, *node2 = NULL;
	char *name = NULL;
	

	if ( !xml )
	{
		errno = EINVAL;
		return NULL;
	}

	/* get the root element - this is the <doc> thingie*/
	doc = xmlParseDoc ( (const xmlChar *)xml );
	root = xmlDocGetRootElement ( doc );
	if ( !root )
		return NULL;

	/* the root element has to be <doc> */
	if ( !root->name )
		return NULL;

	if ( strcmp(root->name, "doc") )
		return NULL;

	d = couch_doc_create ( NULL );
	if ( !d )
		return NULL;

	d->docid = strdup ( xmlGetProp(root, "id") );
	d->revid = strdup ( xmlGetProp(root, "rev") );

	if ( !d->docid || !d->revid )
	{
		couch_doc_cleanup ( &d );
		return NULL;
	}

	/* okay, parse all children (fields) now */
	for ( node1 = root->children; node1; node1 = node1->next )
	{
		if ( !node1->name )
			continue;

		if ( strcmp(node1->name, "field") )
			continue;

		name = xmlGetProp ( node1, "name" );
		if ( !name )
			continue;

		field = couch_doc_field_create ( d, name );

		/* parse field content */
		for ( node2 = node1->children; node2; node2 = node2->next )
		{
			if ( !node2->name )
				continue;

			if ( !strcmp(node2->name, "text") )
				couch_doc_field_content_create ( field, 0, node2->children->content );
			else if ( !strcmp(node2->name, "number") )
				couch_doc_field_content_create ( field, 1, node2->children->content );
		}
	}


	/* cleanup */
	xmlFreeDoc ( doc );

	return d;
}


List *couch_xml_parse_doc_get_all ( char *xml )
{
	CouchDocument *d = NULL;
	List *l = NULL;
	xmlDoc *doc = NULL;
	xmlNode *root = NULL, *node = NULL;

	if ( !xml )
	{
		errno = EINVAL;
		return NULL;
	}

	/* get the root element - this is the <table> thingie*/
	doc = xmlParseDoc ( (const xmlChar *)xml );
	root = xmlDocGetRootElement ( doc );
	if ( !root )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	/* the root element has to be <table> */
	if ( !root->name )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	if ( strcmp(root->name, "table") )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	if ( !list_init(&l) )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	/* okay, parse all children (rows) now */
	for ( node = root->children; node; node = node->next )
	{
		if ( !node->name )
			continue;

		if ( strcmp(node->name, "row") )
			continue;


		d = couch_doc_create ( NULL );
		if ( !d )
			break;

		d->docid = strdup ( xmlGetProp(node, "docid") );
		d->revid = strdup ( xmlGetProp(node, "rev") );

		if ( !d->docid || !d->revid )
		{
			couch_doc_cleanup ( &d );
			continue;
		}

		if ( !list_append(l, d) )
		{
			couch_doc_cleanup ( &d );
			break;
		}
	}


	/* cleanup */
	xmlFreeDoc ( doc );

	return l;
}

List *couch_xml_parse_db_get_all ( char *xml )
{
	List *l = NULL;
	xmlDoc *doc = NULL;
	xmlNode *root = NULL, *node = NULL;
	char *db = NULL;

	if ( !xml )
	{
		errno = EINVAL;
		return NULL;
	}

	/* get the root element - this is the <dbs> thingie*/
	doc = xmlParseDoc ( (const xmlChar *)xml );
	root = xmlDocGetRootElement ( doc );
	if ( !root )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	/* the root element has to be <dbs> */
	if ( !root->name )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	if ( strcmp(root->name, "dbs") )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	if ( !list_init(&l) )
	{
		xmlFreeDoc ( doc );
		return NULL;
	}

	/* okay, parse all children (db) now */
	for ( node = root->children; node; node = node->next )
	{
		if ( !node->name )
			continue;

		if ( strcmp(node->name, "db") )
			continue;

		db = strdup ( node->children->content );
		if ( !db )
			break;

		if ( !list_append(l, db) )
			break;
	}

	/* cleanup */
	xmlFreeDoc ( doc );

	return l;
}

int couch_xml_parse_response ( Couch *c )
{
	CouchResult *r = NULL;
	xmlDoc *doc = NULL;
	xmlNode *root = NULL;

	if ( !c || !c->xml )
	{
		errno = EINVAL;
		return -1;
	}

	/* get the root element - this is the <xml> thingie*/
	doc = xmlParseDoc ( (const xmlChar *)c->xml );
	root = xmlDocGetRootElement ( doc );
	if ( !root )
	{
		xmlFreeDoc ( doc );
		return -1;
	}


	/* the root element has to be <success> or <error> */
	if ( !root->name )
	{
		xmlFreeDoc ( doc );
		return -1;
	}

	r = c->result;

	if ( !strcmp(root->name, "success") )
	{
		/* we had success! :-) */
		r->type = 1;

		/* do we have an updateinfo (documents) or just the success? */
		if ( root->children )
		{
			/* the first child must be the updateinfo */
			if ( root->children->name )
			{
				/* get doc & revision id */
				r->docid = strdup ( xmlGetProp(root->children, "docid") );
				r->revid = strdup ( xmlGetProp(root->children, "rev") );

				if ( !r->docid || !r->revid )
				{
					if ( r->docid )
					{
						free ( r->docid );
						r->docid = NULL;
					}

					if ( r->revid )
					{
						free ( r->revid );
						r->revid = NULL;
					}

					xmlFreeDoc ( doc );
					return -1;
				}
			}

			/* simply don't parse other children.. */
		}
	}
	else if ( !strcmp(root->name, "error") )
	{
		/* error occured :-/ */
		r->type = 0;

		/* every error has a reason, get it! */
		if ( !root->children )
		{
			xmlFreeDoc ( doc );
			return -1;
		}

		r->error = strdup ( root->children->content );
		if ( !r->error )
		{
			xmlFreeDoc ( doc );
			return -1;
		}

		/* ignore other xml stuff.. */
		xmlFreeDoc ( doc );
		return 0;
	}
	else
	{
		xmlFreeDoc ( doc );
		return -1;
	}

	/* cleanup */
	xmlFreeDoc ( doc );

	return 1;
}


List *couch_xml_parse_table ( char *xml )
{
	List *l = NULL;
	CouchDocument *d = NULL;
	CouchDocumentField *f = NULL;
	xmlDoc *doc = NULL;
	xmlNode *root = NULL, *node1 = NULL, *node2 = NULL, *node3 = NULL;
	char *name = NULL;

	if ( !xml )
	{
		errno = EINVAL;
		return NULL;
	}

	if ( !list_init(&l) )
		return NULL;

	/* get the root element - this is the <table> thingie */
	doc = xmlParseDoc ( (const xmlChar *)xml );
	root = xmlDocGetRootElement ( doc );
	if ( !root )
	{
		xmlFreeDoc ( doc );
		list_cleanup ( &l );
		return NULL;
	}

	/* the root element has to be <table> */
	if ( !root->name )
	{
		xmlFreeDoc ( doc );
		list_cleanup ( &l );
		return NULL;
	}

	if ( strcmp(root->name, "table") )
	{
		xmlFreeDoc ( doc );
		list_cleanup ( &l );
		return NULL;
	}

	/* level 1: parse rows */
	for ( node1 = root->children; node1; node1 = node1->next )
	{
		if ( !node1->name )
			continue;

		if ( strcmp(node1->name, "row") )
			continue;

		d = couch_doc_create ( NULL );
		if ( !d )
		{
			xmlFreeDoc ( doc );
			couch_doc_list_cleanup ( &l );
			return NULL;
		}

		d->docid = strdup ( xmlGetProp(node1, "docid") );
		d->revid = strdup ( xmlGetProp(node1, "rev" ) );

		if ( !d->docid || !d->revid )
		{
			xmlFreeDoc ( doc );
			couch_doc_list_cleanup ( &l );
			return NULL;
		}

		if ( !list_append(l, d) )
		{
			xmlFreeDoc ( doc );
			couch_doc_list_cleanup ( &l );
			return NULL;
		}

		/* level 2: parse columns */
		for ( node2 = node1->children; node2; node2 = node2->next )
		{
			if ( !node2->name )
				continue;

			if ( strcmp(node2->name, "column") )
				continue;

			name = xmlGetProp ( node2, "name" );
			if ( !name )
				continue;

			f = couch_doc_field_create ( d, name );

			/* level 3: parse content */
			for ( node3 = node2->children; node3; node3 = node3->next )
			{
				if ( !node3->name )
					continue;

				if ( !strcmp(node3->name, "text") )
					couch_doc_field_content_create ( f, 0, node3->children->content );
				else if ( !strcmp(node3->name, "number") )
					couch_doc_field_content_create ( f, 1, node3->children->content );
			}
		}
	}

	/* cleanup */
	xmlFreeDoc ( doc );

	return l;
}
